
/**********************************************************************
 * $Id: exampleCom.c,v 1.14 92/04/20 16:06:17 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 **********************************************************************/

#include <stdio.h>

#include <xerion/commands.h>
#include "exampleParse.h"

/***********************************************************************
 *	Name:		sumUnits
 *	Description:	increments an integer (address passed in as data)
 *			by the number of units in the group
 *	Parameters:	
 *		Group	group - the group to sum
 *		void	*data - a pointer to an int to increment.
 *	Return Value:	NONE
 ***********************************************************************/
static void	sumUnits (group, data)
  Group		group ;
  void		*data ;
{
  *(int *)data += group->numUnits ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		command_addExamples
 *	Description:	adds examples to a specified example set in
 *			the currentNet.
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_addExamples - 0 on failure, 1 on success
 ***********************************************************************/
int	command_addExamples (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  ExampleSet	exampleSet ;
  int		idx, newMask, setMask ;
  char		*fileName ;
  FILE		*fp ;
  int		numClamped, numTargets, numSlices  ;


  IUsage("-type <type> [ <file> ]") ;
  if (GiveHelp(tokc)) {
    ISynopsis("adds examples to the specified example set of a net") ;
    IHelp
      (IHelpArgs,
       "addExamples reads an input file of example records and  adds them to",
       "an example set. The example set is specified by <type>  which should",
       "be one of   TESTING, TRAINING, or  VALIDATION.  Each  record  in the",
       "input file should be of the form:",
       "",
       "\tinput_1 ... input_n , target_1 ... target_m ;",
       "",
       "where n is the number of inputs and  m the number of  targets.  Each",
       "example is added  to the end  of  the specified example  list in the",
       "current net.  If no input file is specified, standard input is read.",
       "The data file is free format.",
       "",
       "If the network to which the examples are being added is a time delay",
       "network each example  consists  of vectors  of  inputs followed  by",
       "vectors  of targets.  Each vector   must be  preceded  by the range,",
       "surrounded by  square brackets, of  time slices it applies  to (e.g.",
       "[1-4,6]). Special keywords are also allowed in the range:",
       "",
       "\tn-m    - the range of numbers n through m (m >= n)",
       "\tall    - specifies that all time slices use this vector.",
       "\tlast n - specifies that the last n time slices use this vector.",
       "\tnext n - specifies that the next n time slices use this vector.",
       "",
       "These keywords  may be combined, separated by   commas, in a  single",
       "range of units.",
       "",
       "There need not be one input and one target for every time  slice. If",
       "an  input vector  is  not specified for   a  particular  time, it is",
       "assumed that all input units are unclamped at that time.  Similarly,",
       "for  targets it  is assumed that  the units have no  target value at",
       "that time. To specify that an individual unit has  no input (target)",
       "at a given time, set its value to the string \"NaN\" (Not a Number).",
       "EXAMPLES",
       "",
       "\txerion-> addExamples -type TESTING file.ex",
       "",
       "For a normal (not time delay) network with 5 inputs  and  2 targets,",
       "file.ex may contain:",
       "",
       "\t0.0 1.0 0.0 0.0 1.0, 1.0 1.0 ;",
       "\t0.0 0.0 0.0 0.0 0.0, 1.0 0.0 ;",
       "",
       "For the same network, but with 10 time slices:",
       "",
       "\t[all]\t\t0.0 1.0 0.0 0.0 1.0,",
       "\t[last 3]\t1.0 1.0 ;",
       "",
       "\t[0]\t\t0.0 0.0 0.0 0.0 0.0",
       "\t[1]\t\t0.2 0.0 0.0 0.0 0.0",
       "\t[next 2]\t0.5 0.0 0.0 0.0 0.0",
       "\t[next 2]\t1.0 0.0 0.0 0.0 0.0,",
       "\t[0,last 2]\t1.0 0.0 ;",
       "",
       "SEE ALSO",
       "deleteExamples",
       NULL);
    return 0;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;
  
  if (tokc < 3 || tokc > 4)
    IErrorAbort(IPrintUsage(tokv[0], usage)) ;

  setMask = 0 ;
  for (idx = 1 ; 
       idx < tokc && 
       strncmp(tokv[idx], "-type", strlen(tokv[idx])) == 0 ; idx += 2) {
    newMask = findClassMask(EXAMPLESET_CLASS, tokv[idx+1]) ;
    if (newMask == 0) {
      IErrorAbort("Unknown example set type: \"%s\".", tokv[idx+1]) ;
      return 0 ;
    }    
    setMask |= newMask ;
  }
  
  if (setMask != TRAINING && setMask != VALIDATION && setMask != TESTING)
    IErrorAbort("Unknown example set type: \"0x%0x\".", setMask) ;
  
  exampleSet = netGetExampleSet(currentNet, setMask) ;

  fileName = NULL ;
  if (idx = tokc - 1)
    fileName = tokv[idx] ;
  else if (idx != tokc)
    IErrorAbort(IPrintUsage(tokv[0], usage)) ;

  /* set the proper input source */
  if (fileName == NULL) {
    fp = din ;
  } else {
    fp = IOpenFileOrAbort(fileName, "r", NULL) ;
  }

  if (exampleSet == NULL)
    IErrorAbort("No example set specified for net \"%s\".", currentNet->name) ;

  numClamped = 0 ;
  netForAllGroups(exampleSet->net, 
		  exampleSet->clampMask, sumUnits, &numClamped) ;

  numTargets = 0 ;
  netForAllGroups(exampleSet->net, 
		  exampleSet->targetMask, sumUnits, &numTargets) ;

  numSlices = exampleSet->net->timeSlices ;

  /* read the records, which are of the form:
   * input_1 input_2 ... input_n, target_1 target_2 ... target_m ;
   */
  readExampleFile(fp, exampleSet, numSlices, numClamped, numTargets) ;

  if (fp != din)
    fclose(fp) ;

  markToRebuildDisplay(EXAMPLE_DISPLAY) ;

  return 1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		command_deleteExamples
 *	Description:	clears all examples from the specified set in
 *			the currentNet.
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_deleteExamples - 0 on failure, 1 on success
 ***********************************************************************/
int	command_deleteExamples (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  int		idx, newMask, setMask ;

  IUsage("-type <type>") ;
  if (GiveHelp(tokc)) {
    ISynopsis("deletes examples from the specified example set") ;
    IHelp
      (IHelpArgs,
       "deleteExamples deletes all examples from the  specified example set.",
       "<type>  specifies which  example set to  use. It   should be  one of",
       "TRAINING, TESTING, or VALIDATION.",
       "EXAMPLE",
       "\txerion-> deleteExamples -type TRAINING",
       "",
       "removes all training examples from the current net.",
       "SEE ALSO",
       "addExamples",
       NULL);
    return 1;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  if (tokc != 3)
    IErrorAbort(IPrintUsage(tokv[0],usage));

  setMask = 0 ;
  for (idx = 1 ; 
       idx < tokc && 
       strncmp(tokv[idx], "-type", strlen(tokv[idx])) == 0 ; idx += 2) {
    newMask = findClassMask(EXAMPLESET_CLASS, tokv[idx+1]) ;
    if (newMask == 0) {
      IErrorAbort("Unknown example set type: \"%s\".", tokv[idx+1]) ;
      return 0 ;
    }    
    setMask |= newMask ;
  }
  
  if (setMask == TRAINING) {
    destroyExampleSet(currentNet->trainingExampleSet) ;
    if ((currentNet->trainingExampleSet 
	= createExampleSet("Training", TRAINING, currentNet)) == NULL)
      IErrorAbort("Cannot create exampleSet") ;
  } else if (setMask == TESTING) {
    destroyExampleSet(currentNet->testingExampleSet) ;
    if ((currentNet->testingExampleSet 
	 = createExampleSet("Testing", TESTING, currentNet)) == NULL)
      IErrorAbort("Cannot create exampleSet") ;
  } else if (setMask == VALIDATION) {
    destroyExampleSet(currentNet->validationExampleSet) ;
    if ((currentNet->validationExampleSet 
	 = createExampleSet("Validation", VALIDATION, currentNet)) == NULL)
      IErrorAbort("Cannot create exampleSet") ;
  } else {
    IErrorAbort("Unknown example set type: \"0x%0x\".", setMask) ;
  }
  
  markToRebuildDisplay(EXAMPLE_DISPLAY) ;

  return 1 ;
}
/**********************************************************************/
