
/**********************************************************************
 * $Id: actDisplay.c,v 1.20 92/04/20 16:07:09 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <math.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xutil.h>

#include <X11/Shell.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>

#include <xerion/Network.h>
#include <xerion/RowColumn.h>
#include <xerion/Selection.h>
#include <xerion/MenuPane.h>
#include <xerion/display.h>

#include "display2Itf.h"
#include "display2Sim.h"
#include "displayUtils.h"
#include "traceDisplay.h"
#include "check.h"

typedef enum FieldType { 
  activityField, 
  targetField 
}  FieldType ;

typedef struct {
  Widget	shell ;		/* the popup shell for the display 	*/
  String	titleFormat ;	/* the format for the title string	*/
  Widget	stepWidget ;	/* the label widget showing the step	*/
  String	stepLabel ;	/* the label for the time step widget	*/
  Widget	back[32] ;	/* the widget backgrounds available	*/
  XtStyle	unitStyle ;	/* the style to draw units in		*/
  Widget	set[32] ;	/* the example sets available		*/
  Widget	setLabelWidget ;/* the label for the example list	*/
  int		setIdx ;	/* the index for the set in use		*/
  Widget	listWidget ;	/* the example list			*/
  Widget	activationNet ;	/* the activation network		*/
  Widget	targetNet ;	/* the target network			*/
} ActivationDisplayRec, *ActivationDisplay ;

/* Routines for redisplaying the whole display */
static void 	resetDisplayCB ARGS((Widget, XtPointer, XtPointer)) ;
static void 	setSensitive   ARGS((ActivationDisplay)) ;

/* routines for redisplaying networks */
static void	resetNet       ARGS((Widget, FieldType)) ;
static void	rebuildNet     ARGS((Widget, UnitDisplayRec *, int)) ;

/* routines for handling the network appearance */
static void 	miscellaneousCB    ARGS((Widget, XtPointer, XtPointer)) ;
static void 	changeBackgroundCB ARGS((Widget, XtPointer, XtPointer)) ;
static void 	checkBackgroundCB  ARGS((Widget, XtPointer, XtPointer)) ;

/* routines for handling the examples */
static void	changeExampleSetCB ARGS((Widget, XtPointer, XtPointer)) ;
static void 	checkSetCB	   ARGS((Widget, XtPointer, XtPointer)) ;
static void	doExampleCB        ARGS((Widget, XtPointer, XtPointer)) ;
static void	resetExamples	   ARGS((Widget, int)) ;

/* routines for handling the time slices in recurrent nets */
static void 	incrementTimeCB ARGS((Widget, XtPointer, XtPointer)) ;
static void 	setTimeLabel    ARGS((ActivationDisplay)) ;

/* miscellaneous routines */
static void 	moveUnitCB 	 ARGS((Widget, XtPointer, XtPointer)) ;
static void 	changeActivityCB ARGS((Widget, XtPointer, XtPointer)) ;
static void 	changeTargetCB   ARGS((Widget, XtPointer, XtPointer)) ;
static void 	printActivityCB  ARGS((Widget, XtPointer, XtPointer)) ;
static void 	printTargetCB    ARGS((Widget, XtPointer, XtPointer)) ;
static void	popdownWidgetCB  ARGS((Widget, XtPointer, XtPointer)) ;
static void	doCheck          ARGS((Widget, int)) ;

static ActivationDisplay	actDisplay ;

static XtStyle	styleOption[] = { XtShadow, XtLines, XtLines2, XtWeave,
				    XtPlain, XtSolid } ;
static String	styleName[] = { "shadow", "lines", "lines2", "weave",
				  "plain", "solid" } ;

/***********************************************************************
 *	Name:		createActivationDisplay
 *	Description:	creates the window for the network display
 *			and all of the sub windows and necessary methods
 *	Parameters:	
 *		Widget	popup - the popup shell we are going in
 *	Return Value:	
 *		NONE
 ***********************************************************************/
export void	createActivationDisplay(popup)
  Widget	popup ;
{
  Widget	menuPane, menuButton, menu, line ;
  Widget	selection, net ;
  Widget	form, vpane, subForm, viewport, updateShell ;
  Widget	label, button ;
  String	*list, name ;
  int		idx, currentTime ;

  actDisplay = (ActivationDisplay)XtMalloc(sizeof(ActivationDisplayRec)) ;
  actDisplay->shell = popup ;
  
  /********************************************
   * Network Name label 
   */
  startArgs() ;
  actDisplay->titleFormat = getTitle(actDisplay->shell) ;
  getNetValue(NAME, &name) ;
  titlePrintf(actDisplay->shell, actDisplay->titleFormat, name) ;

  startArgs() ;
  form = createManagedWidget("form", formWidgetClass, popup) ;

  /********************************************
   * Display Updates 
   */
  updateShell = createTracePopup(popup, "updateActDisplay") ;

  /********************************************
   * Pulldown menu pane 
   */
  startArgs() ;
  menuPane = createManagedWidget("menuPane", menuPaneWidgetClass, form) ;

  /**********************
   * Options menu 
   */
  startArgs() ;
  setArg(XtNmenuName, "optionsMenu") ;
  menuButton = createManagedWidget("options", menuButtonWidgetClass, menuPane);

  startArgs() ;
  menu = createPopupShell("optionsMenu", simpleMenuWidgetClass, menuButton) ;
  
  startArgs() ;
  button = createManagedWidget("updates", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, popupCenteredCB, (XtPointer)updateShell) ;

  startArgs() ;
  button = createManagedWidget("reset", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, resetDisplayCB, (XtPointer)actDisplay) ;

  startArgs() ;
  line = createManagedWidget("line", smeLineObjectClass, menu) ;

  startArgs() ;
  button = createManagedWidget("close", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, popdownWidgetCB, (XtPointer)popup) ;


  /**********************
   * "Example Set" menu 
   */
  startArgs() ;
  setArg(XtNmenuName, "setMenu") ;
  menuButton = createManagedWidget("set", menuButtonWidgetClass, menuPane);

  startArgs() ;
  menu = createPopupShell("setMenu", simpleMenuWidgetClass, menuButton) ;
  XtAddCallback(menu, XtNpopupCallback, 
		checkSetCB, (XtPointer)actDisplay->set) ;
  
  list = listExampleSets() ;
  startArgs() ;
  setArg(XtNleftMargin, check_width) ;
  for (idx = 0 ; idx < 32 && list[idx] != NULL ; ++idx) {
    Widget	set ;
    set = createManagedWidget(list[idx], smeBSBObjectClass, menu) ;
    XtAddCallback(set, XtNcallback, changeExampleSetCB, (XtPointer)idx) ;
    actDisplay->set[idx] = set ;
  }
  actDisplay->set[idx] = NULL ;


  /**********************
   * Network menu 
   */
  startArgs() ;
  setArg(XtNmenuName, "networkMenu") ;
  menuButton = createManagedWidget("network",
				   menuButtonWidgetClass, menuPane);
  startArgs() ;
  menu = createPopupShell("networkMenu", simpleMenuWidgetClass, menuButton);
  XtAddCallback(menu, XtNpopupCallback, 
		checkBackgroundCB, (XtPointer)actDisplay->back) ;
  
  startArgs() ;
  setArg(XtNleftMargin, check_width) ;
  for (idx = 0 ; idx < XtNumber(styleOption) ; ++idx) {
    Widget	back ;
    back = createManagedWidget(styleName[idx], smeBSBObjectClass, menu) ;
    XtAddCallback(back, XtNcallback, 
		  changeBackgroundCB, (XtPointer)styleOption[idx]) ;
    actDisplay->back[idx] = back ;
  }
  actDisplay->back[idx] = NULL ;
  actDisplay->unitStyle = XtShadow ;

  startArgs() ;
  line = createManagedWidget("line", smeLineObjectClass, menu) ;

  startArgs() ;
  button = createManagedWidget("misc", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, miscellaneousCB, (XtPointer)actDisplay) ;

  /**********************
   * Time step button
   */
  startArgs() ;
  button = createManagedWidget("step", commandWidgetClass, menuPane) ;
  XtAddCallback(button, XtNcallback, incrementTimeCB, (XtPointer)actDisplay) ;
  actDisplay->stepWidget = button ;
  XtSetMappedWhenManaged(actDisplay->stepWidget, FALSE) ;

  actDisplay->stepLabel = getLabel(actDisplay->stepWidget) ;
  getNetValue(CURRENT_TIME, &currentTime) ;
  labelPrintf(actDisplay->stepWidget, actDisplay->stepLabel, currentTime) ;


  /********************************************
   * Example lists
   */
  actDisplay->setIdx = 0 ;
  /* Example Set list of examples */
  startArgs() ;
  setArg(XtNlist, listExamples(actDisplay->setIdx))  ;
  selection = createManagedWidget("exampleList", selectionWidgetClass, form) ;
  XtAddCallback(selection, XtNcallback, doExampleCB, (XtPointer)selection) ;
  actDisplay->listWidget = selection ;

  actDisplay->setLabelWidget = XtNameToWidget(selection, "listLabel") ;
  startArgs() ;
  setArg(XtNlabel, &name) ;
  getValues(actDisplay->set[actDisplay->setIdx]) ;
  labelPrintf(actDisplay->setLabelWidget, "%-10s", name) ;

  /********************************************
   * The Networks 
   */
  startArgs() ;
  setArg(XtNtop, 	XtChainTop) ;
  setArg(XtNbottom, 	XtChainBottom) ;
  setArg(XtNright,  	XtChainRight) ;
  vpane = createManagedWidget("vpane", panedWidgetClass, form) ;

  /**********************
   * Target Outputs 
   */
  startArgs() ;
  subForm  = createManagedWidget("form",	formWidgetClass,     vpane) ;
  label    = createManagedWidget("targetLabel",	labelWidgetClass,    subForm) ;
  viewport = createManagedWidget("targetViewport", 
				 viewportWidgetClass, subForm) ;
  net   = createManagedWidget("targets", networkWidgetClass,  viewport);
  XtAddCallback(net, XtNqueryCallback,	     printTargetCB, (XtPointer)net) ;
  XtAddCallback(net, XtNchangeValueCallback, 
		changeTargetCB,	(XtPointer)net) ;
  XtAddCallback(net, XtNchangeValueCallback, 
		resetDisplayCB,	(XtPointer)actDisplay) ;
  actDisplay->targetNet = net ;

  /**********************
   * Network Activation 
   */
  startArgs() ;
  subForm  = createManagedWidget("form", formWidgetClass, vpane) ;
  label    = createManagedWidget("activityLabel", labelWidgetClass, subForm) ;
  viewport = createManagedWidget("activityViewport", 
				 viewportWidgetClass, subForm) ;
  net 	   = createManagedWidget("activities", networkWidgetClass, viewport);
  XtAddCallback(net, XtNqueryCallback,  printActivityCB,(XtPointer)net) ;
  XtAddCallback(net, XtNmotionCallback, moveUnitCB,     (XtPointer)net) ;
  XtAddCallback(net, XtNmotionCallback, resetDisplayCB, (XtPointer)actDisplay);
  XtAddCallback(net, XtNchangeValueCallback,
		changeActivityCB,	(XtPointer)net) ;
  XtAddCallback(net, XtNchangeValueCallback, 
		resetDisplayCB,		(XtPointer)actDisplay) ;
  actDisplay->activationNet = net ;

  /********************************************
   * Do all the popup CB's for the first time 
   */
  XtAddCallback(popup, XtNpopupCallback, 
		resetDisplayCB, (XtPointer)actDisplay) ;

  resetDisplayCB(topLevel, (XtPointer)actDisplay, (XtPointer)NULL) ;
  XtInstallAllAccelerators(form, form) ;
  startArgs() ;
  setArg(XtNcreatePopupChildProc, NULL) ;
  setValues(popup) ;
}
/**********************************************************************/



/***********************************************************************
 *	Name:		updateActivationDisplay
 *	Description:	this little hack allows outside users to
 *			update the activation display without clicking
 *			or popping anything
 *	Parameters:	NONE
 *	Return Value:	NONE
 ***********************************************************************/
export void updateActivationDisplay()
{
  if (topLevel == NULL || actDisplay == NULL)
    return ;

  resetDisplayCB(topLevel, (XtPointer)actDisplay, (XtPointer)NULL) ;

  dispatchExposures() ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		doCheck
 *	Description:	puts a check beside a smeBSB (or clears it)
 *			depending on the flag passed in
 *	Parameters:
 *	  Widget	widget - the widget to put the check in
 *	  Boolean	flag   - TRUE - put check, FALSE - clear it
 *	Return Value:
 *	  static void	doCheck - NONE
 *********************************************************************/
static void	doCheck(widget, flag)
  Widget	widget ;
  int		flag ;
{
  static Pixmap	check = NULL ;

  if (check == NULL) {
    check = XCreateBitmapFromData(XtDisplayOfObject(widget),
				  XtWindowOfObject(widget), 
				  check_bits, check_width, check_height) ;
  }
  startArgs() ;
  setArg(XtNleftMargin, check_width) ;
  if (flag) {
    setArg(XtNleftBitmap, check) ;
  } else {
    setArg(XtNleftBitmap, NULL) ;
  }
  setValues(widget) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		changeBackgroundCB
 *	Description:	changes the unit background used in the networks
 *	Parameters:
 *	  Widget	widget     - UNUSED
 *	  XtPointer	clientData - the style of background to use
 *	  XtPointer	callData   - UNUSED
 *	Return Value:
 *	  static void	changeBackgroundCB - 
 *********************************************************************/
static void	changeBackgroundCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  actDisplay->unitStyle = (XtStyle)clientData ;

  rebuildActFlag = TRUE ;
  resetNet(actDisplay->activationNet, activityField) ;

  rebuildTargetFlag = TRUE ;
  resetNet(actDisplay->targetNet, targetField) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		checkBackgroundCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData - 
 *	Return Background:
 *	  static void	checkBackgroundCB - 
 *********************************************************************/
static void	checkBackgroundCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	*option = (Widget *)clientData ;
  int		idx ;

  for (idx = 0 ; option[idx] != NULL ; ++idx) {
    if (actDisplay->unitStyle == styleOption[idx])
      doCheck(option[idx], TRUE) ;
    else
      doCheck(option[idx], FALSE) ;
  }
}
/********************************************************************/


/*********************************************************************
 *	Name:		setTimeLabel
 *	Description:	gets the time slice of the network
 *	Parameters:
 *	  ActivationDisplay	display - the display object
 *	Return Value:
 *	  static void	incrementTimeCB - NULL 
 *********************************************************************/
static void		setTimeLabel(display)
  ActivationDisplay	display ;
{
  int		currentTime ;

  getNetValue(CURRENT_TIME, &currentTime) ;
  labelPrintf(display->stepWidget, display->stepLabel, currentTime) ;
  resetNet(display->activationNet, activityField) ;
  resetNet(display->targetNet,	   targetField) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		incrementTimeCB
 *	Description:	increments the time slice of the network
 *	Parameters:
 *	  Widget	widget - the widget callback is attached to
 *	  XtPointer	clientData - the display object
 *	  XtPointer	callData   - UNUSED
 *	Return Value:
 *	  static void	incrementTimeCB - NULL 
 *********************************************************************/
static void	incrementTimeCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  ActivationDisplay	display = (ActivationDisplay)clientData ;
  int			currentTime = incrementNetTime() ;

  labelPrintf(display->stepWidget, display->stepLabel, currentTime) ;
  resetNet(display->activationNet, activityField) ;
  resetNet(display->targetNet,	   targetField) ;
}
/********************************************************************/


/***********************************************************************
 *	Name:		resetNet
 *	Description:	rebuilds a network widget showing the target
 *			network outputs from the current net
 *	  Widget	network - the target network
 * 	  FieldType	type    - the field displayed. Either
 *				activityField or targetField
 *	Return Value:
 *		NONE
 ***********************************************************************/
static void	resetNet(network, type)
  Widget	network ;
  FieldType	type ;
{
  int		numUnits ;
  Boolean	*flagPtr ;

  static UnitDisplayRec	*unitArray = NULL ;
  static int		maxUnits = 0 ;

  numUnits = 0 ;
  /* We got a network, so calculate the necessary stuff */
  switch (type) {
  case activityField:
    flagPtr = &rebuildActFlag ;
    if (netExists())
      unitArray = getNetworkActivations(unitArray, &numUnits, &maxUnits) ;
    break ;
  case targetField:
  default:
    flagPtr = &rebuildTargetFlag ;
    if (netExists())
      unitArray = getNetworkTargets(unitArray, &numUnits, &maxUnits) ;
    break ;
  } 

  if (numUnits == 0) {
    /* No network, blank out the widget */
    rebuildNet(network, NULL, 0) ;
  } else if (*flagPtr == True) {
    rebuildNet(network, unitArray, numUnits) ;
    *flagPtr = False ;
  } else {
    XtNetworkSetUnits(network, unitArray) ;
  }
}
/**********************************************************************/


/***********************************************************************
 *	Name:		rebuildNet
 *	Description:	rebuilds the units and size of a network widget
 *	  Widget		network    - the network widget to update
 *	  UnitDisplayRec	*unitArray - the array of units to  update with
 *	  int		numUnits   - the number of units
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void	rebuildNet(network, unitArray, numUnits)
  Widget		network ;
  UnitDisplayRec	*unitArray ;
  int			numUnits ;
{
  int		height, width, margin ;

  getNetValue(ACTIVATION_HEIGHT, &height) ;
  getNetValue(ACTIVATION_WIDTH,  &width) ;
  getNetValue(ACTIVATION_MARGIN, &margin) ;

  XtUnmanageChild(network) ;
  startArgs() ;
  setArg(XtNunitBackgroundStyle, actDisplay->unitStyle) ;
  setArg(XtNunitHeight,	height) ;
  setArg(XtNunitWidth, 	width) ;
  setArg(XtNunitSep,	margin) ;
  setArg(XtNnumUnits,	numUnits) ;
  setArg(XtNunitArray,	unitArray) ;
  setValues(network) ;
  XtManageChild(network) ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		popdownWidgetCB
 *	Description:	pops down a popup shell which is passed in as
 *			the client data
 *	Parameters:
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the shell to pop down
 *	  XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	popdownWidgetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	popup = (Widget)clientData ;
  XtPopdown(popup) ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		checkSetCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData - 
 *	Return Value:
 *	  static void	checkFieldCB - 
 *********************************************************************/
static void	checkSetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	*option = (Widget *)clientData ;
  int		idx ;

  for (idx = 0 ; option[idx] != NULL ; ++idx) {
    if (actDisplay->setIdx == idx)
      doCheck(option[idx], TRUE) ;
    else
      doCheck(option[idx], FALSE) ;
  }
}
/********************************************************************/


/**********************************************************************
 *	Name:		changeExampleSetCB
 *	Description:	sets the index of what ExampleSet to get examples
 *			from
 *	Parameters:
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - NONE
 *	  XtPointer	callData   - the call data (XawListReturnStruct *)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	changeExampleSetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  String	label ;

  startArgs() ;
  setArg(XtNlabel, &label) ;
  getValues(widget) ;
  labelPrintf(actDisplay->setLabelWidget, "%-10s", label) ;

  actDisplay->setIdx = (int)clientData ;
  rebuildExampleFlag = True ;
  resetExamples(actDisplay->listWidget, actDisplay->setIdx) ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		setSensitive
 *	Description:	sets the sensitive field to true or false for
 *			all of the widgets that care.
 *	Parameters:
 *	  ActivationDisplay	display - the display
 *	Return Value:
 *	  static void	setSensitiveCB - NONE
 *********************************************************************/
static void	setSensitive(display)
  ActivationDisplay	display ;
{
  Widget	*set     = display->set ;
  Widget	menuPane = XtNameToWidget(display->shell, "form.menuPane") ;
  int		idx, timeSlices ;

  /* if no network, can't set the background of units */
  startArgs() ;
  if (netExists()) {
    setArg(XtNsensitive, TRUE) ;
  } else {
    setArg(XtNsensitive, FALSE) ;
  }
  setValues(XtNameToWidget(menuPane, "network")) ;
  setValues(XtNameToWidget(menuPane, "set")) ;

  /* if no network, or no timeslices, can't step through time */
#if 0
  if (XtIsRealized(display->stepWidget)) {
    if (netExists())
      getNetValue(TIME_SLICES, &timeSlices) ;
    else
      timeSlices = 0 ;

    if (timeSlices > 1) 
      XtMapWidget(display->stepWidget) ;
    else
      XtUnmapWidget(display->stepWidget) ;
  }
#endif

  /* check which examples sets are empty */
  for (idx = 0 ; set[idx] != NULL ; ++idx) {
    String	*list = listExamples(idx) ;
    startArgs() ;
    if (list == NULL || *list == NULL) {
      setArg(XtNsensitive, FALSE) ;
    } else {
      setArg(XtNsensitive, TRUE) ;
    }
    setValues(set[idx]) ;
  }
  /* make sure that we reset the list */
  listExamples(display->setIdx) ;
}
/********************************************************************/


/***********************************************************************
 *	Name:		resetExamples
 *	Description:	rebuilds the list of examples
 *		Widget	listWidget - the list (or selection) widget to update
 *		int	index - the index of the example set to use
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void	resetExamples(listWidget, idx)
  Widget	listWidget ;
  int		idx ;
{
  if (rebuildExampleFlag == True) {
    rebuildExampleFlag = False ;
    startArgs() ;
    setArg(XtNlist, listExamples(idx)) ;
    setValues(listWidget) ;
  }
}
/**********************************************************************/


/***********************************************************************
 *	Name:		doExampleCB
 *	Description:	performs an example on the current net. It
 *			gets the index of the example to do from
 *			the callData which must be XawListReturnStruct *
 *	Parameters:	
 *	  Widget	widget 	   - the list the callback is attached to
 *	  XtPointer	clientData - unused
 *	  XtPointer	callData   - the call data (XawListReturnStruct *)
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void	doExampleCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  XawListReturnStruct *returnStruct = (XawListReturnStruct *)callData ;

  if (returnStruct == NULL || returnStruct->list_index == XAW_LIST_NONE)
    return ;

  doExample(actDisplay->setIdx, returnStruct->list_index) ;

  setTimeLabel(actDisplay) ;
  resetNet(actDisplay->activationNet, activityField) ;
  resetNet(actDisplay->targetNet,     targetField) ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		printActivityCB
 *	Description:	prints out the activation level of a unit
 *	Parameters:	
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the network widget
 *	  XtPointer	callData   - NetworkReturnStruct *
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void printActivityCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget		network = (Widget)clientData ;
  NetworkReturnStruct	*returnStruct = (NetworkReturnStruct *)callData ;
  UnitDisplayRec	*UnitPtr ;

  UnitPtr = XtNetworkGetUnit(network, returnStruct->index) ;
  if (UnitPtr == NULL) {
    fprintf(stderr, "Yow! This unit doesn't seem to exist!\n") ;
  } else {
    fprintf(stdout, "\nUnit \"%s\" has activation %.4g\n",
	    (char *)UnitPtr->userData, UnitPtr->value) ;
  }
  XtFree((char *)UnitPtr) ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		printTargetCB
 *	Description:	prints out the target level of a unit
 *	Parameters:	
 *		Widget		widget - the widget the callback is attached to
 *		XtPointer	clientData - the network widget
 *		XtPointer	callData   - NetworkReturnStruct *
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void printTargetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget		network = (Widget)clientData ;
  NetworkReturnStruct	*returnStruct = (NetworkReturnStruct *)callData ;
  UnitDisplayRec	*UnitPtr ;

  UnitPtr = XtNetworkGetUnit(network, returnStruct->index) ;
  if (UnitPtr == NULL) {
    fprintf(stderr, "Yow! This unit doesn't seem to exist!\n") ;
  } else {
    fprintf(stdout, "\nUnit \"%s\" has target %.4g\n",
	    (char *)UnitPtr->userData, UnitPtr->value) ;
  }
  XtFree((char *)UnitPtr) ;
}
/**********************************************************************/
static void moveUnitCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget		network = (Widget)clientData ;
  NetworkReturnStruct	*returnStruct = (NetworkReturnStruct *)callData ;
  UnitDisplayRec	*unitPtr ;

  unitPtr = XtNetworkGetUnit(network, returnStruct->index) ;
  moveUnit(unitPtr) ;
}
/**********************************************************************/
static void changeActivityCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget		network = (Widget)clientData ;
  NetworkReturnStruct	*returnStruct = (NetworkReturnStruct *)callData ;
  UnitDisplayRec	*unitPtr ;

  unitPtr = XtNetworkGetUnit(network, returnStruct->index) ;
  changeActivity(unitPtr) ;
  updateNetActivities() ;
}
/**********************************************************************/
static void changeTargetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget		network = (Widget)clientData ;
  NetworkReturnStruct	*returnStruct = (NetworkReturnStruct *)callData ;
  UnitDisplayRec	*unitPtr ;

  unitPtr = XtNetworkGetUnit(network, returnStruct->index) ;
  changeTarget(unitPtr) ;
  updateNetActivities() ;
}
/*********************************************************************
 *	Name:		resetDisplayCB
 *	Description:	resets the display (i.e. sets the sensitivity
 *			of the widgets and rebuilds nets and example
 *			lists
 *	Parameters:
 *	  Widget	widget     - the widget the cb is attached to
 *	  XtPointer	clientData - UNUSED
 *	  XtPointer	callData   - UNUSED
 *	Return Value:
 *	  static void resetDisplayCB - NONE
 *********************************************************************/
static void resetDisplayCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  String	name ;

  getNetValue(NAME, &name) ;
  titlePrintf(actDisplay->shell, actDisplay->titleFormat, name) ;

  setSensitive(actDisplay) ;
  resetNet(actDisplay->activationNet,   activityField) ;
  resetNet(actDisplay->targetNet,	targetField) ;
  resetExamples(actDisplay->listWidget, actDisplay->setIdx) ;
}
/********************************************************************/

typedef struct {
  Widget	shell ;
  Widget	fixedMax ;
  Widget	maxValue ;
  Widget	size ;
  Widget	border ;
  Widget	margin ;
  Widget	activationNet ;
  Widget	targetNet ;
} *MiscNetworkDialog ;

/*********************************************************************
 *	Name:		setMiscellaneousCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData -
 *	Return Value:
 *	  static void	setMiscellaneousCB - 
 *********************************************************************/
static void	setMiscellaneousCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  MiscNetworkDialog	dialog = (MiscNetworkDialog)clientData ;
  Boolean		fixedMax ;
  int			size, margin, border ;
  float			maxValue ;
  String		string ;

  string = getText(dialog->maxValue) ;
  if (string != NULL)
    maxValue = atof(string) ;
  else
    maxValue = 0.0 ;

  startArgs() ;
  setArg(XtNstate, &fixedMax) ;
  getValues(dialog->fixedMax) ;

  string = getText(dialog->size) ;
  if (string != NULL) {
    size = atol(string) ;
    setNetValue(ACTIVATION_WIDTH,  &size) ;
    setNetValue(ACTIVATION_HEIGHT, &size) ;
  } else {
    getNetValue(ACTIVATION_WIDTH,  &size) ;
    getNetValue(ACTIVATION_HEIGHT, &size) ;
  }

  string = getText(dialog->border) ;
  if (string != NULL)
    border = atol(string) ;
  else
    XtVaGetValues(dialog->targetNet, XtNunitBorderWidth, &border, NULL) ;

  string = getText(dialog->margin) ;
  if (string != NULL) {
    margin = atol(string) ;
    setNetValue(ACTIVATION_MARGIN,  &margin) ;
  } else {
    getNetValue(ACTIVATION_MARGIN,  &margin) ;
  }

  startArgs() ;
  setArg(XtNfixedMax,	fixedMax) ;
  setArg(XtNmaxValue,	&maxValue) ;
  setArg(XtNunitHeight,	size) ;
  setArg(XtNunitWidth,	size) ;
  setArg(XtNunitSep,	margin) ;
  setArg(XtNunitBorderWidth,	border) ;

  setValues(dialog->activationNet) ;
  setValues(dialog->targetNet) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		miscellaneousCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData - 
 *	Return Value:
 *	  static void	miscellaneousCB - 
 *********************************************************************/
static void	miscellaneousCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  ActivationDisplay	display = (ActivationDisplay)clientData ;
  float		*max ;
  Dimension	size, margin, border ;
  Boolean	fixedMax ;

  static MiscNetworkDialog	dialog = NULL ;

  if (dialog == NULL) {
    Widget	form, button, table ;

    dialog = (MiscNetworkDialog)XtMalloc(sizeof(*dialog)) ;

    startArgs() ;
    dialog->shell = createPopupShell("miscShell", 
				     transientShellWidgetClass, 
				     display->shell) ;
    startArgs() ;
    form = createManagedWidget("form", formWidgetClass, dialog->shell) ;

    table = startTable(form) ;
    dialog->maxValue = addTableItem("maxValue", &table, TRUE) ;
    dialog->size     = addTableItem("size",   	&table, FALSE) ;
    dialog->border   = addTableItem("border",	&table, FALSE) ;
    dialog->margin   = addTableItem("margin",	&table, FALSE) ;

    dialog->fixedMax = XtNameToWidget(form, "maxValueToggle") ;

    startArgs() ;
    setArg(XtNfromVert,	dialog->margin) ;
    button = createManagedWidget("ok", commandWidgetClass, form) ;
    XtAddCallback(button, XtNcallback, setMiscellaneousCB, (XtPointer)dialog) ;
    XtAddCallback(button, XtNcallback,
		  popdownWidgetCB, (XtPointer)dialog->shell) ;

    startArgs() ;
    setArg(XtNfromVert,	dialog->margin) ;
    setArg(XtNfromHoriz,button) ;
    button = createManagedWidget("cancel", commandWidgetClass, form) ;
    XtAddCallback(button, XtNcallback, 
		  popdownWidgetCB, (XtPointer)dialog->shell) ;

    XtInstallAllAccelerators(form, form) ;
  }
  dialog->activationNet = display->activationNet ;
  dialog->targetNet     = display->targetNet ;

  startArgs() ;
  setArg(XtNfixedMax,	&fixedMax) ;
  setArg(XtNmaxValue,	&max) ;
  setArg(XtNunitHeight, &size) ;
  setArg(XtNunitSep,	&margin) ;
  setArg(XtNunitBorderWidth,	&border) ;
  getValues(dialog->activationNet) ;

  startArgs() ;
  setArg(XtNstate, fixedMax) ;
  setValues(dialog->fixedMax) ;
    
  textPrintf(dialog->maxValue,	"%g", *max) ;
  textPrintf(dialog->size,	"%d", size) ;
  textPrintf(dialog->border,	"%d", border) ;
  textPrintf(dialog->margin,	"%d", margin) ;

  popupCentered(dialog->shell, XtGrabExclusive) ;
}
/********************************************************************/
