
/**********************************************************************
 * $Id: display.c,v 1.20 92/04/20 16:07:21 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <stdio.h>

#include <X11/ShellP.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Command.h>
#include <X11/cursorfont.h>

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Toggle.h>

/* necessary to circumvent taps private assert routines */
#define _ASSERT_
#include <xerion/useful.h>
#include <xerion/display.h>

#include "display2Itf.h"
#include "actDisplay.h"
#include "weightDisplay.h"
#include "graphDisplay.h"
#include "varDisplay.h"
#include "traceDisplay.h"

extern void	createMinimizeDisplay ARGS((Widget)) ;

extern int	IDoSingleCommand ARGS((FILE *inStream, FILE *outStream)) ;
extern int	updateActivationDisplay() ;
extern int	updateWeightDisplay() ;
extern int	updateMinimizeDisplay() ;
extern int	updateVarDisplays() ;


/* Private functions */
static int	enableCommandLineInput ARGS((FILE *in, FILE *out)) ;
static Window	createBusyWindow       ARGS((Widget)) ;
static Boolean	doWorkProc	       ARGS((XtPointer	clientData)) ;

static Widget	createMainWindow   ARGS((Widget)) ;
static void	commandLineCB      ARGS((XtPointer, int *, XtInputId *)) ;
static void	popupWidgetCB	   ARGS((Widget, XtPointer, XtPointer)) ;
static void	popdownWidgetCB	   ARGS((Widget, XtPointer, XtPointer)) ;
static void	popupRaiseWidgetCB ARGS((Widget, XtPointer, XtPointer)) ;
static void	unselectCB	   ARGS((Widget, XtPointer, XtPointer)) ;
static void	quitCB		   ARGS((Widget, XtPointer, XtPointer)) ;
static void	confirmQuitCB	   ARGS((Widget, XtPointer, XtPointer)) ;
static Boolean	createPopupChild   ARGS((XtPointer clientData)) ;

/* Top Level widget for the application */
Widget		topLevel ;
Boolean		rebuildActFlag     = True ;
Boolean		rebuildExampleFlag = True ;
Boolean		rebuildWeightFlag  = True ;
Boolean		rebuildTargetFlag  = True ;
void		blockDeviceEvents() ;
void		unblockDeviceEvents() ;

/* input id for command line input */
static XtInputId	commandId ;

#define MAX_SHELLS	20
static Widget	topLevelShell[MAX_SHELLS] ;

/* 
 * these are passed in from the itf routines 
 * I'd rather not have them here, but there is no easy
 * way around it right now.
 */
static FILE		*InStream, *OutStream ;

/**********************************************************************
 *	Name:		initGraphics
 *	Description:	initializes the Xt toolkit and opens the 
 *			display. We can't use XtAppInitialize
 *			because it exits when it can't open the display
 *			and we want to continue, but without graphics.
 *	Parameters:
 *		int	*argc  - pointer to command line argc
 *		char	**argv - command line argv
 *	Return Value:
 *		int	initGraphics - 0 on failure, 1 on success
 **********************************************************************/
int	initGraphics(argc, argv)
  int		*argc ;
  char		**argv ;
{
  XtAppContext	appContext ;
  Display	*display ;

  XtToolkitInitialize() ;
  appContext = XtCreateApplicationContext() ;
  display    = XtOpenDisplay(appContext, NULL,
			     argv[0], "Xerion", NULL, 0, argc, argv) ;

  if (display == NULL) {
    XtDestroyApplicationContext(appContext) ;
    fprintf (stderr, "Unable to open display, graphics disabled.\n") ;
    return 0 ;
  }

  startArgs() ;
  topLevel = createAppShell(argv[0], "Xerion", 
			    applicationShellWidgetClass, display) ;
  return 1 ;
}
/**********************************************************************/


typedef struct {
  void		(*proc)	ARGS((XtPointer)) ;
  XtPointer	clientData ;
} *WorkProc ;

/*********************************************************************
 *	Name:		doWorkProc
 *	Description:	does a registered work Procedure
 *	Parameters:
 *	  XtPointer	clientData - a pointer to a WorkProc data
 *			structure.
 *	Return Value:
 *	  static Boolean	doWorkProc - Always TRUE, never repeat
 *********************************************************************/
static Boolean	doWorkProc(clientData)
  XtPointer	clientData ;
{
  WorkProc	workProc = (WorkProc)clientData ;

  workProc->proc(workProc->clientData) ;

  XtFree((void *)workProc) ;

  return TRUE ;
}
/********************************************************************/
  

/*********************************************************************
 *	Name:		addWorkProc
 *	Description:	adds a work procedure (or does it if topLevel
 *			is NULL)
 *	Parameters:
 *	  void		(*proc)(XtPointer) - the proc to do.
 *	  XtPointer	clientData - the data to pass to id
 *	Return Value:
 *	  int	addWorkProc - 1
 *********************************************************************/
int	addWorkProc(proc, clientData)
  void		(*proc)	ARGS((XtPointer)) ;
  XtPointer	clientData ;
{
  WorkProc	workProc = (WorkProc)XtMalloc(sizeof(*workProc)) ;

  workProc->proc = proc ;
  workProc->clientData = clientData ;

  if (topLevel)
    XtAppAddWorkProc(XtWidgetToApplicationContext(topLevel),
		     doWorkProc, workProc) ;
  else
    doWorkProc(workProc) ;

  return 1 ;
}
/********************************************************************/

/***********************************************************************
 *	Name:		buildGraphics
 *	Description:	builds the graphical displays for the simulator.
 *			IInit and initGraphics MUST be called before this 
 *			is called. NOT safe to call multiple times unless
 *			destroyGraphics has been called.
 *	Parameters:	NONE
 *	Return Value:	
 *		int	buildGraphics - 0 on failure, 1 on success
 ***********************************************************************/
int	buildGraphics() 
{
  
  if (topLevel == NULL)
    return 0 ;

  createMainWindow(topLevel) ;
  XtRealizeWidget(topLevel);
  return 1 ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		destroyGraphics
 *	Description:	unmanages and destroys the topLevel widget and
 *			all its children, then sets topLevel to NULL
 *			and closes the display.
 *	Parameters:	NONE
 *	Return Value:
 *	  int	destroyGraphics - 0 if topLevel == NULL, 1 otherwise
 *********************************************************************/
int	destroyGraphics() 
{
  int		idx ;
  Display	*display ;

  if (topLevel == NULL)
    return 0 ;

  /* get the display and save it, we'll need it later */
  display = XtDisplay(topLevel) ;

  /* unmap all the top level shells */
  for (idx = 0 ; idx < XtNumber(topLevelShell) ; ++idx) {
    if (topLevelShell[idx] != NULL && XtIsRealized(topLevelShell[idx])) {
      XtUnmapWidget(topLevelShell[idx]) ;
    }
    topLevelShell[idx] = NULL ;
  }

  /* Now destroy the top levelest and close the display */
  XtDestroyWidget(topLevel) ;
  XFlush(display) ;
  XtCloseDisplay(display) ;
  topLevel = NULL ;

  return 1 ;
}
/********************************************************************/


/**********************************************************************
 *	Name:		graphicsLoop
 *	Description:	the main event loop for using with the graphics.
 *			It also initializes the alternate input stream.
 *	Parameters:
 *		FILE	*inStream  - the stream for input (NULL if you
 *					don't want one)
 *		FILE	*outStream - the stream for output
 *	Return Value:
 *			0 on error, otherwise NEVER RETURNS.
 **********************************************************************/
int	graphicsLoop(inStream, outStream)
  FILE	*inStream, *outStream;
{
  XtAppContext	appContext ;

  appContext =  XtWidgetToApplicationContext(topLevel) ;

  if (inStream != NULL) {
    if (outStream == NULL)
      outStream = stdout ;
    if (enableCommandLineInput(inStream, outStream) == 0)
      return 0 ;
  }

  setSignalHandler() ;
  XtAppMainLoop(appContext) ;
  resetSignalHandler() ;

  return 0 ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		delayDisplayReset
 *	Description:	sets a flag saying to save the display resets
 *			till later
 *	Name:		catchupOnDisplayResets
 *	Description:	unsets a flag saying to save the display resets
 *			till later, then does any resets.
 *	Parameters:	NONE
 *	Return Value:
 *	  int	delayDisplayReset      - TRUE
 *	  int	catchupOnDisplayResets - FALSE
 *********************************************************************/
static	int	delayDisplayResetFlag = FALSE ;
/**********************************************************************/
int	delayDisplayResets() 
{
  delayDisplayResetFlag = TRUE ;
  return delayDisplayResetFlag ;
}
/********************************************************************/
int	catchupOnDisplayResets() 
{
  delayDisplayResetFlag = FALSE ;

  updateActivationDisplay() ;
  updateWeightDisplay() ;  
  updateMinimizeDisplay() ;
  updateVarDisplays() ;
  redrawGraphDisplays() ;
  
  return delayDisplayResetFlag ;
}
/********************************************************************/


/***********************************************************************
 *	Name:		resetDisplay
 *	Description:	sets flags saying that the network widgets
 *			must be rebuilt (should be called after major
 *			changes to the network or whenever the display
 *			params are changed), and if resets are not being
 *			delays
 *	Parameters:	
 *		int	which - a mask set saying which displays should
 *				be redrawn. May contain one or more of:
 *				ACTIVITY_DISPLAY
 *				EXAMPLE_DISPLAY
 *				CONNECTION_DISPLAY
 *				ALL_DISPLAYS
 *	Return Value:	1
 ***********************************************************************/
int	markToRebuildDisplay(which) 
  int	which ;
{
  
  if (topLevel == NULL)
    return 0 ;

  if (which & ACTIVITY_DISPLAY)
    rebuildActFlag     = True ;
  if (which & EXAMPLE_DISPLAY)
    rebuildExampleFlag = True ;
  if (which & CONNECTION_DISPLAY)
    rebuildWeightFlag  = True ;
  if (which & ACTIVITY_DISPLAY)
    rebuildTargetFlag  = True ;

  if (delayDisplayResetFlag == FALSE)
    catchupOnDisplayResets() ;

  return (int)1 ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		dispatchExposures
 *	Description:	flushes the output buffers and dispatches all
 *			exposure events in the input queue. This should
 *			be called intermittently by commands that take
 *			a VERY long time.
 *	Parameters:	NONE
 *	Return Value:	1
 ***********************************************************************/
int	dispatchExposures() {
  XEvent	event ;

  while (XCheckMaskEvent(XtDisplay(topLevel), 
			 SubstructureNotifyMask|SubstructureRedirectMask
			 |ResizeRedirectMask|ExposureMask, &event) != False) {
    XtDispatchEvent(&event) ;
  }
  return 1 ;
}
/**********************************************************************/
  

/*********************************************************************
 *	Name:		addUserDisplay
 *	Description:	allows a user to register a name and creation 
 *			procedure for a top level display
 *	Parameters:
 *	  char	*name - the base name for the display 
 *			(eg. "usr" -> "usrDisplay")
 *	  void	(*createProc)() - the creation procedure, will be passed
 *			the shell to create children of and to pop up.
 *	Return Value:
 *	  int	addUserDisplay - 0 on failure, 1 on success
 *********************************************************************/
typedef struct UserDisplay {
  char	*name ;
  void	(*createProc)() ;
} UserDisplay ;
/********************************************************************/
static UserDisplay	*userDisplayList = NULL ;
static int		numUserDisplays = 0 ;
/********************************************************************/
int	addUserDisplay(name, createProc)
  char	*name ;
  void	(*createProc)() ;
{
  ++numUserDisplays ;
  if (userDisplayList == NULL)
    userDisplayList = (UserDisplay *)XtMalloc(sizeof(UserDisplay)) ;
  else
    userDisplayList 
      = (UserDisplay *)XtRealloc((char *)userDisplayList,
				 numUserDisplays*sizeof(UserDisplay)) ;

  if (userDisplayList == NULL) {
    numUserDisplays = 0 ;
    return 0 ;
  }
  userDisplayList[numUserDisplays-1].name       = XtNewString(name) ;
  userDisplayList[numUserDisplays-1].createProc = createProc ;

  return 1 ;
}
/**********************************************************************/

/**********************************************************************
 *	Name:		createMainWindow
 *	Description:	creates a window with push buttons for popping
 *			up the display  windows for the simulator
 *	Parameters:
 *		Widget	parent - the parent widget to this widget
 *	Return Value:
 *		Widget	createMainWindow - the widget created, NULL on 
 *				error.
 **********************************************************************/
#define TOP_LEVEL	((int)0)
#define TRANSIENT	((int)1)
#define createTopLevel(w, s, f)	createDisplay(w, s, f, TOP_LEVEL)
#define createDialog(w, s, f)	createDisplay(w, s, f, TRANSIENT)
/**********************************************************************/
static Widget	createDisplay(parent, name, createFunction, type)
  Widget	parent ;
  char		*name ;
  void		(*createFunction)() ;
  int		type ;
{
  XtAppContext	appContext = XtWidgetToApplicationContext(topLevel) ;
  Widget	button, shell ;
  char		displayName[BUFSIZ] ;

  startArgs() ;
  button = createManagedWidget(name, commandWidgetClass, parent) ;

  sprintf(displayName, "%sDisplay", name) ;
  startArgs() ;
  if (createFunction != NULL)
    setArg(XtNcreatePopupChildProc, createFunction) ;

  if (type == TOP_LEVEL)
    shell = createPopupShell(displayName, topLevelShellWidgetClass, topLevel) ;
  else if (type == TRANSIENT)
    shell = createPopupShell(displayName, transientShellWidgetClass, topLevel);
  else
    return NULL ;

  XtAppAddWorkProc(appContext, createPopupChild, (XtPointer)shell) ;

  if (type == TRANSIENT)
    XtAddCallback(button, XtNcallback,  popupCenteredCB, (XtPointer)shell) ;
  else
    XtAddCallback(button, XtNcallback,  popupRaiseWidgetCB, (XtPointer)shell) ;

  XtSetKeyboardFocus(button, parent);
  return shell ;
}
/**********************************************************************/
static Widget	createMainWindow(parent)
  Widget	parent ;
{
  Widget	box, widget ;
  XtAppContext	appContext = XtWidgetToApplicationContext(topLevel) ;
  int		idx, userIdx ;

  startArgs() ;
  box = createManagedWidget("displayBox", boxWidgetClass, parent) ;

  for (idx = 0 ; idx < XtNumber(topLevelShell) ; ++idx)
    topLevelShell[idx] = NULL ;

  /* set the top level shell */
  idx = 0 ;
  topLevelShell[idx++] = topLevel ;

  /* all of the network displays */
  startArgs() ;
  widget = createManagedWidget("displays", labelWidgetClass, box) ;
  topLevelShell[idx++] = createTopLevel(box, "activation",  
					createActivationDisplay) ;
  topLevelShell[idx++] = createTopLevel(box, "weight",      
					createWeightDisplay) ;

  topLevelShell[idx++] = createTopLevel(box, "mini",      
					createMinimizeDisplay) ;

  /* User defined displays (added by addUserDisplay) */
  for (userIdx = 0 ; userIdx < numUserDisplays ; ++userIdx)
    topLevelShell[idx++] = createTopLevel(box, 
					  userDisplayList[userIdx].name, 
					  userDisplayList[userIdx].createProc);

  /* The graph displays */
#if 0
  startArgs() ;
  widget = createManagedWidget("graphs", labelWidgetClass, box) ;
#endif
  topLevelShell[idx++] = createDialog(box, "graph", createGraphQuery) ;

  /* The variable etc. lists */
#if 0
  startArgs() ;
  widget = createManagedWidget("lists", labelWidgetClass, box) ;
  topLevelShell[idx++] = createTopLevel(box, "netParam",  	
					createNetParamDisplay) ;
  topLevelShell[idx++] = createTopLevel(box, "miniParam", 
					createMiniParamDisplay) ;
#endif
  topLevelShell[idx++] = createTopLevel(box, "variable",  
					createVariableDisplay) ;

  /* The quit button */
  startArgs() ;
  widget = createManagedWidget("quit", commandWidgetClass, box) ;
  topLevelShell[idx] 
    = createPopupShell("quitDisplay", transientShellWidgetClass, topLevel) ;
  XtAppAddWorkProc(appContext, 
		   createPopupChild, (XtPointer)topLevelShell[idx]) ;
  XtAddCallback(widget, XtNcallback,  
		confirmQuitCB, (XtPointer)topLevelShell[idx]) ;
  XtSetKeyboardFocus(widget, parent);

  XtInstallAllAccelerators(box, box) ;

  return box ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		popupShell
 *	Description:	pops up one of the displays (in GrabNone mode)
 *	Parameters:
 *	  char	*name - the name of the shell (display)
 *	Return Value:
 *	  int	popupShell - 1 on success, 0 on failure.
 *********************************************************************/
int	popupShell(name, geometry)
  char	*name ;
  char	*geometry ;
{
  Widget	widget ;

  if (topLevel == NULL)
    return 0 ;

  widget = XtNameToWidget(topLevel, name) ;
  if (widget == NULL)
    return 0 ;

  if (geometry) {
    startArgs() ;
    setArg(XtNgeometry, geometry) ;
    setValues(widget) ;
  }

  XtPopup(widget, XtGrabNone);
  return 1 ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		createPopupChild
 *	Description:	work procedure to call the createPopup proc for
 *			a popup shell
 *	Parameters:
 *	  XtPointer	clientData - the popup shell
 *	Return Value:
 *	  static Boolean	CreatePopupChild - TRUE
 *********************************************************************/
static Boolean	createPopupChild(clientData)
  XtPointer	clientData ;
{
  Widget	shell = (Widget)clientData ;
  void		(*create)() ;

  startArgs() ;
  setArg(XtNcreatePopupChildProc, &create) ;
  getValues(shell) ;

  if (create != NULL)
    (*create)(shell) ;
  return TRUE ;
}
/********************************************************************/


/**********************************************************************
 *	Name:		popupRaiseWidgetCB
 *	Description:	pops up a popup shell which is passed in as
 *			the client data
 *	Parameters:
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the shell to pop up
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	popupRaiseWidgetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  ShellWidget	popup = (ShellWidget)clientData ;

  if (popup->shell.popped_up == TRUE) {
    XBell(XtDisplay(topLevel), (int)0) ;
    XRaiseWindow(XtDisplay(topLevel), XtWindow(popup)) ;
  } else {
    XtPopup((Widget)popup, XtGrabNone) ;
  }
}
/**********************************************************************/


/**********************************************************************
 *	Name:		popupWidgetCB
 *	Description:	pops up a popup shell which is passed in as
 *			the client data
 *	Parameters:
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the shell to pop up
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	popupWidgetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	popup = (Widget)clientData ;
  XtPopup(popup, XtGrabNone) ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		popdownWidgetCB
 *	Description:	pops down a popup shell which is passed in as
 *			the client data
 *	Parameters:
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the shell to pop down
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	popdownWidgetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	popup = (Widget)clientData ;
  XtPopdown(popup) ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		unselectCB
 *	Description:	unselects a toggle button
 *	Parameters:
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the toggle button to unselect
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	unselectCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	toggle = (Widget)clientData ;
  startArgs() ;
  setArg(XtNstate, False) ;
  setValues(toggle) ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		confirmQuitCB
 *	Description:	quits the application BUT CONFIRMS FIRST
 *	Parameters:	
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the clientData (unused)
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 ***********************************************************************/
static void	confirmQuitCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  static Widget	shell ;

  if (shell == NULL) {
    Widget	dialog, button ;
    startArgs() ;
    shell = createPopupShell("quitDialog", transientShellWidgetClass, topLevel) ;
    
    startArgs() ;
    dialog = createManagedWidget("quitDialog", dialogWidgetClass, shell) ;

    startArgs() ;
    button = createManagedWidget("ok", commandWidgetClass, dialog) ;
    XtAddCallback(button, XtNcallback, popdownWidgetCB, (XtPointer)shell) ;
    XtAddCallback(button, XtNcallback, quitCB	      , (XtPointer)topLevel) ;

    button = createManagedWidget("cancel", commandWidgetClass, dialog) ;
    XtAddCallback(button, XtNcallback, popdownWidgetCB, (XtPointer)shell) ;
  }
  popupCentered(shell, XtGrabExclusive) ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		quitCB
 *	Description:	quits the entire application
 *	Parameters:
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the client data (unused)
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	quitCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  XtDestroyWidget(topLevel) ;
  fprintf(stderr, "Bye!\n") ;
  fflush(stderr) ;
  exit(0) ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		enableCommandLineInput
 *	Description:	adds command line input to the graphics events
 *			and adds the callback to handle the event.
 *	Parameters:
 *		FILE	*inStream - the stream for input
 *		FILE	*outStream - the stream for output
 *	Return Value:
 *		int	enableCommandLineInput - +1 on success, 0 on 
 *			failure.
 **********************************************************************/
static int	enableCommandLineInput(inStream, outStream)
  FILE	*inStream, *outStream;
{
  int			fileDescriptor ;
  XtAppContext		appContext ;

  if (topLevel == NULL)
    return 0 ;

  InStream  = inStream ;
  OutStream = outStream ;
 
  fileDescriptor = fileno(InStream) ;
  if (fileDescriptor < 0)
    return 0 ;

  appContext =  XtWidgetToApplicationContext(topLevel) ;

  commandId = XtAppAddInput(appContext, fileDescriptor, 
			    (XtPointer)XtInputReadMask, 
			    commandLineCB, (XtPointer)NULL) ;
  if (commandId == 0) 
    return 0 ;

  return 1 ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		commandLineCB
 *	Description:	the callback for when a command is entered
 *			to the commandLine. This calls the routines
 *			from the itf module which actually reads the
 *			command. It must be added to the event list
 *			by XtAppAddInput.
 *	Parameters:
 *		XtPointer		clientData - unused
 *		int		sourceFile - the file descriptor of
 *				the input file
 *		XtInputId	id - the id returned by XtAppAddInput
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	commandLineCB(clientData, sourceFile, id)
  XtPointer	clientData ;
  int		*sourceFile ;
  XtInputId	*id ;
{
  blockDeviceEvents() ;
  resetSignalHandler() ;
  if (IDoSingleCommand(InStream, OutStream) == 0)
    exit (0) ;
  setSignalHandler() ;
  unblockDeviceEvents() ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		blockDeviceEvents / unblockDeviceEvents
 *	Description:	blocks device events to the application and 
 *			shows a busy cursor, then unblocks it (if it 
 *			is blocked)
 *	Parameters:	NONE
 *	Return Value:	NONE
 ***********************************************************************/
static Window	busyWindow[MAX_SHELLS] ;
static Boolean	deviceEventsBlocked = False ;

void	blockDeviceEvents() {
  int	idx ;

  for (idx = 0 ; idx < XtNumber(topLevelShell) ; ++idx) {
    if (topLevelShell[idx] != NULL 
	&& XtIsRealized(topLevelShell[idx]) == True) {
      if (busyWindow[idx] == NULL)
	busyWindow[idx] = createBusyWindow(topLevelShell[idx]) ;

      XMapWindow(XtDisplay(topLevelShell[idx]),   busyWindow[idx]) ;
      XRaiseWindow(XtDisplay(topLevelShell[idx]), busyWindow[idx]) ;
      XFlush(XtDisplay(topLevelShell[idx])) ;
    }
  }
  deviceEventsBlocked = True ;
}
/**********************************************************************/
void	unblockDeviceEvents() {
  int	idx ;

  if (deviceEventsBlocked == False)
    return ;

  for (idx = 0 ; idx < XtNumber(topLevelShell) ; ++idx) {
    if (topLevelShell[idx] != NULL && XtIsRealized(topLevelShell[idx]) == True
	&& busyWindow[idx] != NULL)
      XUnmapWindow(XtDisplay(topLevelShell[idx]),   busyWindow[idx]) ;
  }	
  deviceEventsBlocked = False ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		createBusyWindow
 *	Description:	creates a window as large as the screen which
 *			should stop all deviceEvents to the given shell
 *			while it is mapped and raised
 *	Parameters:	
 *		Widget	shell - the shell to be busy for
 *	Return Value:	
 *		Window	createBusyWindow - the newly created window
 ***********************************************************************/
static Window	createBusyWindow(shell)
  Widget	shell ;
{
  unsigned long 	valuemask ;
  XSetWindowAttributes	attributes ;
  Window		window ;

  /* Ignore device events while the busy cursor is displayed. */
  valuemask = CWDontPropagate | CWCursor ;
  attributes.do_not_propagate_mask =  (KeyPressMask | KeyReleaseMask |
				       ButtonPressMask | ButtonReleaseMask |
				       PointerMotionMask) ;
  attributes.cursor = XCreateFontCursor(XtDisplay(shell), XC_watch) ;

  /* The window will be as big as the display screen, and clipped by
     its own parent window, so we never have to worry about resizing */
  window = XCreateWindow(XtDisplay(shell), XtWindow(shell), 0, 0,
			 WidthOfScreen(XtScreen(shell)), 
			 HeightOfScreen(XtScreen(shell)),
			 (unsigned int) 0, CopyFromParent, InputOnly,
			 CopyFromParent, valuemask, &attributes) ;
  return window ;
}
/**********************************************************************/
