
/**********************************************************************
 * $Id: displayUtils.c,v 1.1 92/04/20 16:32:32 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <varargs.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Form.h>

#include <xerion/display.h>

/***********************************************************************
 *	Name:		titlePrintf
 *	Description:	same as printf, but writes to the XtNtitle
 *			resource of a shell widget. It doesn't update
 *			the field if it the string is unchanged.
 *	Parameters:	va_alist - this uses varargs, so be careful
 *				that the format string is correct.
 *	Return Value:	whatever vsprintf returns.
 ***********************************************************************/
int	titlePrintf(va_alist)
  va_dcl 
{
  va_list 	args ;
  Widget	widget ;
  int		returnValue ;
  char		*format, *oldString, newString[BUFSIZ] ;

  va_start(args) ;
  widget = va_arg(args, Widget) ;
  format = va_arg(args, char *) ;

  XtVaGetValues(widget, XtNtitle, &oldString, NULL) ;

  returnValue = vsprintf(newString, format, args) ;

  if (strcmp(newString, oldString))
    XtVaSetValues(widget, XtNtitle, newString, NULL) ;

  va_end(args) ;
  return returnValue ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		labelPrintf
 *	Description:	same as printf, but writes to the XtNlabel
 *			resource of a widget. It doesn't update
 *			the field if it the string is unchanged.
 *	Parameters:	va_alist - this uses varargs, so be careful
 *				that the format string is correct.
 *	Return Value:	whatever vsprintf returns.
 ***********************************************************************/
int	labelPrintf(va_alist)
  va_dcl 
{
  va_list 	args ;
  Widget	widget ;
  int		returnValue ;
  char		*format, *oldString, newString[BUFSIZ] ;

  va_start(args) ;
  widget = va_arg(args, Widget) ;
  format = va_arg(args, char *) ;

  XtVaGetValues(widget, XtNlabel, &oldString, NULL) ;

  returnValue = vsprintf(newString, format, args) ;

  if (strcmp(newString, oldString))
    XtVaSetValues(widget, XtNlabel, newString, NULL) ;

  va_end(args) ;
  return returnValue ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		textPrintf
 *	Description:	same as printf, but writes to the XtNstring
 *			resource of a text widget. It doesn't update
 *			the field if it the string is unchanged.
 *	Parameters:	va_alist - this uses varargs, so be careful
 *				that the format string is correct.
 *	Return Value:	whatever vsprintf returns.
 ***********************************************************************/
int	textPrintf(va_alist)
  va_dcl 
{
  va_list 	args ;
  Widget	widget ;
  int		returnValue ;
  char		*format, *oldString, newString[BUFSIZ] ;

  va_start(args) ;
  widget = va_arg(args, Widget) ;
  format = va_arg(args, char *) ;

  XtVaGetValues(widget, XtNstring, &oldString, NULL) ;

  returnValue = vsprintf(newString, format, args) ;

  if (strcmp(newString, oldString))
    XtVaSetValues(widget, XtNstring, newString, NULL) ;

  va_end(args) ;
  return returnValue ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		getTitle
 *	Description:	returns a copy of the title resource from a shell
 *			widget
 *	Parameters:
 *	  Widget	widget - the widget to get the title from
 *	Return Value:
 *	  char		*getTitle - a copy of the string
 *********************************************************************/
char		*getTitle(widget)
  Widget	widget ;
{
  char		*ptr ;

  XtVaGetValues(widget, XtNtitle,  &ptr, NULL) ;
  if (ptr != NULL)
    return XtNewString(ptr) ;
  else
    return NULL ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		getLabel
 *	Description:	returns a copy of the label in a widget
 *	Parameters:
 *	  Widget	widget - the widget to get the label from
 *	Return Value:
 *	  char		*getLabel - a copy of the string
 *********************************************************************/
char		*getLabel(widget)
  Widget	widget ;
{
  char		*ptr ;

  XtVaGetValues(widget, XtNlabel,  &ptr, NULL) ;
  if (ptr != NULL)
    return XtNewString(ptr) ;
  else
    return NULL ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		getText
 *	Description:	returns a copy of the string in a text widget
 *	Parameters:
 *	  Widget	widget - the widget to get the text from
 *	Return Value:
 *	  char		*getText - the string, *not* a copy
 *********************************************************************/
char		*getText(widget)
  Widget	widget ;
{
  char		*ptr ;

  XtVaGetValues(widget, XtNstring,  &ptr, NULL) ;
  return ptr ;
}
/********************************************************************/


static Widget	tableParent ;

/*********************************************************************
 *	Name:		startTable
 *	Description:	initializes a table of options
 *	Parameters:
 *	  Widget	form - the form to be the parent
 *	Return Value:
 *	  Widget	startTable - NULL
 *********************************************************************/
Widget		startTable(form)
  Widget	form ;
{
  tableParent = form ;
  return NULL ;
}
/********************************************************************/

/*********************************************************************
 *	Name:		addTableItem
 *	Description:	creates a labeled edit widget with 
 *	Parameters:
 *	  String	name - the name for the edit widget (Label will
 *				be called name"Label"
 *	  Widget	*table - the value returned from startTable
 *	  Boolean	toggleToo - if TRUE add a toggle to the right
 *				called name"Toggle"
 *	Return Value:
 *	  static Widget	addTableItem - the edit widget
 *********************************************************************/
Widget		addTableItem(name, table, toggleToo)
  String	name ;
  Widget	*table ;
  Boolean	toggleToo ;
{
  char		labelName[BUFSIZ] ;
  Widget	label, edit, toggle ;

  sprintf(labelName, "%sLabel", name) ;

  startArgs() ;
  if (*table) {
    setArg(XtNfromVert, *table) ;
  }
  label = createManagedWidget(labelName, labelWidgetClass, tableParent) ;

  startArgs() ;
  if (*table) {
    setArg(XtNfromVert,		*table) ;
  }
  setArg(XtNfromHoriz,	label) ;
  setArg(XtNtype,	XawAsciiString) ;
  setArg(XtNeditType,	XawtextEdit) ;
  edit = createManagedWidget(name, asciiTextWidgetClass, tableParent) ;
  if (toggleToo) {
    sprintf(labelName, "%sToggle", name) ;
    startArgs() ;
    if (*table) {
      setArg(XtNfromVert,	*table) ;
    }
    setArg(XtNfromHoriz,	edit) ;
    toggle = createManagedWidget(labelName, toggleWidgetClass, tableParent) ;
  }
  *table = edit ;
  return edit ;
}
/********************************************************************/
