
/**********************************************************************
 * $Id: graphDisplay.c,v 1.2 92/04/20 16:07:56 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <math.h>
#include <varargs.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xutil.h>

#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Text.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>

#include <xerion/Graph.h>
#include <xerion/MenuPane.h>
#include <xerion/display.h>

#include "display2Itf.h"
#include "display2Sim.h"
#include "displayUtils.h"
#include "traceDisplay.h"

typedef struct {
  Widget	shell ;
  Widget	fixedMin ;
  Widget	fixedMax ;
  Widget	maxValue ;
  Widget	minValue ;
  Widget	maxFormat ;
  Widget	minFormat ;
  Widget	numColumns ;
  Widget	jumpScroll ;
  Widget	graph ;
} *MiscGraphDialog ;

typedef struct GraphDisplay {
  Widget	popupShell ;	/* the shell containing the display	 */
  char		*objectName ;	/* the name of the object being plotted	 */
  String	titleFormat ;	/* the format string to print the title	 */
  Widget	graphWidget ;	/* the graph widget			 */
  Widget	graphMenu ;	/* the graphOptions menuButton		 */
  Widget	plotButton ;	/* the button for plotting the object	 */
  MiscGraphDialog miscDialog ;	/* holds all the stuff for the misc menu */
} GraphDisplayObj, *GraphDisplay ;


/* private functions */
static void	createGraphCB   ARGS((Widget, XtPointer, XtPointer)) ;
static void	changeGraphCB   ARGS((Widget, XtPointer, XtPointer)) ;
static void	destroyGraphCB  ARGS((Widget, XtPointer, XtPointer)) ;
static void	popdownWidgetCB ARGS((Widget, XtPointer, XtPointer)) ;
static void	plotValueCB     ARGS((Widget, XtPointer, XtPointer)) ;
static void	rescaleGraphCB  ARGS((Widget, XtPointer, XtPointer)) ;
static void	shiftGraphCB    ARGS((Widget, XtPointer, XtPointer)) ;
static void	clearGraphCB    ARGS((Widget, XtPointer, XtPointer)) ;
static void 	setSensitiveCB  ARGS((Widget, XtPointer, XtPointer)) ;
static void 	miscellaneousCB ARGS((Widget, XtPointer, XtPointer)) ;

#define MAXGRAPHS	64
static GraphDisplayObj	graphDisplay[MAXGRAPHS] ;

/*********************************************************************
 *	Name:		createGraphQuery
 *	Description:	creates the graph query popup (asks for variable
 *			to graph, and adds, buttons to make the graph)
 *	Parameters:
 *	  Widget	popup - the popup shell that is the parent
 *	Return Value:
 *	  export void	createGraphQuery - NONE
 *********************************************************************/
export void	createGraphQuery(popup)
  Widget	popup ;
{
  Widget	dialog, value, ok, cancel ;
  Position	x, y ;
  
  startArgs() ;
  dialog = createManagedWidget("graphDialog", dialogWidgetClass, popup) ;

  startArgs() ;
  setArg(XtNvalue, "currentNet.error") ;
  setValues(dialog) ;

  startArgs() ;
  ok     = createManagedWidget("ok",     commandWidgetClass, dialog) ;
  cancel = createManagedWidget("cancel", commandWidgetClass, dialog) ;

  XtAddCallback(ok,     XtNcallback, createGraphCB,   (XtPointer)dialog) ;
  XtAddCallback(ok,     XtNcallback, popdownWidgetCB, (XtPointer)popup) ;
  XtAddCallback(cancel, XtNcallback, popdownWidgetCB, (XtPointer)popup) ;

  value = XtNameToWidget(dialog, "value") ;
  startArgs() ;
  setArg(XtNwrap, XawtextWrapLine) ;
  setValues(value) ;
  XtSetKeyboardFocus(dialog, value) ;
  XtInstallAllAccelerators(value, dialog) ;
  
  /* center the popup in the screen */
  x = WidthOfScreen(XtScreen(topLevel))/2  - 150 ;
  y = HeightOfScreen(XtScreen(topLevel))/2 - 100 ;
    
  startArgs() ;
  setArg(XtNx, x) ;
  setArg(XtNy, y) ;
  setValues(popup) ;

  startArgs() ;
  setArg(XtNcreatePopupChildProc, NULL) ;
  setValues(popup) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		popupGraph
 *	Description:	pops up a graph (already created) given the
 *			name of the object being graphed
 *	Parameters:
 *	  char	*objectName - the name of the object being graphed
 *	Return Value:
 *	  int	popupGraph - 0 on failure, 1 on success.
 *********************************************************************/
int	popupGraph(objectName, geometry)
  char	*objectName ;
  char	*geometry ;
{
  int	idx ;

  for (idx = 0 ; 
       idx < MAXGRAPHS 
       && strcmp(objectName, graphDisplay[idx].objectName) ; ++idx)
    ;
  
  if (idx == MAXGRAPHS)
    return 0 ;

  if (geometry) {
    startArgs() ;
    setArg(XtNgeometry, geometry) ;
    setValues(graphDisplay[idx].popupShell) ;
  }

  XtPopup(graphDisplay[idx].popupShell, XtGrabNone) ;
  return 1 ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		createGraph
 *	Description:	creates a graph display and returns a pointer
 *			to the popup shell surrounding it.
 *	Parameters:
 *	  char	*objectName - the name of the object to graph
 *	Return Value:
 *	  Widget 	createGraph - the popup shell around the graph
 *********************************************************************/
Widget 	createGraph(objectName)
  char	*objectName ;
{
  Widget	menuPane, menuButton, menu, button, line ;
  Widget	popup, form, graph, updateShell, changeShell ;
  char		command[128], widgetName[128] ;
  String	ptr ;
  int		idx ;

  if (topLevel == NULL)
    return NULL ;

  for (idx = 0 ; 
       idx < MAXGRAPHS && graphDisplay[idx].graphWidget != NULL ; ++idx)
    ;
  
  if (idx == MAXGRAPHS)
    return NULL ;

  graphDisplay[idx].objectName = XtNewString(objectName) ;

  /********************************************
   * the shell and top level form
   */
  startArgs() ;
  popup = createPopupShell("graphShell", topLevelShellWidgetClass, topLevel) ;
  graphDisplay[idx].popupShell = popup ;
  graphDisplay[idx].titleFormat = getTitle(popup) ;

  titlePrintf(graphDisplay[idx].popupShell, 
	      graphDisplay[idx].titleFormat, graphDisplay[idx].objectName) ;

  startArgs() ;
  form = createManagedWidget("graphForm", formWidgetClass, popup) ;

  /********************************************
   * Display Updates 
   */
  sprintf(command, "updateGraphDisplay %d", idx) ;
  startArgs() ;
  updateShell = createTracePopup(popup, command) ;

  /********************************************
   * Pulldown menu pane 
   */
  startArgs() ;
  menuPane = createManagedWidget("menuPane", menuPaneWidgetClass, form) ;

  /**********************
   * Options menu 
   */
  startArgs() ;
  setArg(XtNmenuName, "optionsMenu") ;
  menuButton = createManagedWidget("options", menuButtonWidgetClass, menuPane);

  startArgs() ;
  menu = createPopupShell("optionsMenu", simpleMenuWidgetClass, menuButton) ;
  
  startArgs() ;
  button = createManagedWidget("updates", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, popupCenteredCB, (XtPointer)updateShell) ;

  startArgs() ;
  line = createManagedWidget("line", smeLineObjectClass, menu) ;

  startArgs() ;
  button = createManagedWidget("close", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, popdownWidgetCB, (XtPointer)popup) ;

  /**********************
   * Graph menu 
   */
  startArgs() ;
  setArg(XtNmenuName, "graphMenu") ;
  menuButton = createManagedWidget("graph", menuButtonWidgetClass, menuPane);
  graphDisplay[idx].graphMenu = menuButton ;

  startArgs() ;
  menu = createPopupShell("graphMenu", simpleMenuWidgetClass, menuButton) ;
  
  startArgs() ;
  button = createManagedWidget("rescale", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, rescaleGraphCB, 
		(XtPointer)(&graphDisplay[idx])) ;

  startArgs() ;
  button = createManagedWidget("shift", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, shiftGraphCB,
		(XtPointer)&graphDisplay[idx]) ;

  startArgs() ;
  button = createManagedWidget("reset", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, clearGraphCB,
		(XtPointer)&graphDisplay[idx]) ;
  XtAddCallback(button, XtNcallback, setSensitiveCB,
		(XtPointer)&graphDisplay[idx]) ;

  startArgs() ;
  button = createManagedWidget("misc", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, miscellaneousCB,
		(XtPointer)&graphDisplay[idx]) ;

  /**********************
   * Object menu 
   */
  startArgs() ;
  setArg(XtNmenuName, "objectMenu") ;
  menuButton = createManagedWidget("object", menuButtonWidgetClass, menuPane);

  startArgs() ;
  menu = createPopupShell("objectMenu", simpleMenuWidgetClass, menuButton) ;
  
  startArgs() ;
  button = createManagedWidget("plotValue", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, plotValueCB,
		(XtPointer)&graphDisplay[idx]) ;
  graphDisplay[idx].plotButton = menuButton ;

  startArgs() ;
  button = createManagedWidget("changeObject", smeBSBObjectClass, menu) ;
  changeShell = createPopupShell("changeShell",
				 transientShellWidgetClass, popup) ;
  XtAddCallback(button, XtNcallback, popupCenteredCB, (XtPointer)changeShell) ;

  createGraphQuery(changeShell) ;
  button = XtNameToWidget(changeShell, "graphDialog.ok") ;
  XtRemoveAllCallbacks(button, XtNcallback) ;
  XtAddCallback(button, XtNcallback, popdownWidgetCB, (XtPointer)changeShell) ;
  XtAddCallback(button, XtNcallback, changeGraphCB,
		(XtPointer)&graphDisplay[idx]) ;
  XtAddCallback(button, XtNcallback, setSensitiveCB,
		(XtPointer)&graphDisplay[idx]) ;

  /********************************************
   * The graph itself
   */

  /**********************
   * name the graph the object name with '.' replaced by '_'
   */
  strncpy(widgetName, graphDisplay[idx].objectName, 128) ;
  for (ptr = widgetName ; *ptr && ptr - widgetName < 128 ; ++ptr) {
    if (*ptr == '.')
      *ptr = '_' ;
  }

  /**********************
   * make the graph
   */
  startArgs() ;
  setArg(XtNlabel, "") ;
  graph = createManagedWidget(widgetName, graphWidgetClass, form) ;
  graphDisplay[idx].graphWidget = graph ;

  /********************************************
   * att callbacks to the display itself
   */
  XtAddCallback(popup, XtNpopdownCallback, destroyGraphCB,
		(XtPointer)&graphDisplay[idx]) ;

  setSensitiveCB(popup, (XtPointer)&graphDisplay[idx], (XtPointer)NULL) ;
  XtInstallAllAccelerators(graph, form) ;

  return graphDisplay[idx].popupShell ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		createGraphCB
 *	Description:	actually builds a graph display, given the name
 *			of the object to graph
 *	Parameters:
 *	  Widget	widget - the widget the callback is attached to
 *	  XtPointer	clientData - the dialog widget to get the object
 *				name from
 *	  XtPointer	callData - UNUSED
 *	Return Value:
 *	  static void	createGraphCB - NONE
 *********************************************************************/
static void	createGraphCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	popup, dialog = (Widget)clientData ;
  String	objectName ;

  objectName = XawDialogGetValueString(dialog) ;
  if (objectName == NULL || objectName[0] == '\0')
    return ;
  
  popup = createGraph(objectName) ;

  XtPopup(popup, XtGrabNone) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		changeGraphCB
 *	Description:	changes the object being plotted int the graph
 *	Parameters:
 *	  Widget	widget     - a button in a dialog widget
 *	  XtPointer	clientData - the graphDisplay
 *	  XtPointer	callData   - UNUSED
 *	Return Value:
 *	  static void	changeGraphCB - NONE
 *********************************************************************/
static void	changeGraphCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  GraphDisplay	display = (GraphDisplay)clientData ;
  String	objectName ;
  Widget	parent = XtParent(widget) ;

  objectName = XawDialogGetValueString(parent) ;
  if (objectName == NULL || objectName[0] == '\0')
    return ;

  if (display->objectName)
    XtFree(display->objectName) ;
  display->objectName = XtNewString(objectName) ;

  titlePrintf(display->popupShell, display->titleFormat, display->objectName) ;

  graphReset(display->graphWidget) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		setSensitiveCB
 *	Description:	sets the sensitive field to true or false for
 *			all of the widgets that care.
 *	Parameters:
 *	  Widget	widget -  the widget the callback is attached to
 *	  XtPointer	clientData - a pointer to the graphDisplay object
 *	  XtPointer	callData   - UNUSED
 *	Return Value:
 *	  static void	setSensitiveCB - NONE
 *********************************************************************/
static void	setSensitiveCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  GraphDisplay	display = (GraphDisplay)clientData ;
  char		*valueString ;

  valueString = getSimpleObjectValue(display->objectName, "%.8g") ;

  startArgs() ;
  if (valueString == NULL || strcmp(valueString, "") == 0) {
    setArg(XtNsensitive, FALSE) ;
  } else {
    setArg(XtNsensitive, TRUE) ;
  }
  setValues(display->graphMenu) ;
  setValues(display->plotButton) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		destroyGraphCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData - 
 *	Return Value:
 *	  static void	destroyGraphCB - 
 *********************************************************************/
static void	destroyGraphCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	graph = ((GraphDisplay)clientData)->graphWidget ;
  int		idx ;

  for (idx = 0 ;
       idx < MAXGRAPHS && graphDisplay[idx].graphWidget != graph ; ++idx)
    ;
  
  if (idx == MAXGRAPHS)
    return ;

  XtDestroyWidget(graphDisplay[idx].popupShell) ;
  XtFree((char *)graphDisplay[idx].objectName) ;
  XtFree((char *)graphDisplay[idx].miscDialog) ;
  memset(&graphDisplay[idx], 0, sizeof(graphDisplay[idx])) ;
}
/********************************************************************/


/**********************************************************************
 *	Name:		popdownWidgetCB
 *	Description:	pops down a popup shell which is 
 *			passed in as the client data
 *	Parameters:
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the shell to pop down
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	popdownWidgetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  XtPopdown((Widget)clientData) ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		updateGraphDisplay
 *	Description:	this little hack allows outside users to
 *			update the graph display without clicking
 *			or popping anything
 *	Parameters:	
 *		int	idx - the index of the display to update
 *	Return Value:	NONE
 ***********************************************************************/
export void	updateGraphDisplay(idx)
  int		idx ;
{
  if (topLevel == NULL || idx >= MAXGRAPHS)
    return ;

  if (graphDisplay[idx].graphWidget == NULL)
    return ;

  setSensitiveCB(topLevel, (XtPointer)&graphDisplay[idx], (XtPointer)NULL) ;
  plotValueCB(topLevel,    (XtPointer)&graphDisplay[idx], (XtPointer)NULL) ;

  dispatchExposures() ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		redrawGraphDisplays
 *	Description:	this little hack allows outside users to
 *			update the graph display without clicking
 *			or popping anything
 *	Parameters:	NONE
 *	Return Value:	NONE
 ***********************************************************************/
export void	redrawGraphDisplays()
{
  int		idx ;

  if (topLevel == NULL)
    return ;

  for (idx = 0 ; idx < MAXGRAPHS ; ++idx) {
    if (graphDisplay[idx].graphWidget != NULL)
      setSensitiveCB(topLevel, (XtPointer)&graphDisplay[idx], (XtPointer)NULL) ;
  }
  dispatchExposures() ;
}
/**********************************************************************/



/***********************************************************************
 *	Name:		clearGraphCB
 *	Description:	clears a graph (passed in as client data)
 *	Parameters:	
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the graph Widget
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:	NONE
 ***********************************************************************/
static void	clearGraphCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  graphReset(((GraphDisplay)clientData)->graphWidget) ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		rescaleGraphCB
 *	Description:	rescales a graph (passed in as client data)
 *	Parameters:	
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the graph Widget
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:	NONE
 ***********************************************************************/
static void	rescaleGraphCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  graphRescale(((GraphDisplay)clientData)->graphWidget) ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		shiftGraphCB
 *	Description:	shifts a graph (passed in as client data)
 *	Parameters:	
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the graph Widget
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:	NONE
 ***********************************************************************/
static void	shiftGraphCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  graphShift(((GraphDisplay)clientData)->graphWidget) ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		plotValueCB
 *	Description:	adds values to the graph widget using 
 *			updateGraphDisplay from above
 *	Parameters:
 *		Widget	widget 	   - the widget the callback is attached to
 *		XtPointer	clientData - the graph widget
 *		XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	plotValueCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  GraphDisplay	display = (GraphDisplay)clientData ;
  char		*valueString ;
  float		value ;

  valueString = getSimpleObjectValue(display->objectName, "%.8g") ;
  if (valueString == NULL)
    return ;

  value = atof(valueString) ;
  graphAddValue(display->graphWidget, value) ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		setMiscellaneousCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData -
 *	Return Value:
 *	  static void	setMiscellaneousCB - 
 *********************************************************************/
static void	setMiscellaneousCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  MiscGraphDialog	dialog = (MiscGraphDialog)clientData ;
  Boolean		fixedMin, fixedMax ;
  float			minValue, maxValue ;
  int			numColumns, jumpScroll ;
  String		maxFormat, minFormat, string ;

  minValue   = 0.0 ;
  maxValue   = 0.0 ;
  numColumns = 0 ;
  jumpScroll = 0 ;
  minFormat  = NULL ;
  maxFormat  = NULL ;

  string = getText(dialog->minValue) ;
  if (string != NULL)
    minValue = atof(string) ;

  string = getText(dialog->maxValue) ;
  if (string != NULL)
    maxValue = atof(string) ;

  string = getText(dialog->maxFormat) ;
  if (string != NULL)
    maxFormat = string ;

  string = getText(dialog->minFormat) ;
  if (string != NULL)
    minFormat = string ;

  string = getText(dialog->numColumns) ;
  if (string != NULL)
    numColumns = atof(string) ;

  string = getText(dialog->jumpScroll) ;
  if (string != NULL)
    jumpScroll = atof(string) ;
    
  startArgs() ;
  setArg(XtNstate, &fixedMin) ;
  getValues(dialog->fixedMin) ;

  startArgs() ;
  setArg(XtNstate, &fixedMax) ;
  getValues(dialog->fixedMax) ;

  startArgs() ;
  setArg(XtNminValue,	&minValue) ;
  setArg(XtNmaxValue,	&maxValue) ;
  setArg(XtNmaxFormat,	maxFormat) ;
  setArg(XtNminFormat,	minFormat) ;
  setArg(XtNjumpScroll, jumpScroll) ;
  setArg(XtNnumColumns, numColumns) ;
  setArg(XtNfixedMin,	fixedMin) ;
  setArg(XtNfixedMax,	fixedMax) ;
  setValues(dialog->graph) ;
}

/*********************************************************************
 *	Name:		miscellaneousCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData - 
 *	Return Value:
 *	  static void	miscellaneousCB - 
 *********************************************************************/
static void	miscellaneousCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  GraphDisplay	display = (GraphDisplay)clientData ;
  float		*min, *max ;
  Boolean	fixedMin, fixedMax ;
  int		numColumns, jumpScroll ;
  String	minFormat, maxFormat ;
  MiscGraphDialog	dialog ;
  extern Widget	startTable() ;
  extern Widget	addTableItem() ;

  if (display->miscDialog == NULL) {
    Widget	form, button, table ;
    Position	x, y ;
    
    dialog = (MiscGraphDialog)XtMalloc(sizeof(*dialog)) ;
    display->miscDialog = dialog ;

    startArgs() ;
    dialog->shell = createPopupShell("miscShell", 
				     transientShellWidgetClass, 
				     display->popupShell) ;
    startArgs() ;
    form = createManagedWidget("form", formWidgetClass, dialog->shell) ;

    table = startTable(form) ;
    dialog->minValue   = addTableItem("minValue",   &table, TRUE) ;
    dialog->maxValue   = addTableItem("maxValue",   &table, TRUE) ;
    dialog->maxFormat  = addTableItem("maxFormat",  &table, FALSE) ;
    dialog->minFormat  = addTableItem("minFormat",  &table, FALSE) ;
    dialog->numColumns = addTableItem("numColumns", &table, FALSE) ;
    dialog->jumpScroll = addTableItem("jumpScroll", &table, FALSE) ;

    dialog->fixedMax   = XtNameToWidget(form, "maxValueToggle") ;
    dialog->fixedMin   = XtNameToWidget(form, "minValueToggle") ;

    startArgs() ;
    setArg(XtNfromVert,	dialog->jumpScroll) ;
    button = createManagedWidget("ok", commandWidgetClass, form) ;
    XtAddCallback(button, XtNcallback, setMiscellaneousCB, (XtPointer)dialog) ;
    XtAddCallback(button, XtNcallback, popdownWidgetCB, 
		  (XtPointer)dialog->shell) ;

    startArgs() ;
    setArg(XtNfromVert,	dialog->jumpScroll) ;
    setArg(XtNfromHoriz,button) ;
    button = createManagedWidget("cancel", commandWidgetClass, form) ;
    XtAddCallback(button, XtNcallback, popdownWidgetCB, 
		  (XtPointer)dialog->shell) ;

    XtInstallAllAccelerators(form, form) ;

    /* center the popup in the screen */
    x = WidthOfScreen(XtScreen(topLevel))/2  - 100 ;
    y = HeightOfScreen(XtScreen(topLevel))/2 - 100 ;
    
    startArgs() ;
    setArg(XtNx, x) ;
    setArg(XtNy, y) ;
    setValues(dialog->shell) ;
  }
  dialog = display->miscDialog ;

  dialog->graph = display->graphWidget ;

  startArgs() ;
  setArg(XtNfixedMin,	&fixedMin) ;
  setArg(XtNfixedMax,	&fixedMax) ;
  setArg(XtNminValue,	&min) ;
  setArg(XtNmaxValue,	&max) ;
  setArg(XtNmaxFormat,	&maxFormat) ;
  setArg(XtNminFormat,	&minFormat) ;
  setArg(XtNnumColumns, &numColumns) ;
  setArg(XtNjumpScroll, &jumpScroll) ;
  getValues(dialog->graph) ;

  textPrintf(dialog->minValue,   "%g", *min) ;
  textPrintf(dialog->maxValue,   "%g", *max) ;
  textPrintf(dialog->maxFormat,	 "%s", maxFormat) ;
  textPrintf(dialog->minFormat,	 "%s", minFormat) ;
  textPrintf(dialog->numColumns, "%d", numColumns) ;
  textPrintf(dialog->jumpScroll, "%d", jumpScroll) ;

  startArgs() ;
  setArg(XtNstate, fixedMin) ;
  setValues(dialog->fixedMin) ;

  startArgs() ;
  setArg(XtNstate, fixedMax) ;
  setValues(dialog->fixedMax) ;

  popupCentered(dialog->shell, XtGrabExclusive) ;
}
/********************************************************************/
