
/**********************************************************************
 * $Id: trace.c,v 1.1 92/02/27 11:44:04 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xutil.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Toggle.h>

#include <xerion/display.h>

#include "display2Itf.h"
#include "traceDisplay.h"

typedef struct Update {
  char		*commandName ;
  char		*traceName ;
  Widget	toggle ;
} UpdateRec, *Update, **UpdateList ;

/* private functions */
static void	resetTogglesCB  ARGS((Widget, XtPointer, XtPointer)) ;
static void	unsetTogglesCB  ARGS((Widget, XtPointer, XtPointer)) ;
static void	popdownWidgetCB ARGS((Widget, XtPointer, XtPointer)) ;
static void	changeTracesCB  ARGS((Widget, XtPointer, XtPointer)) ;

/*********************************************************************
 *	Name:		createTracePopup
 *	Description:	creates a "Display Updates" dialog with all the
 *			trace names as toggles
 *	Parameters:
 *	  Widget	popup - the popup shell parent
 *	  String	commandName - the command to add to the traces
 *	Return Value:
 *	  export void	createTracePopup - NONE
 *********************************************************************/
export Widget	createTracePopup(parent, commandName)
  Widget	parent ;
  String	commandName ;
{
  Widget	popup, form, radioBox ;
  Widget	label, button, toggle ;
  char		**list, *privateCommandName ;
  int		idx ;
  UpdateList	updateList ;

  privateCommandName = XtNewString(commandName) ;

  popup = createPopupShell("tracePopup", transientShellWidgetClass, parent) ;

  startArgs() ;
  form = createManagedWidget("form", formWidgetClass, popup) ;

  /* allocate the arrays for the update list */
  list = listTraces() ;
  for (idx = 0 ; list[idx] != NULL ; ++idx)
    ;
  updateList = (UpdateList)XtCalloc(idx + 1, sizeof(Update)) ;
  
  startArgs() ;
  radioBox = createManagedWidget("box", boxWidgetClass, form) ;

  startArgs() ;
  label = createManagedWidget("label", labelWidgetClass, radioBox) ;

  for (idx = 0 ; list[idx] != NULL ; ++idx) {
    updateList[idx] = XtNew(UpdateRec) ;
    startArgs() ;
    toggle = createManagedWidget(list[idx], toggleWidgetClass, radioBox) ;
    updateList[idx]->toggle      = toggle ;
    updateList[idx]->commandName = privateCommandName ;
    updateList[idx]->traceName   = XtNewString(list[idx]) ;
  }

  /* Buttons */
  startArgs() ;
  button = createManagedWidget("ok", commandWidgetClass, form) ;
  XtAddCallback(button, XtNcallback, changeTracesCB,  (XtPointer)updateList) ;
  XtAddCallback(button, XtNcallback, popdownWidgetCB, (XtPointer)popup) ;

  startArgs() ;
  button = createManagedWidget("clear", commandWidgetClass, form) ;
  XtAddCallback(button, XtNcallback, unsetTogglesCB, (XtPointer)updateList) ;

  startArgs() ;
  button = createManagedWidget("cancel", commandWidgetClass, form) ;
  XtAddCallback(button, XtNcallback, popdownWidgetCB, (XtPointer)popup) ;

  XtAddCallback(popup, XtNpopupCallback, 
		resetTogglesCB, (XtPointer)updateList) ;

  XtInstallAllAccelerators(form, form) ;

  return popup ;
}
/********************************************************************/


/***********************************************************************
 *	Name:		resetToggle
 *	Description:	changes the state of a toggle button if necessary
 *	Parameters:	
 *	  Widget	toggle  - the toggle button
 *	  Boolean	desired - the desired state of the button (True/False)
 *	Return Value:	NONE
 ***********************************************************************/
static void	resetToggle(toggle, desired)
  Widget	toggle ;
  Boolean	desired ;
{
  Boolean	actual ;
  
  startArgs() ;
  setArg(XtNstate, &actual) ;
  getValues(toggle) ;
  if (actual != desired) {
    startArgs() ;
    setArg(XtNstate, desired) ;
    setValues(toggle) ;
  }
}
/**********************************************************************/


/***********************************************************************
 *	Name:		resetTogglesCB
 *	Description:	resets all the toggles in an update display
 *			to either set or unset, depending on whether
 *			the display update is on that trace
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the client data (updateList)
 *	  XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 ***********************************************************************/
static void	resetTogglesCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  UpdateList	updateList = (UpdateList)clientData ;
  int		idx ;

  for (idx = 0 ; updateList[idx] != NULL ; ++idx) {
    resetToggle(updateList[idx]->toggle, 
		traceOn(updateList[idx]->traceName,
			updateList[idx]->commandName)) ;
  }
}
/**********************************************************************/


/***********************************************************************
 *	Name:		unsetTogglesCB
 *	Description:	unsets all the toggles in an update display
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the client data (UpdateList)
 *	  XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 ***********************************************************************/
static void	unsetTogglesCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  UpdateList	updateList = (UpdateList)clientData ;
  int		idx ;

  for (idx = 0 ; updateList[idx] != NULL ; ++idx) {
    resetToggle(updateList[idx]->toggle, False) ;
  }
}
/**********************************************************************/


/*********************************************************************
 *	Name:		changeTracesCB
 *	Description:	either adds or deletes traces (depending on the
 *			state of the toggles in the Update struct)
 *	Parameters:
 *	  Widget	widget     - a toggle widget
 *	  XtPointer	clientData - an 'Update' structure list.
 *	  XtPointer	callData   - UNUSED
 *	Return Value:
 *	  static void	changeTraceCB - NONE
 *********************************************************************/
static void	changeTracesCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  UpdateList	updateList = (UpdateList)clientData ;
  int		idx ;
  Boolean	set ;

  for (idx = 0 ; updateList[idx] != NULL ; ++idx) {
    Update	update = updateList[idx] ;
    startArgs() ;
    setArg(XtNstate, &set) ;
    getValues(update->toggle) ;
    if (set)	addTrace   (update->traceName, update->commandName) ;
    else	deleteTrace(update->traceName, update->commandName) ;
  }
}
/********************************************************************/


/*********************************************************************
 *	Name:		popupCenteredCB
 *	Description:	pops up the "Display Updates" shell in the
 *			center of its parent
 *	Parameters:
 *	  Widget	widget - the widget the cb is attached  to
 *	  XtPointer	clientData - the popup shell to pop up
 *	  XtPointer	callData - UNUSED
 *	Return Value:
 *	  export void	popupCenteredCB - NONE
 *********************************************************************/
export void	popupCenteredCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	popup  = (Widget)clientData ;

  if (popup == NULL)
    return ;

  popupCentered(popup, XtGrabExclusive) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		popupCentered
 *	Description:	pops up a shell centered in it's parent 
 *	Parameters:
 *	  Widget	popup - the popup shell
 *	  XtGrabKind	grabKind - the grab kind.
 *	Return Value:
 *	  export void	popupCentered - NONE
 *********************************************************************/
export void	popupCentered(popup, grabKind)
  Widget	popup ;
  XtGrabKind	grabKind ;
{
  Position	x, y ;
  Dimension	width, height ;

  startArgs() ;
  setArg(XtNx, &x) ;
  setArg(XtNy, &y) ;
  setArg(XtNwidth, &width) ;
  setArg(XtNheight, &height) ;
  getValues(XtParent(popup)) ;
  
  /* center the popup in the center of its parent */
  x += width/2  ;
  y += height/2 ;
  
  /* make sure the popup doesn't get managed, then realize it,
   * so we can get its dimensions */
  startArgs() ;
  setArg(XtNmappedWhenManaged, FALSE) ;
  setValues(popup) ;
  XtRealizeWidget(popup) ;

  startArgs() ;
  setArg(XtNwidth, &width) ;
  setArg(XtNheight, &height) ;
  getValues(popup) ;

  x -= width/2 ;
  y -= height/2 ;

  startArgs() ;
  setArg(XtNx, x) ;
  setArg(XtNy, y) ;
  setValues(popup) ;

  XtPopup(popup, grabKind) ;
}
/********************************************************************/


/**********************************************************************
 *	Name:		popdownWidgetCB
 *	Description:	pops down a popup shell which is passed in as
 *			the client data
 *	Parameters:
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the shell to pop down
 *	  XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	popdownWidgetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  XtPopdown((Widget)clientData) ;
}
/**********************************************************************/
