
/**********************************************************************
 * $Id: varDisplay.c,v 1.12 92/04/20 16:08:06 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Command.h>

#include <xerion/Selection.h>
#include <xerion/display.h>

#include "display2Itf.h"
#include "display2Sim.h"

#define NO_SELECTION(selectionPtr) \
  ((selectionPtr) == NULL || (selectionPtr)->list_index == XAW_LIST_NONE)

/* private functions */
static void	popdownWidgetCB ARGS((Widget, XtPointer, XtPointer)) ;

static void	resetListCB    ARGS((Widget, XtPointer, XtPointer)) ;
static void 	setSensitiveCB ARGS((Widget, XtPointer, XtPointer)) ;

static void	showValueCB ARGS((Widget, XtPointer, XtPointer)) ;
static void	saveValueCB ARGS((Widget, XtPointer, XtPointer)) ;
static void	showHelpCB  ARGS((Widget, XtPointer, XtPointer)) ;

static Widget	variableList ;

export int	updateVarDisplays() {
  if (topLevel == NULL)
    return 0 ;

  dispatchExposures() ;
  return 1 ;
}

/**********************************************************************
 *	Name:		createVariableDisplay
 *	Description:	creates a list widget with a list of variables
 *			and some buttons for updating the list and
 *			closing the widget
 *	Parameters:	
 *		Widget	popup - the popup shell we are going in
 *	Return Value:
 *		NONE
 **********************************************************************/
export void	createVariableDisplay(popup)
  Widget	popup ;
{
  static char	**objectList ;
  Widget	selection, button ;

  objectList = listSimpleObjectNames() ;

  selection = XtVaCreateManagedWidget("selectionBox",
				      selectionWidgetClass, popup,
				      XtNlist,	objectList,
				      XtNvalue,	"",
				      NULL) ;
  XtAddCallback(selection, XtNcallback, showValueCB,    (XtPointer)selection) ;
  variableList = selection ;

  button = XtSelectionAddButton(selection, "close", popdownWidgetCB, popup) ;

  button = XtSelectionAddButton(selection, "save", saveValueCB, selection) ;
  XtAddCallback(button, XtNcallback, showValueCB, (XtPointer)selection) ;

  button = XtSelectionAddButton(selection, "reset", resetListCB, selection) ;
  XtAddCallback(button, XtNcallback, showValueCB, (XtPointer)selection) ;

  button = XtSelectionAddButton(selection, "help", showHelpCB, selection) ;

  XtAddCallback(popup, XtNpopupCallback, resetListCB, (XtPointer)selection) ;
  XtAddCallback(popup, XtNpopupCallback, showValueCB, (XtPointer)selection) ;

  XtInstallAllAccelerators(XtNameToWidget(selection, "value"), selection) ;

  /* call the necessary callbacks the first time */
  resetListCB(popup, (XtPointer)selection, (XtPointer)NULL) ;
  showValueCB(popup, (XtPointer)selection, (XtPointer)NULL) ;

  XtVaSetValues(popup, XtNcreatePopupChildProc, NULL, NULL) ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		popdownWidgetCB
 *	Description:	pops down a popup shell which is passed in as
 *			the client data
 *	Parameters:
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the shell to pop down
 *	  XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	popdownWidgetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	popup = (Widget)clientData ;
  XtPopdown(popup) ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		setSensitiveCB
 *	Description:	sets the sensitive field to true or false for
 *			all of the widgets that care.
 *	Parameters:
 *	  Widget	widget -  the widget the callback is attached to
 *	  XtPointer	clientData - a pointer to the Selection widget
 *	  XtPointer	callData   - UNUSED
 *	Return Value:
 *	  static void	setSensitiveCB - NONE
 *********************************************************************/
static void	setSensitiveCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	selection = (Widget)clientData ;

  /* if no network, can't save the field */
  if (netExists())
    XtSetSensitive(XtNameToWidget(selection, "save"), TRUE) ;
  else
    XtSetSensitive(XtNameToWidget(selection, "save"), FALSE) ;
}
/********************************************************************/


/**********************************************************************
 *	Name:		resetListCB
 *	Description:	resets the list of a selection box to the proper
 *			values.
 *	Parameters:
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the selection box whose list is to
 *					be updated.
 *	  XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	resetListCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	selectionWidget = (Widget)clientData ;

  XtVaSetValues(selectionWidget, XtNlist, listSimpleObjectNames(), NULL) ;

  XtSelectionHighlight(selectionWidget, 0) ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		showValueCB
 *	Description:	shows the value of the selected item in the
 *			value widget
 *	Parameters:
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the selection box
 *	  XtPointer	callData   - the call data UNUSED
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	showValueCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget		selectionWidget = (Widget)clientData ;
  XawListReturnStruct	*selected ;
  char			*object, *value ;
  
  selected = XtSelectionGetSelected(selectionWidget) ;

  if (NO_SELECTION(selected))
    return ;

  object   = selected->string ;
  value    = getSimpleObjectValue(object, NULL) ;

  XtSelectionSetValueString((Widget)clientData, value) ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		saveCurrentValueCB
 *	Description:	saves the value in the value widget of a 
 *			selection box to the selected object.
 *	Parameters:
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the selection box widget
 *	  XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	saveValueCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget		selectionWidget = (Widget)clientData ;
  XawListReturnStruct	*selected ;
  char			*object, *value ;
  
  selected = XtSelectionGetSelected(selectionWidget) ;

  if (NO_SELECTION(selected))
    return ;

  object   = selected->string ;
  value    = XtSelectionGetValueString(selectionWidget) ;
  if (value != NULL && object != NULL) {
    saveSimpleObjectValue(object, value) ;
  }
}
/**********************************************************************/


/*********************************************************************
 *	Name:		showHelpCB
 *	Description:	prints the "extraHelp" string for a variable.
 *			the extraHelp string is registered by the itf
 *			module 
 *	Parameters:
 *	  Widget	widget - the widget the cb is attached to
 *	  XtPointer	clientData - the Selection widget to get the
 *				variable name from.
 *	  XtPointer	callData - UNUSED
 *	Return Value:
 *	  static void	showHelpCB - NONE
 *********************************************************************/
static void	showHelpCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget		selectionWidget = (Widget)clientData ;
  XawListReturnStruct	*selected ;
  char			**helpStrings ;

  selected = XtSelectionGetSelected(selectionWidget) ;
  if (NO_SELECTION(selected)) {
    fprintf(stdout, "\nSelect an item from the list, then press \"help\"\n") ;
    return ;
  }

  helpStrings = getHelpStrings(selected->string) ;

  fprintf(stdout, "\n%s:\n", selected->string) ;
  if (helpStrings != NULL) {
    while (*helpStrings != NULL)
      fprintf(stdout, "%s\n", *helpStrings++);
  } else {
    fprintf(stdout, "Sorry, no help available\n") ;
  }
}
/********************************************************************/
