
/**********************************************************************
 * $Id: weightDisplay.c,v 1.19 92/04/20 16:08:13 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <math.h>
#include <varargs.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xutil.h>

#include <X11/Shell.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>

#include <xerion/Network.h>
#include <xerion/MenuPane.h>
#include <xerion/RowColumn.h>
#include <xerion/display.h>

#include "display2Itf.h"
#include "display2Sim.h"
#include "displayUtils.h"
#include "traceDisplay.h"
#include "check.h"

extern int	updateActivationDisplay ARGS(()) ;
extern int	touchUpMajorDimensions  ARGS((int, int)) ;

typedef struct {
  Widget	shell ;		/* the popup shell for the display */
  Widget	rowCol ;	/* the row column widget to put the nets in */
  Widget	scaleWidget ;	/* the label widget showing scale etc. */
  String	titleLabel ;	/* the label for the widget */
  String	scaleLabel ;	/* the label for the widget */
  XtStyle	unitStyle ;	/* the style to draw units in */
  int		fieldIdx ;	/* the index of the link field to display */
  float		maxValue ;
  Dimension	border ;
  Boolean	fixedMax ;
} WeightDisplayRec, *WeightDisplay ;


/* private functions */
static void 	toggleShowLabelCB ARGS((Widget,	XtPointer, XtPointer)) ;
static void 	checkBackgroundCB ARGS((Widget,	XtPointer, XtPointer)) ;
static void 	checkFieldCB	ARGS((Widget,	XtPointer, XtPointer)) ;
static void 	changeFieldCB	ARGS((Widget,	XtPointer, XtPointer)) ;
static void 	setBackgroundCB	ARGS((Widget, 	XtPointer, XtPointer)) ;
static void	setFieldIdxCB	ARGS((Widget,	XtPointer, XtPointer)) ;
static void 	setSensitiveCB	ARGS((Widget,	XtPointer, XtPointer)) ;
static void	popdownWidgetCB	ARGS((Widget,	XtPointer, XtPointer)) ;
static void 	printWeightCB	ARGS((Widget,	XtPointer, XtPointer)) ;
static void 	miscellaneousCB ARGS((Widget, XtPointer, XtPointer)) ;

static void	displayFanIn ARGS((int	unitIidx, Widget  *network)) ;
static void	doCheck	     ARGS((Widget	widget, int	flag)) ;

static void	updateDisplay ARGS((WeightDisplay)) ;
static void	rebuildNet ARGS((Widget  network,
				UnitDisplayRec  *unitArray, int  numUnits)) ;

static WeightDisplay	weightDisplay = NULL ;

static XtStyle	styleOption[] = { XtShadow, XtLines, XtLines2, XtWeave,
				    XtPlain, XtSolid } ;
static String	styleName[] = { "shadow", "lines", "lines2", "weave",
				  "plain", "solid" } ;

/***********************************************************************
 *	Name:		createWeightDisplay
 *	Description:	creates the window for the network display
 *			and all of the sub windows and necessary methods
 *	Parameters:	
 *	  Widget	popup - the popup shell that is the parent
 *	Return Value:	
 *			NONE
 ***********************************************************************/
export void	createWeightDisplay(popup)
  Widget	popup ;
{
  Widget	menuPane, menuButton, menu, line ;
  Widget	form, viewport, shell, rowCol ;
  Widget	button, close, updates, reset, *field, *back ;
  String	*list, name ;
  int		idx ;

  weightDisplay = (WeightDisplay)malloc(sizeof(WeightDisplayRec)) ;
  weightDisplay->shell = popup ;

  weightDisplay->fixedMax = FALSE ;
  weightDisplay->border   = -1 ;
  weightDisplay->maxValue = 0.0 ;

  /********************************************
   * Network Name label 
   */
  weightDisplay->titleLabel = getTitle(weightDisplay->shell) ;
  getNetValue(NAME, &name) ;
  titlePrintf(weightDisplay->shell, weightDisplay->titleLabel, name) ;

  startArgs() ;
  form = createManagedWidget("form", formWidgetClass, popup) ;

  /********************************************
   * Pulldown menu pane 
   */
  startArgs() ;
  menuPane = createManagedWidget("menuPane", menuPaneWidgetClass, form) ;

  /**********************
   * Options menu 
   */
  startArgs() ;
  setArg(XtNmenuName, "optionsMenu") ;
  menuButton = createManagedWidget("options", menuButtonWidgetClass, menuPane);

  startArgs() ;
  menu = createPopupShell("optionsMenu", simpleMenuWidgetClass, menuButton) ;
  
  startArgs() ;
  updates = createManagedWidget("updates", smeBSBObjectClass,  menu) ;
  reset	  = createManagedWidget("reset",   smeBSBObjectClass,  menu) ;
  line	  = createManagedWidget("line",	   smeLineObjectClass, menu) ;
  close	  = createManagedWidget("close",   smeBSBObjectClass,  menu) ;

  /**********************
   * "Field" menu 
   */
  startArgs() ;
  setArg(XtNmenuName, "fieldMenu") ;
  menuButton = createManagedWidget("field", menuButtonWidgetClass, menuPane);

  startArgs() ;
  menu = createPopupShell("fieldMenu", simpleMenuWidgetClass, menuButton) ;
  
  list = listConnectionFields() ;
  for (idx = 0 ; list[idx] != NULL ; ++idx)
    ;
  field = (Widget *)XtCalloc(idx + 1, sizeof(Widget)) ;

  startArgs() ;
  setArg(XtNleftMargin, check_width) ;
  for (idx = 0 ; list[idx] != NULL ; ++idx)
    field[idx] = createManagedWidget(list[idx], smeBSBObjectClass, menu) ;
  XtAddCallback(menu, XtNpopupCallback, checkFieldCB, (XtPointer)field) ;

  /**********************
   * Network menu 
   */
  startArgs() ;
  setArg(XtNmenuName, "networkMenu") ;
  menuButton = createManagedWidget("network",
				   menuButtonWidgetClass, menuPane);
  startArgs() ;
  menu = createPopupShell("networkMenu", simpleMenuWidgetClass, menuButton);
  
  back = (Widget *)XtCalloc(XtNumber(styleOption) + 1, sizeof(Widget)) ;
  startArgs() ;
  setArg(XtNleftMargin, check_width) ;
  for (idx = 0 ; idx < XtNumber(styleOption) ; ++idx)
    back[idx] = createManagedWidget(styleName[idx], smeBSBObjectClass, menu) ;

  startArgs() ;
  line = createManagedWidget("line", smeLineObjectClass, menu) ;

  startArgs() ;
  setArg(XtNlabel, "Hide Label") ;
  button = createManagedWidget("showLabel", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, 
		toggleShowLabelCB, (XtPointer)weightDisplay) ;

  startArgs() ;
  button = createManagedWidget("misc", smeBSBObjectClass, menu) ;
  XtAddCallback(button, XtNcallback, 
		miscellaneousCB, (XtPointer)weightDisplay) ;
  XtAddCallback(menu, XtNpopupCallback, checkBackgroundCB, (XtPointer)back) ;

  /********************************************
   * Display Updates 
   */
  shell = createTracePopup(popup, "updateWeightDisplay") ;

  startArgs() ;
  weightDisplay->scaleWidget 
    = createManagedWidget("scaleLabel", labelWidgetClass, form) ;
  weightDisplay->scaleLabel = getLabel(weightDisplay->scaleWidget) ;
  labelPrintf(weightDisplay->scaleWidget, 
	      weightDisplay->scaleLabel, 0.0, 0.0, 0.0) ;

  /********************************************
   * Network 
   */
  startArgs() ;
  viewport = createManagedWidget("viewport", viewportWidgetClass, form) ;
  startArgs() ;
  rowCol = createManagedWidget("rowColumn", rowColumnWidgetClass, viewport) ;
  weightDisplay->rowCol = rowCol ;

  /********************************************
   * Add all the callbacks 
   */
  XtAddCallback(updates, XtNcallback, 
		popupCenteredCB, (XtPointer)shell) ;
  XtAddCallback(reset, XtNcallback, 
		setSensitiveCB,	(XtPointer)weightDisplay) ;
  XtAddCallback(close, XtNcallback, 
		popdownWidgetCB, (XtPointer)popup) ;
  XtAddCallback(popup, XtNpopupCallback, 
		setSensitiveCB, (XtPointer)weightDisplay) ;
  weightDisplay->unitStyle = XtShadow ;
  for (idx = 0 ; back[idx] != NULL ; ++idx) {
    XtAddCallback(back[idx], XtNcallback, 
		  setBackgroundCB, (XtPointer)styleOption[idx]) ;
  }
  weightDisplay->fieldIdx = 0 ;
  for (idx = 0 ; list[idx] != NULL ; ++idx) {
    XtAddCallback(field[idx], XtNcallback, setFieldIdxCB, (XtPointer)idx) ;
  }

  XtInstallAllAccelerators(form, form) ;

  /********************************************
   * Do all the popup CB's for the first time 
   */
  setSensitiveCB(popup,  (XtPointer)weightDisplay, (XtPointer)NULL) ;

  startArgs() ;
  setArg(XtNcreatePopupChildProc, NULL) ;
  setValues(popup) ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		doCheck
 *	Description:	puts a check beside a smeBSB (or clears it)
 *			depending on the flag passed in
 *	Parameters:
 *	  Widget	widget - the widget to put the check in
 *	  Boolean	flag   - TRUE - put check, FALSE - clear it
 *	Return Value:
 *	  static void	doCheck - NONE
 *********************************************************************/
static void	doCheck(widget, flag)
  Widget	widget ;
  int		flag ;
{
  static Pixmap	check = NULL ;

  if (check == NULL) {
    check = XCreateBitmapFromData(XtDisplayOfObject(widget),
				  XtWindowOfObject(widget), 
				  check_bits, check_width, check_height) ;
  }
  startArgs() ;
  setArg(XtNleftMargin, check_width) ;
  if (flag) {
    setArg(XtNleftBitmap, check) ;
  } else {
    setArg(XtNleftBitmap, NULL) ;
  }
  setValues(widget) ;
}
/********************************************************************/

static void 	toggleShowLabelCB (widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	label = ((WeightDisplay)clientData)->scaleWidget ;
  String	toggleLabel ;
  
  toggleLabel = getLabel(widget) ;
  if (strcmp(toggleLabel, "Hide Label") == 0) {
    XtUnmapWidget(label) ;
    labelPrintf(widget, "Show Label") ;
  } else {
    XtMapWidget(label) ;
    XRaiseWindow(XtDisplay(label), XtWindow(label)) ;
    labelPrintf(widget, "Hide Label") ;
  }
  XtFree(toggleLabel) ;
}


/**********************************************************************
 *	Name:		setFieldIdxCB
 *	Description:	sets the mask of what connection field to
 *			display
 *	Parameters:
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the index to set it to
 *	  XtPointer	callData   - the call data (XtListReturnStruct *)
 *	Return Value:
 *		NONE
 
 **********************************************************************/
static void	setFieldIdxCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  weightDisplay->fieldIdx = (int)clientData ;
  updateDisplay(weightDisplay) ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		checkFieldCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData - 
 *	Return Value:
 *	  static void	checkFieldCB - 
 *********************************************************************/
static void	checkFieldCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	*option = (Widget *)clientData ;
  int		idx ;

  for (idx = 0 ; option[idx] != NULL ; ++idx) {
    if (weightDisplay->fieldIdx == idx)
      doCheck(option[idx], TRUE) ;
    else
      doCheck(option[idx], FALSE) ;
  }
}
/********************************************************************/


/*********************************************************************
 *	Name:		setBackgroundCB
 *	Description:	sets the variable 'unitStyle' through the
 *			allowable background styles
 *	Parameters:
 *	  Widget	widget - the widget the callback is attached to
 *	  XtPointer	clientData - the style to use
 *	  XtPointer	callData   - UNUSED
 *	Return Value:
 *	  static void	setBackgrouncCB - NONE
 *********************************************************************/
static void	setBackgroundCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  weightDisplay->unitStyle = (XtStyle)clientData ;
  rebuildWeightFlag = TRUE ;
  updateDisplay(weightDisplay) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		checkBackgroundCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData - 
 *	Return Background:
 *	  static void	checkBackgroundCB - 
 *********************************************************************/
static void	checkBackgroundCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	*option = (Widget *)clientData ;
  int		idx ;

  for (idx = 0 ; option[idx] != NULL ; ++idx) {
    if (weightDisplay->unitStyle == styleOption[idx])
      doCheck(option[idx], TRUE) ;
    else
      doCheck(option[idx], FALSE) ;
  }
}
/********************************************************************/


/***********************************************************************
 *	Name:		updateWeightDisplay
 *	Description:	this little hack allows outside users to
 *			update the weight display without clicking
 *			or popping anything
 *	Parameters:	NONE
 *	Return Value:	NONE
 ***********************************************************************/
export void updateWeightDisplay()
{
  if (topLevel == NULL || weightDisplay == NULL)
    return ;

  setSensitiveCB(weightDisplay->shell, 
		 (XtPointer)weightDisplay, (XtPointer)NULL) ;
  dispatchExposures() ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		updateDisplay
 *	Description:	rebuilds a network Widget showing the current
 *			networks weight levels
 *		Widget	widget 	   - the widget the callback is attached to
 *	Return Value:
 *		NONE
 ***********************************************************************/
static double	scale, mean, var ;
static Widget	rowCol ;
static void	updateDisplay(display)
  WeightDisplay	display ;
{
  int		netIdx ;

  static Widget	*Network = NULL ;
  static int	 MaxNets, NumNets ;

  rowCol = display->rowCol ;

  if (!netExists() || rebuildWeightFlag == TRUE) {
    int		height, width, margin ;

    getNetValue(MAJOR_HEIGHT, &height) ;
    getNetValue(MAJOR_WIDTH,  &width) ;
    getNetValue(MAJOR_MARGIN, &margin) ;
    XtUnmanageChild(rowCol) ;
    /* reinitialize the row and column sizes, destroy old network widgets */
    startArgs() ;
    setArg(XtNcolumnWidth,   width) ;
    setArg(XtNrowHeight,     height) ;
    setArg(XtNvertDistance,  margin) ;
    setArg(XtNhorizDistance, margin) ;
    setValues(rowCol) ;
    for (netIdx = 0 ; netIdx < NumNets ; ++netIdx) {
      if (Network[netIdx] != NULL)
	XtDestroyWidget(Network[netIdx]) ;
      Network[netIdx] = NULL ;
    }
  }

  /* We got a network, so calculate the necessary stuff */
  NumNets = getNumberOfUnits() ;

  if (NumNets > MaxNets) {
    Network  = (Widget *)XtRealloc((char *)Network, NumNets*sizeof(Widget)) ;
    for (netIdx = MaxNets ; netIdx < NumNets ; ++netIdx)
      Network[netIdx] = NULL ;
    MaxNets = NumNets ;
  }

  if (!netExists()) {
    char	*name ;
    getNetValue(NAME, &name) ;
    titlePrintf(display->shell, display->titleLabel, name) ;
    labelPrintf(display->scaleWidget,
		display->scaleLabel, 0.0, 0.0, 0.0) ;
    if (display->fixedMax != TRUE) 
      display->maxValue = 0.0 ;

  } else {
    char	*name ;
    getWeightParams(display->fieldIdx, &mean, &var, &scale) ;
    if (display->fixedMax != TRUE) 
      display->maxValue = scale ;

    getNetValue(NAME, &name) ;
    titlePrintf(display->shell, display->titleLabel, name) ;
    labelPrintf(display->scaleWidget, 
		display->scaleLabel, display->maxValue, mean, var);
    displayForAllUnits(displayFanIn, Network) ; 

    for (netIdx = NumNets ; netIdx < MaxNets ; ++netIdx) {
      if (Network[netIdx] != NULL)
	XtUnmanageChild(Network[netIdx]) ;
    }
  }
  XtManageChild(rowCol) ;
  rebuildWeightFlag = FALSE ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		displayFanIn 
 *	Description:	I'm really sorry about this bit of code.
 *			Please accept my apologies, but I had to
 *			hack it together really quick and I know
 *			it totally sucks.
 *	Parameters:	
 *	Return Value:	
 ***********************************************************************/
static void	displayFanIn (unitIdx, network)
  int		unitIdx ;
  Widget	*network ;
{
  UnitDisplayRec	*unitArray ;
  int			idx, numUnits, numOldUnits ;

  if (getUnitMajorRow() < 0 || getUnitMajorColumn() < 0) 
    return ;

  unitArray = getFanIn(weightDisplay->fieldIdx, &unitArray, &numUnits) ;
  for (idx = 0 ; idx < numUnits ; ++idx)
    unitArray[idx].value /= weightDisplay->maxValue ;

  if (network[unitIdx] == NULL) {
    startArgs() ;
    setArg(XtNrow,      getUnitMajorRow()) ;
    setArg(XtNcolumn,   getUnitMajorColumn()) ;
    network[unitIdx] = createManagedWidget("weightNet", 
					   networkWidgetClass, rowCol) ;
    XtAddCallback(network[unitIdx], XtNchangeValueCallback, 
		  changeFieldCB, (XtPointer)weightDisplay) ;
    XtAddCallback(network[unitIdx], XtNqueryCallback, 
		  printWeightCB, (XtPointer)network[unitIdx]) ;
  }
  XtVaGetValues(network[unitIdx], XtNnumUnits, &numOldUnits, NULL) ;

  if (numUnits == 0) {
    /* No network, blank out the widget */
    XtUnmanageChild(network[unitIdx]) ;
  } else if (rebuildWeightFlag == TRUE || numUnits != numOldUnits) {
    rebuildNet(network[unitIdx], unitArray, numUnits) ;
  } else {
    XtNetworkSetUnits(network[unitIdx], unitArray) ;
    XFlush(XtDisplay(network[unitIdx])) ;
  }
}
/**********************************************************************/

/***********************************************************************
 *	Name:		rebuildNet
 *	Description:	rebuilds the units and size of a network widget
 *	  Widget		network    - the network widget to update
 *	  UnitDisplayRec	*unitArray - the array of units to  update with
 *	  int		numUnits   - the number of units
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void	rebuildNet(network, unitArray, numUnits)
  Widget		network ;
  UnitDisplayRec	*unitArray ;
  int			numUnits ;
{
  int		height, width, margin ;
  int		majorHeight, majorWidth ;

  getNetValue(MAJOR_HEIGHT, &majorHeight) ;
  getNetValue(MAJOR_WIDTH,  &majorWidth) ;
  getNetValue(MINOR_HEIGHT, &height) ;
  getNetValue(MINOR_WIDTH,  &width) ;
  getNetValue(MINOR_MARGIN, &margin) ;

  XtUnmanageChild(network) ;
  startArgs() ;
  setArg(XtNunitBackgroundStyle, weightDisplay->unitStyle) ;
  if ((short)weightDisplay->border >= 0) {
    setArg(XtNunitBorderWidth, weightDisplay->border) ;
  }
  setArg(XtNheight,	majorHeight) ;
  setArg(XtNwidth,	majorWidth) ;
  setArg(XtNunitHeight,	height) ;
  setArg(XtNunitWidth, 	width) ;
  setArg(XtNunitSep,	margin) ;
  setArg(XtNnumUnits,	numUnits) ;
  setArg(XtNunitArray,	unitArray) ;
  setValues(network) ;
  XtManageChild(network) ;
  
  XtVaGetValues(network, XtNunitBorderWidth, &weightDisplay->border, NULL) ;
}
/**********************************************************************/


/**********************************************************************
 *	Name:		popdownWidgetCB
 *	Description:	pops down a popup shell which is passed in as
 *			the client data
 *	Parameters:
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the shell to pop down
 *	  XtPointer	callData   - the call data (unused)
 *	Return Value:
 *		NONE
 **********************************************************************/
static void	popdownWidgetCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget	popup = (Widget)clientData ;
  XtPopdown(popup) ;
}
/**********************************************************************/
 

/*********************************************************************
 *	Name:		setSensitiveCB
 *	Description:	sets the sensitive field to true or false for
 *			all of the widgets that care.
 *	Parameters:
 *	  Widget	widget -  the widget the callback is attached to
 *	  XtPointer	clientData - a pointer to the weightDisplay record
 *	  XtPointer	callData   - UNUSED
 *	Return Value:
 *	  static void	setSensitiveCB - NONE
 *********************************************************************/
static void	setSensitiveCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  WeightDisplay	display = (WeightDisplay)clientData ;
  Widget	menuPane ;
  Boolean	sensitive ;

  menuPane = XtNameToWidget(display->shell, "form.menuPane") ;

  if (netExists())
    sensitive = TRUE ;
  else
    sensitive = FALSE ;

  /* if no network, can't set the background of units */
  XtSetSensitive(XtNameToWidget(menuPane, "network"), sensitive) ;

  /* don't forget the label saying what we're showing */
  XtSetSensitive(display->scaleWidget, sensitive) ;
  
  /* If no network, can't change the field being displayed either  */
  XtSetSensitive(XtNameToWidget(menuPane, "field"), sensitive) ;
  
  updateDisplay(display) ;
}
/********************************************************************/


/***********************************************************************
 *	Name:		printWeightCB
 *	Description:	prints out the weight level of a unit
 *	Parameters:	
 *	  Widget	widget 	   - the widget the callback is attached to
 *	  XtPointer	clientData - the network widget
 *	  XtPointer	callData   - NetworkReturnStruct *
 *	Return Value:	
 *		NONE
 ***********************************************************************/
static void printWeightCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  Widget		network = (Widget)clientData ;
  NetworkReturnStruct	*returnStruct = (NetworkReturnStruct *)callData ;
  UnitDisplayRec	*UnitPtr ;

  UnitPtr = XtNetworkGetUnit(network, returnStruct->index) ;

  if (UnitPtr == NULL) {
    fprintf(stderr, "Yow! This unit doesn't seem to exist!\n") ;
  } else {
    fprintf(stdout, "\nLink \"%s\" has value %.4g\n",
	    (char *)UnitPtr->userData, 
	    UnitPtr->value*weightDisplay->maxValue);
  }
  XtFree((char *)UnitPtr) ;
}
/**********************************************************************/
static void changeFieldCB(network, clientData, callData)
  Widget	network ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  WeightDisplay		display = (WeightDisplay)clientData ;
  NetworkReturnStruct	*returnStruct = (NetworkReturnStruct *)callData ;
  UnitDisplayRec	*unitPtr ;

  unitPtr = XtNetworkGetUnit(network, returnStruct->index) ;
  unitPtr->value *= display->maxValue ;
  changeLinkField(unitPtr, display->fieldIdx) ;
  updateNetActivities() ;
  updateActivationDisplay() ;
}
/**********************************************************************/


typedef struct {
  Widget	shell ;
  Widget	fixedMax ;
  Widget	maxValue ;
  Widget	size ;
  Widget	border ;
  Widget	margin ;
  Widget	blockHeight ;
  Widget	fixedBlockHeight ;
  Widget	blockWidth ;
  Widget	fixedBlockWidth ;
  Widget	blockMargin ;
} *MiscNetworkDialog ;

/*********************************************************************
 *	Name:		setMiscellaneousCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData -
 *	Return Value:
 *	  static void	setMiscellaneousCB - 
 *********************************************************************/
static void	setMiscellaneousCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  MiscNetworkDialog	dialog = (MiscNetworkDialog)clientData ;
  Boolean		fixed ;
  int			size, margin ;
  float			maxValue ;
  String		string ;

  string = getText(dialog->maxValue) ;
  if (string != NULL) {
    maxValue = atof(string) ;
    weightDisplay->maxValue = maxValue ;
  }

  startArgs() ;
  setArg(XtNstate, &fixed) ;
  getValues(dialog->fixedMax) ;
  if (weightDisplay->fixedMax != fixed) {
    weightDisplay->fixedMax = fixed ;
  }

  string = getText(dialog->size) ;
  if (string != NULL) {
    size = atol(string) ;
    setNetValue(MINOR_WIDTH,  &size) ;
    setNetValue(MINOR_HEIGHT, &size) ;
  }

  string = getText(dialog->border) ;
  if (string != NULL)
    weightDisplay->border = atol(string) ;
  else
    weightDisplay->border = -1 ;

  string = getText(dialog->margin) ;
  if (string != NULL) {
    margin = atol(string) ;
    setNetValue(MINOR_MARGIN,  &margin) ;
  }

  startArgs() ;
  setArg(XtNstate, &fixed) ;
  getValues(dialog->fixedBlockHeight) ;
  if (fixed) {
    string = getText(dialog->blockHeight) ;
    if (string != NULL) {
      size = atol(string) ;
      setNetValue(MAJOR_HEIGHT,  &size) ;
    }
  } else {
    touchUpMajorDimensions(FALSE, TRUE) ;
  }

  startArgs() ;
  setArg(XtNstate, &fixed) ;
  getValues(dialog->fixedBlockWidth) ;
  if (fixed) {
    string = getText(dialog->blockWidth) ;
    if (string != NULL) {
      size = atol(string) ;
      setNetValue(MAJOR_WIDTH,  &size) ;
    }
  } else {
    touchUpMajorDimensions(TRUE, FALSE) ;
  }

  string = getText(dialog->blockMargin) ;
  if (string != NULL) {
    size = atol(string) ;
    setNetValue(MAJOR_MARGIN,  &size) ;
  }

  rebuildWeightFlag = TRUE ;
  updateDisplay(weightDisplay) ;
}


/*********************************************************************
 *	Name:		miscellaneousCB
 *	Description:	
 *	Parameters:
 *	  Widget	widget - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData - 
 *	Return Value:
 *	  static void	miscellaneousCB - 
 *********************************************************************/
static void	miscellaneousCB(widget, clientData, callData)
  Widget	widget ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  WeightDisplay	display = (WeightDisplay)clientData ;
  int		size ;
  Boolean	fixedMax ;
  
  static MiscNetworkDialog	dialog = NULL ;

  if (dialog == NULL) {
    Widget	form, button, table ;
    Position	x, y ;

    dialog = (MiscNetworkDialog)XtMalloc(sizeof(*dialog)) ;

    startArgs() ;
    dialog->shell = createPopupShell("miscShell", 
				     transientShellWidgetClass, 
				     display->shell) ;

    startArgs() ;
    form = createManagedWidget("form", formWidgetClass, dialog->shell) ;

    table = startTable(form) ;
    dialog->maxValue	= addTableItem("maxValue",    &table, TRUE) ;
    dialog->size	= addTableItem("size", 	      &table, FALSE) ;
    dialog->border	= addTableItem("border",      &table, FALSE) ;
    dialog->margin	= addTableItem("margin",      &table, FALSE) ;
    dialog->blockHeight	= addTableItem("blockHeight", &table, TRUE) ;
    dialog->blockWidth	= addTableItem("blockWidth",  &table, TRUE) ;
    dialog->blockMargin	= addTableItem("blockMargin", &table, FALSE) ;

    dialog->fixedMax    = XtNameToWidget(form, "maxValueToggle") ;
    dialog->fixedBlockWidth  = XtNameToWidget(form, "blockWidthToggle") ;
    dialog->fixedBlockHeight = XtNameToWidget(form, "blockHeightToggle") ;

    startArgs() ;
    setArg(XtNfromVert,	dialog->blockMargin) ;
    button = createManagedWidget("ok", commandWidgetClass, form) ;
    XtAddCallback(button, XtNcallback, setMiscellaneousCB, (XtPointer)dialog) ;
    XtAddCallback(button, XtNcallback, 
		  popdownWidgetCB, (XtPointer)dialog->shell) ;

    startArgs() ;
    setArg(XtNfromVert,	dialog->blockMargin) ;
    setArg(XtNfromHoriz,button) ;
    button = createManagedWidget("cancel", commandWidgetClass, form) ;
    XtAddCallback(button, XtNcallback, 
		  popdownWidgetCB, (XtPointer)dialog->shell) ;
    
    XtInstallAllAccelerators(form, form) ;

    /* center the popup in the screen */
    x = WidthOfScreen(XtScreen(topLevel))/2  - 150 ;
    y = HeightOfScreen(XtScreen(topLevel))/2 - 100 ;
    
    startArgs() ;
    setArg(XtNx, x) ;
    setArg(XtNy, y) ;
    setValues(dialog->shell) ;
  }

  if (netExists()) {
    fixedMax = display->fixedMax ;
    startArgs() ;
    setArg(XtNstate, fixedMax) ;
    setValues(dialog->fixedMax) ;
    
    textPrintf(dialog->maxValue, "%g", display->maxValue) ;

    getNetValue(MINOR_WIDTH, &size) ;
    textPrintf(dialog->size, "%d", size) ;

    textPrintf(dialog->border, "%d", display->border) ;

    getNetValue(MINOR_MARGIN, &size) ;
    textPrintf(dialog->margin, "%d", size) ;

    getNetValue(MAJOR_WIDTH, &size) ;
    textPrintf(dialog->blockWidth, "%d", size) ;

    getNetValue(MAJOR_HEIGHT, &size) ;
    textPrintf(dialog->blockHeight, "%d", size) ;

    getNetValue(MAJOR_MARGIN, &size) ;
    textPrintf(dialog->blockMargin, "%d", size) ;
  }

  popupCentered(dialog->shell, XtGrabExclusive) ;
}
/********************************************************************/
