
/**********************************************************************
 * $Id: minor.c,v 1.11 92/04/20 16:11:21 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <strings.h>

#include <xerion/commands.h>
#include "layoutUtils.h"
#include "minor.h"

static int	perRow ;
static Unit	forUnit ;
static void	setUnitDefault ARGS((Unit	unit, void	*data)) ;
static void	setLinkDefault ARGS((Link	link, void	*data)) ;

/***********************************************************************
 *	Name:		setMinorOptions
 *	Description:	
 *	Parameters:	
 *	Return Value:	
 ***********************************************************************/
int	setMinorOptions (argv, argc)
  char	**argv ;
  int	argc ;
{
  int	line = layoutInputLine ;

  if (argc == 3 && strcmp(argv[1], "for-unit") == 0) {
    forUnit = unitFromName(currentNet, argv[2]) ;
    if (forUnit == NULL)
      IErrorAbort("Line %d: unknown unit \"%s\"", line, argv[2]) ;
    unitForAllLinks(forUnit, INCOMING, setLinkDefault, NULL) ;

  } else {
    forUnit = NULL ;
    for (--argc, ++argv ; argc > 0 ; --argc, ++argv) {
      char	*ptr, *option = *argv ;

      --argc, ++argv ;
      if (argc == 0) {
	IErrorAbort("Line %d: missing argument for \"%s\"", line, option) ;

      } else if (strcmp(option, "cell-size") == 0) {
	currentNet->minorHeight = strtol(*argv, &ptr, 0) ;
	currentNet->minorWidth  = currentNet->minorHeight ;
	if (*ptr != '\0') 
	  IErrorAbort("Line %d: bad cell size \"%s\"", line, *argv) ;

      } else if (strcmp(option, "margin") == 0) {
	currentNet->minorMargin = strtol(*argv, &ptr, 0) ;
	if (*ptr != '\0') 
	  IErrorAbort("Line %d: bad margin size \"%s\"", line, *argv) ;

      } else if (strcmp(option, "per-row") == 0) {
	perRow = strtol(*argv, &ptr, 0) ;
	if (*ptr != '\0') 
	  IErrorAbort("Line %d: bad units per row \"%s\"", line, *argv) ;

      } else if (strcmp(option, "for-unit") == 0) {
	IErrorAbort("Line %d: \"%s\" cannot appear with other options",
		    line, option);

      } else {
	IErrorAbort("Line %d: bad option \"%s\"", line, option) ;
      }
    }

    netForAllUnits(currentNet, ALL, setUnitDefault, NULL) ;
  }
  
  resetLastLaid() ;
  resetCurrentRow() ;

  return 1 ;
}
/**********************************************************************/
static void	setLinkDefault(link, data)
  Link		link ;
  void		*data ;
{
  link->weightRow    = -1 ;
  link->weightColumn = -1 ;
}
/**********************************************************************/
static void	setUnitDefault(unit, data)
  Unit		unit ;
  void		*data ;
{
  unitForAllLinks(unit, INCOMING, setLinkDefault, NULL) ;
}
/**********************************************************************/


/**********************************************************************/
static Link	findLinkBetween(from, to)
  Unit		from ;
  Unit		to ;
{
  int	idx ;
  int	numIncoming = to->numIncoming ;
  Link	*incoming   = to->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx)
    if (incoming[idx]->preUnit == from)
      break ;

  if (idx < numIncoming)
    return incoming[idx] ;
  else
    return NULL ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		setMinorLine
 *	Description:	
 *	Parameters:	
 *	Return Value:	
 ***********************************************************************/
int	setMinorLine(argv, argc) 
  char	**argv ;
  int	argc ;
{
  int		unitIdx ;
  int		currentColumn ;
  Group		group ;
  Section	*section ;

  section = buildSections(argv, argc) ;
  if (section == NULL)
    return 0 ;

  section = fillSections(section, perRow) ;
  if (section == NULL) 
    return 0 ;

  currentColumn = 0 ;
  while (section->type != End) {
    Link	link ;
    switch (section->type) {
    case All:
    case Sect:
    case Next:
      group = groupFromName(currentNet, section->name) ;
      for (unitIdx = section->from ; 
	   unitIdx < section->from + section->size ; ++unitIdx ) {
	if (forUnit
	    && (link = findLinkBetween(group->unit[unitIdx], forUnit))) {
	  link->weightRow    = getCurrentRow() ;
	  link->weightColumn = currentColumn ;
	} else {
	  int	numOutgoing = group->unit[unitIdx]->numOutgoing ;
	  Link	*outgoing   = group->unit[unitIdx]->outgoingLink ;
	  int	idx ;

	  for (idx = 0 ; idx < numOutgoing ; ++idx) {
	    outgoing[idx]->weightRow    = getCurrentRow() ;
	    outgoing[idx]->weightColumn = currentColumn ;
	  }
	}
	++currentColumn ;
      }
      break ;

    case Blank:
    case Fill:
      currentColumn += section->size ;
      break ;
    default:
      break ;
    }
    ++section ;
  }

  return 1 ;
}
/**********************************************************************/
