/**********************************************************************
 * Copyright 1990,1991 by University of Toronto, Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee, 
 * provided that the above copyright notice appears in all copies and that 
 * both the copyright notice and this permission notice appear in 
 * supporting documentation, and that the name of University of Toronto 
 * not be used in advertising or publicity pertaining to distribution 
 * of the software without specific, written prior permission.  
 * University of Toronto makes no representations about the suitability 
 * of this software for any purpose. It is provided "as is" without 
 * express or implied warranty. 
 *
 * UNIVERSITY OF TORONTO DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS, IN NO EVENT SHALL UNIVERSITY OF TORONTO BE LIABLE FOR ANY 
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER 
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF 
 * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 **********************************************************************/

#include <math.h>
#include <useful.h>
#include <simulator.h>
#include <commands.h>
#include <xerion/minimize.h>
#include <xerion/example.h>
#include "utilsDCP.h"

int command_multi (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  int		u, ex, ex2, example, same, diff, nSame = 2 ;
  Real		prox ;
  static int	numExamples = 0, numUnits = 0  ;
  static Real	**hidden = NULL ;
  static Real	withinError, betweenError, diffProximity, sameProximity ;
  Group		Hidden, Output1, Output2 ; 
  struct ARG_DESC *argd ;

  argd = StartArgs(tokv[0]);
  Args(argd, "[-Pnsame <n>%i%?]", &nSame, "number of examples for each concept [2]") ;
  EndArgs(argd);
  if (GiveHelp(tokc)) {
    ISynopsis("updates within and between errors (TRACE ONLY)") ;
    IHelp
      (tokc, tokv[0], NULL, synopsis,
       "Examples are assumed to be ordered such that all 'nsame' examples" ,
       "involving the same concept occur together and that even-numbered" ,
       "examples are 'within' for Output1 group"  ,
       NULL);
    return 0;
  }
  (void) ParseArgs(argd, tokc, tokv, 0);
  if (!currentNet) IErrorAbort("There is no current net.") ;
  if (currentNet->batchSize <= 0) return ;

  if (currentNet->mz->repetitionCount==0 || 
      currentNet->mz->nIterations%currentNet->mz->repetitionCount == 0) {

    if ((Hidden=groupFromName(currentNet, "Hidden")) == NULL) 
      IErrorAbort("Group \"Hidden\" does not exists.") ;
    if ((Output1=groupFromName(currentNet, "Output1")) == NULL)
      IErrorAbort("Group \"Output1\" does not exists.") ;
    if ((Output2=groupFromName(currentNet, "Output2")) == NULL)
      IErrorAbort("Group \"Output2\" does not exists.") ;

    example = currentNet->trainingExampleSet->currentIdx ;
    if (example < 0) IErrorAbort("No training example loaded.") ;

    if (!hidden || 
	numExamples != currentNet->batchSize || 
	numUnits != Hidden->numUnits) {
      numExamples = currentNet->batchSize ;
      numUnits    = Hidden->numUnits ;
      if (hidden) free(hidden) ;
      hidden = (Real **)malloc(numExamples*sizeof(Real*)) ;
      for (ex = 0; ex < numExamples; ex++) 
	hidden[ex] = (Real *)malloc(numUnits*sizeof(Real)) ;
    }

    if (example == 0) withinError = betweenError = 0;

    saveOutputs(Hidden, hidden[example]) ;
    /* EVEN EXAMPLES ARE WITHIN */
    if (example%2) {
      withinError   += groupError(Output2) ;
      betweenError  += groupError(Output1) ;
    } else {
      withinError  += groupError(Output1) ;
      betweenError += groupError(Output2) ;
    }

    if (example == numExamples-1) {
      sameProximity = diffProximity = 0.0 ;
      same = diff = 0 ;
      for (ex = 0; ex < numExamples-1; ex++)  
	for (ex2 = ex+1; ex2 < numExamples; ex2++) {
	  prox = proximity(hidden[ex], hidden[ex2], numUnits) ;
	  if (floor(ex/nSame) == floor(ex2/nSame)) {
	    sameProximity += prox ;  same++ ;
	  } else {
	    diffProximity += prox ;  diff++ ;
	  }
	}
      
      fprintf(dout, "\tmulti:  wIn= %-9.6g bTn= %-9.6g smP= %-9.6g dfP= %-9.6g\n",
	      withinError, betweenError, sameProximity/same, diffProximity/diff) ;
    }
  }      
  return 1 ;
}
