#ifndef UTILITIES_H
#define UTILITIES_H

extern	Real	now			ARGS(( )) ;
extern	Real	gaussianNoise		ARGS((double sd)) ;
extern	void	corruptWeights		ARGS((Group preGroup, Group postGroup,
					      double sd)) ;
extern	void	restoreWeights		ARGS((Group preGroup, Group postGroup)) ;
extern	Real	proximity		ARGS((Real* v1, Real* v2, int n)) ;
extern	Real	correlation		ARGS((Real* v1, Real* v2, int n)) ;
extern	void	saveOutputs		ARGS((Group group, Real *v)) ;
extern	void	restoreOutputs		ARGS((Group group, Real *v)) ;
extern	Real	proximity		ARGS((Real* v1, Real* v2, int n)) ;
extern	Real	groupOutputProximity	ARGS((Group group, Real* v)) ;
extern	Real	groupError		ARGS((Group group)) ;
extern	void	*getExampleSetData	ARGS((ExampleSet exampleSet)) ;
extern	void	putExampleSetData	ARGS((ExampleSet exampleSet, void* data)) ;
extern	void	initEpoch 		ARGS((Net net, int onset)) ;
extern	int	lastTargetSlice		ARGS((Net net)) ;
extern	void	lesion			ARGS((Group preGroup, Group postGroup,
					      double prob, double minWt,
					      int print)) ;
extern	Boolean	outputCorrect		ARGS((Net net, 
					      double tolerance,
					      int training)) ;

extern	int	maxOutputPosition	ARGS((Group group)) ;
extern	int	cleanupWTA		ARGS((Group group)) ;
extern	void	cleanupThreshold	ARGS((Group group, void *data)) ;

extern	void	Fixate			ARGS((Group Letters,
					      Group Attention,
					      int   saccade,
					      int   attention,
					      int   time)) ;

#define MIN_OUTPUT	0.000001
#define MAX_OUTPUT	0.999999

#define MAX_INPUT	15.0

#define WITH_PROBABILITY(p)	((p)==1.0 || (((Real)random()) / MAXINT) < (p))

#define uniformNoise(lo,hi)	(((Real)random()/MAXINT)*((hi)-(lo))+(lo))

#define randomInteger(hi)	(random()%hi)
			 /* returns int among [0, hi-1] */

#define MIN(x, y)		((x) < (y) ? (x) : (y))
#define MAX(x, y)		((x) > (y) ? (x) : (y))
#define CLIP(x, min, max)	((x) > (max) ? (max) : ((x) < (min) ? (min) : (x)))

#define IS_BIDIRECTIONAL(link)		((link)->type & BIDIRECTIONAL)


#define simpleSigmoid(x) (1.0/(1.0+exp(-(double)(x))))
#define sigmoid(x)       ((x) > MAX_INPUT ? 1.0 : 	\
			  ((x) < -MAX_INPUT ? 0.0 : 	\
			   simpleSigmoid(x)))
#define sigmoidDeriv(x)	((x)*(1.0 - (x)))

#define sigmoidInv(y)	((y)<MIN_OUTPUT ? (-MAX_INPUT) :	\
			 ((y)>MAX_OUTPUT ? MAX_INPUT : 		\
			  log((y)/(1.0-(y)))))

#define CROSS_ENTROPY	0
#define SQUARED_ERROR	1

/* Macros for calculating the information */
#define SIMPLE_INFO(output, target)		\
  (log((target)/(output))*(target) 		\
   + log((1.0 - (target))/(1.0 - (output)))*(1.0 - (target)))

#define INFO(output, target)			\
  ((target) == 0.0 ? -log(1.0 - (output))	\
   : (target) == 1.0 ? -log((output)) : SIMPLE_INFO(output, target))

#define INFO_DERIV(output, target)				\
  ((target) == 0.0 ? 1.0 / (1.0 - (output)) :			\
   ((target) == 1.0 ? -1.0 / (output) :				\
    ((1.0-(target))/(1.0 - (output))) - ((target) / (output))))

#endif				/* UTILITIES_H */
