
/**********************************************************************
 * $Id: Selection.c,v 1.18 92/04/20 16:17:19 drew Exp $
 **********************************************************************/

/**********************************************************************
 *	Copyright 1990,1991,1992 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/


/* NOTE: THIS IS NOT A WIDGET!  Rather, this is an interface to a widget.
   It implements policy, and gives a (hopefully) easier-to-use interface
   than just directly making your own form. */

#include <stdio.h>

#include <X11/Xlib.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Cardinals.h>

#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>	
#include <X11/Xaw/List.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Viewport.h>

#include "SelectionP.h"

/*
 * After we have set the list in the listWidget we set the
 * list to a magic value.  So that when a SetValues request is made
 * on the list we will notice it, and reset it.
 */
#define MAGIC_NUMBER ((String *) 3)

static XtResource resources[] = {
  { XtNcallback,     XtCCallback,      XtRCallback,  sizeof(XtCallbackList),
      XtOffset(SelectionWidget, selection.callback),     XtRCallback, NULL},
  {XtNlistLabel,     XtCListLabel,     XtRString,    sizeof(String),
     XtOffset(SelectionWidget, selection.listLabel),     XtRString,   NULL},
  {XtNlist,          XtCList,          XtRPointer,   sizeof(String *),
     XtOffset(SelectionWidget, selection.list),          XtRString,   NULL},
  {XtNnumberStrings, XtCNumberStrings, XtRInt,       sizeof(int),
     XtOffset(SelectionWidget, selection.numberStrings), XtRString,   "0"},
  {XtNlistHeight,    XtCListHeight,    XtRDimension, sizeof(Dimension),
     XtOffset(SelectionWidget, selection.listHeight),    XtRString,   "200"},
  {XtNvalueLabel,    XtCValueLabel,    XtRString,    sizeof(String),
     XtOffset(SelectionWidget, selection.valueLabel),    XtRString,   NULL},
  {XtNvalue,         XtCValue,         XtRString,    sizeof(String),
     XtOffset(SelectionWidget, selection.value),         XtRString,   NULL},
};

static void	Initialize() ;
static void	ConstraintInitialize() ;
static Boolean	SetValues();
static void	ChangeManaged() ;
static void	CreateSelectionValueWidget();
static void	selectionCB() ;

SelectionClassRec selectionClassRec = {
  {				/* core_class fields */
    /* superclass         */    (WidgetClass) &formClassRec,
    /* class_name         */    "Selection",
    /* widget_size        */    sizeof(SelectionRec),
    /* class_initialize   */    NULL,
    /* class_part init    */    NULL,
    /* class_inited       */    FALSE,
    /* initialize         */    Initialize,
    /* initialize_hook    */    NULL,
    /* realize            */    XtInheritRealize,
    /* actions            */    NULL,
    /* num_actions        */    0,
    /* resources          */    resources,
    /* num_resources      */    XtNumber(resources),
    /* xrm_class          */    NULLQUARK,
    /* compress_motion    */    TRUE,
    /* compress_exposure  */    TRUE,
    /* compress_enterleave*/    TRUE,
    /* visible_interest   */    FALSE,
    /* destroy            */    NULL,
    /* resize             */    XtInheritResize,
    /* expose             */    XtInheritExpose,
    /* set_values         */    SetValues,
    /* set_values_hook    */    NULL,
    /* set_values_almost  */    XtInheritSetValuesAlmost,
    /* get_values_hook    */    NULL,
    /* accept_focus       */    NULL,
    /* version            */    XtVersion,
    /* callback_private   */    NULL,
    /* tm_table           */    NULL,
    /* query_geometry     */	XtInheritQueryGeometry,
    /* display_accelerator*/	XtInheritDisplayAccelerator,
    /* extension          */	NULL
    },
  {				/* composite_class fields */
    /* geometry_manager   */   XtInheritGeometryManager,
    /* change_managed     */   ChangeManaged,
    /* insert_child       */   XtInheritInsertChild,
    /* delete_child       */   XtInheritDeleteChild,
    /* extension          */   NULL
    },
  {				/* constraint_class fields */
    /* subresourses       */   NULL,
    /* subresource_count  */   0,
    /* constraint_size    */   sizeof(SelectionConstraintsRec),
    /* initialize         */   ConstraintInitialize,
    /* destroy            */   NULL,
    /* set_values         */   NULL,
    /* extension          */   NULL
    },
  {				/* form_class fields */
    /* layout             */   XtInheritLayout
    },
  {				/* selection_class fields */
    /* extension          */   NULL
    }
};

WidgetClass selectionWidgetClass = (WidgetClass)&selectionClassRec ;

/*********************************************************************
 *	Name:		Initialize
 *	Description:	initialize procedure for the Selection widget
 *	Parameters:
 *	  Widget	request - the requested format
 *	  Widget	new	 - the accepted format
 *	  ArgList	args	 - the args passed at creation
 *	  Cardinal	*numArgs - the number of args
 *	Return Value:
 *	  static void	Initialize - NONE
 *********************************************************************/
static void	Initialize(request, new, args, numArgs)
  Widget	request ;
  Widget	new;
  ArgList	args ;
  Cardinal	*numArgs ;
{
  SelectionWidget sw = (SelectionWidget)new ;

  sw->selection.listLabelW
    = XtVaCreateManagedWidget("listLabel", labelWidgetClass, new,
			      XtNlabel,		sw->selection.listLabel,
			      XtNtop,		XtChainTop,
			      XtNbottom,	XtChainTop,
			      XtNleft,		XtChainLeft,
			      XtNright,		XtChainLeft,
			      XtNvertDistance, 	0,
			      XtNhorizDistance,	0,
			      NULL) ;

  sw->selection.viewportW 
    = XtVaCreateManagedWidget("viewport", viewportWidgetClass, new,
			      XtNheight,   sw->selection.listHeight,
			      XtNtop,      XtChainTop,
			      XtNbottom,   XtChainBottom,
			      XtNleft,     XtChainLeft,
			      XtNright,    XtChainRight,
			      XtNfromVert, sw->selection.listLabelW,
			      NULL) ;

  sw->selection.listW
    = XtVaCreateManagedWidget("list", listWidgetClass, sw->selection.viewportW,
			      XtNlist,		sw->selection.list,
			      XtNnumberStrings,	sw->selection.numberStrings,
			      NULL) ;

  if (sw->selection.numberStrings == 0)
    XtVaGetValues(sw->selection.listW,
		  XtNnumberStrings, &sw->selection.numberStrings, NULL) ;

  XtSelectionHighlight(sw, 0) ;
      
  sw->selection.list = MAGIC_NUMBER ;

  if (sw->selection.value != NULL) {
    CreateSelectionValueWidget(sw);
  } else {
    sw->selection.valueLabelW = NULL ;
    sw->selection.valueW      = NULL ;
  }

  XtAddCallback(sw->selection.listW, XtNcallback, selectionCB, (XtPointer)sw) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		ConstraintInitialize
 *	Description:	ensures that all the command widgets are at
 *			the bottom and below all of the std widgets
 *	Parameters:
 *	  Widget	request - the requested widget
 *	  Widget	new 	- the accepted one
 *	Return Value:
 *	  static void	ConstraintInitialize - NONE
 *********************************************************************/
static void	ConstraintInitialize(request, new)
  Widget	request ;
  Widget	new ;
{
  SelectionWidget 	sw = (SelectionWidget)new->core.parent ;
  SelectionConstraints 	constraint = 
    (SelectionConstraints)((Widget)new)->core.constraints ;

  /*  if not a button, then just use the defaults */
  if (!XtIsSubclass(new, commandWidgetClass))
    return ;

  constraint->form.top       = XtChainBottom ;
  constraint->form.bottom    = XtChainBottom ;
  constraint->form.left      = XtChainLeft ;
  constraint->form.right     = XtChainLeft ;
  if (sw->selection.valueW == NULL) 
    constraint->form.vert_base = sw->selection.viewportW;
  else
    constraint->form.vert_base = sw->selection.valueW;

  if (sw->composite.num_children > 1) {
    WidgetList	children = sw->composite.children;
    Widget	*childP;

    for (childP = children + sw->composite.num_children - 1 ;
	 childP >= children ; childP-- ) {
      if (   *childP == sw->selection.listLabelW
	  || *childP == sw->selection.viewportW
	  || *childP == sw->selection.valueLabelW 
	  || *childP == sw->selection.valueW)
	break;
      if (XtIsManaged(*childP) && XtIsSubclass(*childP, commandWidgetClass)) {
	constraint->form.horiz_base = *childP;
	break;
      }
    }
  }
}
/********************************************************************/


/*********************************************************************
 *	Name:		ChangeManaged
 *	Description:	makes sure that the viewport and value widget
 *			are the same width and at least as wide as all
 *			the command widgets
 *	Parameters:
 *	  Widget	genericWidget - the widget
 *	Return Value:
 *	  static void	ChangeManaged - NONE
 *********************************************************************/
#define MIN_WIDTH	100
#define SCROLLBAR_WIDTH	14
#define SUPER_CLASS_CHANGE_MANAGED(classRec)	\
  (((CompositeClassRec *)(classRec).core_class.superclass)->composite_class.change_managed)
/********************************************************************/
static void	ChangeManaged(genericWidget)
  Widget	genericWidget ;
{
  SelectionWidget 	widget = (SelectionWidget)genericWidget ;

  Widget	viewport = widget->selection.viewportW ;
  Widget	list     = widget->selection.listW ;
  Dimension	listWidth, buttonsWidth, minWidth ;

  if (viewport == NULL || !XtIsManaged(viewport)) {
    (*SUPER_CLASS_CHANGE_MANAGED(selectionClassRec)) ((Widget)widget);
    return ;
  }

  buttonsWidth = 0 ;
  if (widget->composite.num_children > 1) {
    WidgetList	children = widget->composite.children;
    Widget	*childP;

    for (childP = children + widget->composite.num_children - 1 ;
	 childP >= children ; childP-- ) {
      if (XtIsManaged(*childP) && XtIsSubclass(*childP, commandWidgetClass)) {
	buttonsWidth
	  += (*childP)->core.width
	    + ((FormConstraints)(*childP)->core.constraints)->form.dx ;
      }
    }
  }

  if (list && XtIsManaged(list)) {
    listWidth = list->core.width ;
    if (list->core.height > viewport->core.height)
      listWidth += SCROLLBAR_WIDTH ;
  } else {
    listWidth = 0 ;
  }

  minWidth = listWidth > buttonsWidth ? listWidth : buttonsWidth ;
  if (minWidth < MIN_WIDTH)
    minWidth = MIN_WIDTH ;

  XtVaSetValues(viewport,
		XtNwidth, minWidth, NULL) ;
  if (widget->selection.valueW != NULL)
    XtVaSetValues(widget->selection.valueW,
		  XtNwidth, minWidth, NULL) ;

  (*SUPER_CLASS_CHANGE_MANAGED(selectionClassRec)) ((Widget)widget);
}
/********************************************************************/


/*********************************************************************
 *	Name:		SetValues
 *	Description:	set values proc
 *	Parameters:
 *	  Widget	current - 
 *	  Widget	request - 
 *	  Widget	new - 
 *	  ArgList 	in_args - 
 *	  Cardinal 	*numArgs - 
 *	Return Value:
 *	  static Boolean	SetValues - FALSE (never need redisplay)
 *********************************************************************/
static Boolean	SetValues(current, request, new, in_args, numArgs)
  Widget	current ;
  Widget	request ;
  Widget	new;
  ArgList 	in_args;
  Cardinal 	*numArgs;
{
  SelectionWidget w = (SelectionWidget)new;
  SelectionWidget old = (SelectionWidget)current;

  if (w->selection.listLabel != old->selection.listLabel)
    XtVaSetValues(w->selection.listLabelW,
		  XtNlabel, w->selection.listLabel, NULL) ;

  if (w->selection.listHeight != old->selection.listHeight)
    XtVaSetValues(w->selection.viewportW,
		  XtNheight, w->selection.listHeight, NULL) ;

  if (w->selection.list != old->selection.list
      || w->selection.numberStrings != old->selection.numberStrings ) {
    XtUnmanageChild(w->selection.listW) ;

    if (w->selection.list != MAGIC_NUMBER)
      XawListChange(w->selection.listW, w->selection.list, 0, 0, True) ;
    else
      XtVaSetValues(w->selection.listW,
		    XtNnumberStrings, w->selection.numberStrings, NULL) ;

    XtVaGetValues(w->selection.listW,
		  XtNnumberStrings, &w->selection.numberStrings, NULL) ;

    XtManageChild(w->selection.listW) ;
    XtSelectionHighlight(w, 0) ;

    if (w->selection.valueW != NULL)
      XtSelectionSetValueString((Widget)w, "") ;

    w->selection.list = MAGIC_NUMBER ;
  }

  if (w->selection.valueLabelW != NULL
      && w->selection.valueLabel != old->selection.valueLabel)
      XtVaSetValues(w->selection.valueLabelW,
		    XtNlabel, w->selection.valueLabel, NULL) ;

  if (w->selection.value != old->selection.value) {
    if (w->selection.value == NULL) {
      /* only get here if it wasn't NULL before. */
      XtDestroyWidget(old->selection.valueLabelW);
      XtDestroyWidget(old->selection.valueW);
      w->selection.valueLabelW = NULL ;
      w->selection.valueW      = NULL ;
    } else if (old->selection.value == NULL) { 
      /* create a new value widget. */
      w->core.width  = old->core.width;
      w->core.height = old->core.height;
      CreateSelectionValueWidget( (Widget) w);
    } else {			
      /* Widget ok, just change string. */
      XtSelectionSetValueString((Widget)w, w->selection.value) ;
      w->selection.value = w->selection.privateValue ;
    }
  }
  return False;
}
/********************************************************************/


/*********************************************************************
 *	Name:		CreateSelectionValueWidget
 *	Description:	Creates the selection widget's value widget.
 *			Must be called only when w->selection.value
 *			is non-nil.
 *	Parameters:
 *	  Widget	w - the selection widget.
 *	Return Value:
 *	  static void	CreateSelectionValueWidget - NONE
 *********************************************************************/
static void	CreateSelectionValueWidget(w)
  Widget	w ;
{
  SelectionWidget sw = (SelectionWidget) w;    
  int		  maxLength = XtNumber(sw->selection.privateValue) ;

  sw->selection.valueLabelW 
    = XtVaCreateManagedWidget("valueLabel", labelWidgetClass, w,
			      XtNlabel,	 	sw->selection.valueLabel,
			      XtNtop,		XtChainBottom,
			      XtNbottom,	XtChainBottom,
			      XtNleft,	 	XtChainLeft,
			      XtNright,	 	XtChainLeft,
			      XtNfromVert,	sw->selection.viewportW,
			      NULL) ;

  strncpy(sw->selection.privateValue, sw->selection.value, maxLength) ;
  sw->selection.privateValue[maxLength-1] = '\0' ;
  sw->selection.value = sw->selection.privateValue ;

  sw->selection.valueW 
    = XtVaCreateWidget("value", asciiTextWidgetClass, w,
		       XtNtype,			XawAsciiString,
		       XtNeditType,		XawtextEdit,
		       XtNstring,		sw->selection.privateValue,
		       XtNuseStringInPlace,	True,
		       XtNlength,		maxLength,
		       XtNresizable,		True,
		       XtNleft,			XtChainLeft,
		       XtNright,		XtChainRight,
		       XtNtop,      		XtChainBottom,
		       XtNbottom,   		XtChainBottom,
		       XtNfromVert,		sw->selection.valueLabelW,
		       NULL) ;

  /* if the value widget is being added after buttons,
   * then the buttons need new layout constraints.
   */
  if (sw->composite.num_children > 1) {
    WidgetList	children = sw->composite.children;
    Widget	*childP;

    for (childP = children + sw->composite.num_children - 1;
	 childP >= children; childP-- ) {
      if (*childP == sw->selection.valueLabelW 
	  || *childP == sw->selection.listLabelW
	  || *childP == sw->selection.valueW
	  || *childP == sw->selection.viewportW)
	continue;
      if (XtIsManaged(*childP) && XtIsSubclass(*childP, commandWidgetClass)) {
	((SelectionConstraints)(*childP)->core.constraints)->form.vert_base 
	    = sw->selection.valueW;
      }
    }
  }
  XtManageChild(sw->selection.valueW);

  /* Value widget gets the keyboard focus. */
  XtSetKeyboardFocus(w, sw->selection.valueW);
}
/********************************************************************/


/*********************************************************************
 *	Name:		selectionCB
 *	Description:	
 *	Parameters:
 *	  Widget	w - 
 *	  XtPointer	clientData - 
 *	  XtPointer	callData - 
 *	Return Value:
 *	  static void	selectionCB - 
 *********************************************************************/
static void	selectionCB(w, clientData, callData)
  Widget	w ;
  XtPointer	clientData ;
  XtPointer	callData ;
{
  XtCallCallbacks((Widget)clientData, XtNcallback, callData) ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		XtSelectionAddButton
 *	Description:	adds a button to the selection widget
 *	Parameters:
 *	  Widget	 selection - 
 *	  String	 name - 
 *	  XtCallbackProc function - 
 *	  XtPointer	 param - 
 *	Return Value:
 *	  Widget		 XtSelectionAddButton - 
 *********************************************************************/
Widget		 XtSelectionAddButton(selection, name, function, param)
  Widget	 selection ;
  String	 name ;
  XtCallbackProc function ;
  XtPointer	 param ;
{
  Widget button;

  /* Correct Constraints are all set in ConstraintInitialize(). */
  button = XtVaCreateManagedWidget(name, commandWidgetClass, selection, NULL) ;

  if (function != NULL)		/* don't add NULL callback func. */
    XtAddCallback(button, XtNcallback, function, param);

  return button ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		XtSelectionGetValueString
 *	Description:	
 *	Parameters:
 *	  Widget w - 
 *	Return Value:
 *	  char	 *XtSelectionGetValueString - 
 *********************************************************************/
char	 *XtSelectionGetValueString(w)
  Widget w ;
{
  char *value ;
  
  value = ((SelectionWidget)w)->selection.privateValue ;
  return(value);
}
/********************************************************************/


/*********************************************************************
 *	Name:		XtSelectionSetValueString
 *	Description:	
 *	Parameters:
 *	  Widget genericWidget - 
 *	  String newValue - 
 *	Return Value:
 *	  void     XtSelectionSetValueString - 
 *********************************************************************/
void     XtSelectionSetValueString(genericWidget, newValue)
  Widget genericWidget ;
  String newValue ;
{
  SelectionWidget widget = (SelectionWidget)genericWidget ;
  int		  maxLength = XtNumber(widget->selection.privateValue) ;

  strncpy(widget->selection.privateValue, newValue, maxLength) ;
  widget->selection.privateValue[maxLength-1] = '\0' ;

  if (widget->selection.valueW) {
    XtVaSetValues(widget->selection.valueW,
		  XtNstring, widget->selection.privateValue, NULL) ;

    XawTextDisplay(widget->selection.valueW) ;
  }
}
/********************************************************************/


/*********************************************************************
 *	Name:		XtSelectionGetSelected
 *	Description:	
 *	Parameters:
 *	  Widget		genericWidget - 
 *	Return Value:
 *	  XawListReturnStruct	*XtSelectionGetSelected - 
 *********************************************************************/
XawListReturnStruct	*XtSelectionGetSelected(genericWidget)
  Widget		genericWidget ;
{
  SelectionWidget	widget = (SelectionWidget)genericWidget ;
  
  if (widget->selection.listW)
    return XawListShowCurrent(widget->selection.listW) ;
  else
    return NULL ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		XtSelectionHighlight
 *	Description:	highlights a given item in the selection list
 *	Parameters:
 *	  Widget	genericWidget - the selection widget
 *	  int		item - the item index
 *	Return Value:
 *	  void		XtSelectionHighlight - NONE
 *********************************************************************/
void		XtSelectionHighlight(genericWidget, item)
  Widget	genericWidget ;
  int		item ;
{
  SelectionWidget	widget = (SelectionWidget)genericWidget ;

  if (widget->selection.listW && item < widget->selection.numberStrings)
    XawListHighlight(widget->selection.listW, item) ;
}
/********************************************************************/
