
/**********************************************************************
 * $Id: bp.c,v 1.9 91/09/26 11:05:14 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	  Copyright 1990,1991 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <stdio.h>
#include <math.h>
#include <values.h>
#include <nan.h>
#include <commands.h>
#include <xerion/minimize.h>
#include <xerion/example.h>
#include <xerion/utilsDCP.h>
#include "barak.h"

struct TRACE    inRelaxation ;
struct TRACE	forwardEachTime ;
struct TRACE    afterExample ;

static Proc	updateNetActivity      ARGS((Net	net)) ;
static void	calculateNetErrorDeriv ARGS((Net net, ExampleSet exampleSet)) ;
static void	calculateNetError      ARGS((Net net, ExampleSet exampleSet)) ;

static void	initNet   		ARGS((Net	 net)) ;
static void	deinitNet 		ARGS((Net	 net)) ;
static void	initUnit    		ARGS((Unit	 unit)) ;
static void	deinitUnit  		ARGS((Unit	 unit)) ;
static void	initLink    		ARGS((Link	 link)) ;
static void	deinitLink  		ARGS((Link	 link)) ;
static void	initExampleSet    	ARGS((ExampleSet exampleSet)) ;
static void	deinitExampleSet	ARGS((ExampleSet exampleSet)) ;

/**********************************************************************
 * 				Initialization
 **********************************************************************/
int main (argc, argv)
  int	argc ;
  char	**argv ;
{
  /* Insert any private initialization routines here */
  setCreateNetHook(initNet) ;
  setDestroyNetHook(deinitNet) ;
  setCreateUnitHook(initUnit) ;
  setDestroyUnitHook(deinitUnit) ;
  setCreateLinkHook(initLink) ;
  setDestroyLinkHook(deinitLink) ;
  setCreateExampleSetHook(initExampleSet) ;
  setDestroyExampleSetHook(deinitExampleSet) ;

  /* Perform initialization of the simulator */
  IStandardInit(&argc, argv);

  /* Enter loop that reads commands and handles graphics */
  ICommandLoop(stdin, stdout, NULL);

  return 0 ;
}
/***********************************************************************/
static void	initNet (net)
  Net	net ;
{
  net->calculateErrorDerivProc = calculateNetErrorDeriv ;
  net->calculateErrorProc      = calculateNetError ;
  net->activityUpdateProc      = updateNetActivity ;

  net->extension = (NetExtension)calloc(1, sizeof(NetExtensionRec)) ;

  Mcompiled(net)           	= FALSE ;
  MteacherForcing(net)		= FALSE ;
  MsampleExamples(net)		= FALSE ;
  MzeroErrorRadius(net)   	= 0.1 ;
  McorrectTolerance(net) 	= 0.0 ;
  MinitialUnitOutput(net)	= 0.2 ;
  MinputNoise(net)		= 0.0 ;
  MinputStrength(net)		= 1.0 ;
  MinputClamping(net) 		= 0 ;
  MerrorRamp(net)	   	= 1.0 ;
  MnumSlices(net)	   	= 0 ;
  MgraceTimePerSlice(net)	= 0.0 ;
  MminTimePerSlice(net)		= 0.0 ;
  MmaxTimePerSlice(net)		= 1.0 ;
  MmaxTicks(net)		= 0 ;
  Mdt(net)			= 1.0 ;
  MoutputChangeTolerance(net)	= 0.0 ;
}
/**********************************************************************/
static void	deinitNet (net)
  Net	net ;
{
  if (net->extension != NULL)
    free(net->extension) ;
}
/***********************************************************************/
static void	initUnit(unit)
  Unit	unit ;
{
  unit->extension = (UnitExtension)calloc(1, sizeof(UnitExtensionRec)) ;
  Mlesioned(unit)        = FALSE ;
}

static void	deinitUnit(unit)
  Unit	unit ;
{
  if (unit->extension != NULL)
    free(unit->extension) ;
}
/***********************************************************************/
static void	initLink(link)
  Link	link ;
{
  link->extension = (LinkExtension)calloc(1, sizeof(LinkExtensionRec)) ;
  Mlesioned(link) = FALSE ;
}

static void	deinitLink(link)
  Link	link ;
{
  if (link->extension != NULL)
    free(link->extension) ;
}
/***********************************************************************/
static void	initExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  exampleSet->extension = 
    (ExampleExtension)calloc(1, sizeof(ExampleExtensionRec)) ;
}

static void	deinitExampleSet(exampleSet)
  ExampleSet  exampleSet ;
{
  if (exampleSet->extension != NULL)
    free(exampleSet->extension) ;
}

void	*getExampleSetData(exampleSet)
  ExampleSet  exampleSet ;
{
  return exampleSet->extension->data ;
}

void	putExampleSetData(exampleSet, data)
  ExampleSet  exampleSet ;
  void	*data ;
{
 exampleSet->extension->data = data ;
}
/**********************************************************************
 * 				Utilities
 **********************************************************************/
static void	clampInput(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  Net	net = unit->net ;
  Mask	clamping = MinputClamping(net) ;
  int	s = (int)slice ;
  Real	y ;
  Real	in_y   = Mi_slice(unit)[s] ;
  Real	init_y = MinitialUnitOutput(net) ;

  Mext(unit) = 0.0 ;
  if (isNaN(in_y)) {
    Mclamped(unit) = FALSE ;
    y = init_y ;
  } else {
    y = init_y + MinputStrength(net)*(in_y - init_y) + gaussianNoise(MinputNoise(net)) ;
    y = CLIP(y, 0.0, 1.0) ;
    if (clamping&SOFT) {
      Mext(unit) = sigmoidInv(y) ;
      if (! (clamping&ONESHOT)) y = init_y ;
    }
    Mclamped(unit) = !clamping ;
  }
  My(unit) = My_slice(unit)[s] = y ;
}
/**********************************************************************/
static void	setTarget(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  Md(unit) = Md_slice(unit)[(int)slice] ;
}
/**********************************************************************/
static void	forceTarget(unit, tolerance)
  Unit		unit ;
  void		*tolerance ;
{
  Real	d 	= Md(unit) ;
  Real	tol	= *((Real *)tolerance) ;

  if (!isNaN(d) && fabs(My(unit) - d) > tol) 
    My(unit) = fabs(d - tol) ;
}
/**********************************************************************/
static void	saveOutputAt(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  My_t(unit)[(int)tick] = My(unit) ;
}
/**********************************************************************/
static void	saveTargetAt(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  Md_t(unit)[(int)tick] = Md(unit) ;
}
/**********************************************************************/
static void	saveSliceOutput(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  My_slice(unit)[(int)slice] = My(unit) ;
}
/**********************************************************************/
static void	restoreSliceOutput(unit, slice)
  Unit		unit ;
  void		*slice ;
{
  My(unit) = My_slice(unit)[(int)slice] ;
}
/**********************************************************************/
static void	zeroInput(unit, data)
  Unit		unit ;
  void		*data ;
{
  Mx(unit) = 0.0 ;
}
/**********************************************************************/
static void	zeroDerivChange(unit, data)
  Unit		unit ;
  void		*data ;
{
  MdEdydt(unit) = 0.0 ;
}
/**********************************************************************/
static void	addExternalInput(unit, data)
  Unit		unit ;
  void		*data ;
{
  Mx(unit) += Mext(unit) ;
}
/**********************************************************************/
static void	zeroDeriv(unit, data)
  Unit		unit ;
  void		*data ;
{
  MdEdy(unit) = 0.0 ;
}
/**********************************************************************/
static void	printOutput(unit, data)
  Unit		unit ;
  void		*data ;
{
/*
  fprintf(dout,"%3d", (int)(100.0*My(unit))) ;
*/
  fprintf(dout,"%.3f ", My(unit)) ;
}
/**********************************************************************
 * 				Error
 **********************************************************************/
static Real	unitError(unit)
  const Unit	unit ;
{
  Net	net = unit->net ;
  Real	error, output = My(unit), target = Md(unit) ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    output = CLIP(output, MIN_OUTPUT, MAX_OUTPUT) ;
    error = INFO(output, target) ; 
    break ;
  case SQUARED_ERROR:
    error = (output - target) * (output - target) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      
  error *= MerrorRamp(net) ;
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  return error ;
}
/**********************************************************************/
static Real	unitErrorDeriv(unit, tick)
  const Unit	unit ;
  int		tick ;
{
  Net	net = unit->net ;
  Real	error, output = My_t(unit)[tick], target = Md_t(unit)[tick] ;

  if (isNaN(target) || fabs(output-target) <= MzeroErrorRadius(net))
    return 0.0 ;

  switch (MerrorFunction(net)) {
  case CROSS_ENTROPY:
    output = CLIP(output, MIN_OUTPUT, MAX_OUTPUT) ;
    error = INFO_DERIV(output, target) ; 
    break ;
  case SQUARED_ERROR:
    error = 2.0 * (output - target) ;
    break ;
  default:
    IErrorAbort("Unknown error function: %d", MerrorFunction(net)) ;
  }      
  error *= MerrorRamp(net) ;
  if (!MsampleExamples(net)) error *= MexampleFrequency(net) ;
  return error ;
}
/**********************************************************************/
static void	incrementError(unit, data)
  Unit		unit ;
  void		*data ;
{
  *(Real *)data += unitError(unit) ;
}

Real		groupError(group)
  Group group ;
{
  Real	error = 0 ;
  MerrorRamp(group->net) = 1.0 ;
  groupForAllUnits(group, incrementError, &error) ;
  return error ;
}
/**********************************************************************
 * 				Forward
 **********************************************************************/
static	void	netAccumulateInput(net)
  Net		net ;
{
  int	idx; 
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Unit	*pre, *post ;
  Real	*wt ;

  netForAllUnits(net, ALL, zeroInput, NULL) ;

  if (MinputClamping(net)&SOFT)
    netForAllUnits(net, INPUT, addExternalInput, NULL) ;

  for (idx=0, pre=MpreUnits(net), post=MpostUnits(net), wt=Mwts(net) ;
       idx < numCompiledLinks ; 
       idx++) 
    Mx(*post++) += My(*pre++) * (*wt++) ;
}
/**********************************************************************/
static void	computeOutput(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  Net	net 	= unit->net ;
  Real	x, sigx ;

  if (Mclamped(unit)) return ;
  x			   = Mx(unit) + gaussianNoise(MunitNoise(net));
  sigx			   = sigmoid(x) ;
  My(unit)     		  += Mdt(net) * (sigx - My(unit)) ;
  Mdydx_t(unit)[(int)tick] = sigmoidDeriv(sigx) ;
}
/**********************************************************************/
static void	computeOutputAndChange(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  Net	net 	= unit->net ;
  Real	x, sigx ;

  if (Mclamped(unit)) return ;
  x			   = Mx(unit) + gaussianNoise(MunitNoise(net));
  sigx			   = sigmoid(x) ;

  if (unit->group->type&OUTPUT && fabs(sigx - My(unit)) > MmaxOutputChange(net)) 
    MmaxOutputChange(net) = fabs(sigx - My(unit)) ;

  My(unit)     		  += Mdt(net) * (sigx - My(unit)) ;
  Mdydx_t(unit)[(int)tick] = sigmoidDeriv(sigx) ;
}
/**********************************************************************/
static	void	netForward (net, tick, training)
  Net	net ;
  int	tick ;
  Boolean training ;
{
  int	t = tick % MmaxTicks(net) ;
  
  netAccumulateInput (net) ;
  if (training)
    netForAllOtherUnits(net, BIAS,   computeOutput,  (void *)t) ;
  else {
    MmaxOutputChange(net) = 0.0 ;
    netForAllOtherUnits(net, BIAS,   computeOutputAndChange,  (void *)t) ;
  }
  netForAllUnits     (net, ALL,    saveOutputAt,   (void *)t) ;
  netForAllUnits     (net, OUTPUT, saveTargetAt,   (void *)t) ;
  netForAllUnits     (net, OUTPUT, incrementError, (void *)&(net->error)) ;
  if (!Mrunning(net)) {
    int	idx, delayCount = MdelayCount(net) ;
    netForAllOtherUnits(net, BIAS, saveSliceOutput, (void *)Mslice(net)) ;
    IDoTrace(&inRelaxation) ;
    for (idx = 0 ; idx < delayCount*1000 ; idx++) ;
  }
}
/**********************************************************************
 * 				Backward
 **********************************************************************/
static void	injectError(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  MdEdy(unit) += Mdt(unit->net) * unitErrorDeriv(unit, (int)tick) ;
}
/**********************************************************************/
static void	invertError(unit, tick)
  Unit		unit ;
  void		*tick ;
{
  MdEdx(unit) = MdEdy(unit) * Mdydx_t(unit)[(int)tick] ;
}
/**********************************************************************/
static	void	netComputeDerivsChange(net, lastTick)
  Net	net ;
  int	lastTick ;
{
  int	idx; 
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Unit	*pre, *post ;
  Real 	*wt, *dv ;

  netForAllUnits(net, ALL, zeroDerivChange, NULL) ;

  /* factor of dt has been pulled out and applied by netAdjustWeightDerivs */
  for (idx=0, pre=MpreUnits(net), post=MpostUnits(net), wt=Mwts(net), dv=Mdvs(net) ;
       idx < numCompiledLinks ; 
       idx++) {
    MdEdydt(*pre) += MdEdx(*post)   * (*wt++) ;
    (*dv++)       += MdEdx(*post++) * My_t(*pre++)[lastTick] ;
  }
}
/**********************************************************************/
static void	computeDeriv(unit, data)
  Unit		unit ;
  void		*data ;
{
  MdEdy(unit) += Mdt(unit->net) * (MdEdydt(unit) - MdEdy(unit)) ;
}
/***********************************************************************/
static void	netBackward (net, tick, inject)
  Net		net ;
  int		tick ;
  Boolean	inject ;
{
  int	thisTick = tick     % MmaxTicks(net) ;
  int	lastTick = (tick-1) % MmaxTicks(net) ;

  if (inject)
    netForAllUnits      (net, OUTPUT, injectError,  (void *)thisTick) ;
  netForAllOtherUnits   (net, BIAS,   invertError,  (void *)thisTick) ;
  netComputeDerivsChange(net,                       (void *)lastTick) ;
  netForAllOtherUnits   (net, BIAS,   computeDeriv, NULL) ;
}
/***********************************************************************
 * 				Iterate
 ***********************************************************************/
static void	initOutput(unit, tick) 
  Unit		unit ;
  void		*tick ;
{
  My(unit) = My_t(unit)[(int)tick] = MinitialUnitOutput(unit->net) ;
}
/***********************************************************************/
static	void	netPresentExample(net, tick)
  Net	net ;
  int	tick ;
{
  int t = (tick-1)%MmaxTicks(net) ;
  int s = Mslice(net) ;

  netForAllUnits(net, INPUT,  clampInput,   (void *)s) ;
  netForAllUnits(net, INPUT,  saveOutputAt, (void *)t) ;
  netForAllUnits(net, OUTPUT, setTarget,    (void *)s) ;
}
/**********************************************************************/
static void	netIterate(net, training)
  Net		net ;
  Boolean	training ;
{
  Boolean	forcing  = (training && MteacherForcing(net)) ;
  int	forwTick, backTick, outputPos, prevOutputPos, numSlicesCorrect=0  ;
  Real	backTime, timeWindow, bpWindow, dt = Mdt(net) ;

  MnumSlices(net) = lastTargetSlice(net) ;
  if (training) {
    netCompileTicks(net) ;
    timeWindow = (MtimeWindow(net)>0 ? MtimeWindow(net) :
		  MnumSlices(net) * MmaxTimePerSlice(net)) ;
    bpWindow   = (MbpWindow(net)>0 ? MbpWindow(net) : timeWindow) ;
  }
  netForAllOtherUnits(net, INPUT|BIAS, initOutput, 0) ;
  netForAllUnits(net, BIAS, saveOutputAt, 0) ;
  /*
   * For each input slice...
   */
  for (Mslice(net) = 0, forwTick = 1, prevOutputPos = NULL_POS ;
       Mslice(net) < MnumSlices(net) ; Mslice(net)++) {
    Real	timeThisSlice = dt ;
    Real	timeSinceBP = dt ;
    int		ticksThisSlice ;
    int		maxTicksPerSlice = rint(MmaxTimePerSlice(net)/dt) ;
    Boolean	sliceDone = FALSE ;
    Boolean	correct ;

    netPresentExample(net, forwTick) ;
    /*
     * For each time tick...
     */
    if (Mdebug(net)) fprintf(dout, "\n") ;
    for (ticksThisSlice = 1 ; 
	 ticksThisSlice <= maxTicksPerSlice && !sliceDone ; 
	 ticksThisSlice++, forwTick++, timeThisSlice+=dt, timeSinceBP+=dt) {

      MerrorRamp(net) = ERROR_RAMP(net, timeThisSlice) ;
      /*
       * Run forward pass
       */
      netForward(net, forwTick, training) ;
      correct = outputCorrect(net, McorrectTolerance(net), training) ;

      if (Mdebug(net)) {
	fprintf(dout, "%2d %3.1f %5f %1d ",
		ticksThisSlice, timeThisSlice, net->error, correct) ;
	netForAllUnits(net, OUTPUT, printOutput, NULL) ;
	fprintf(dout, "\n") ;
      }

      sliceDone = ((timeThisSlice >= MminTimePerSlice(net)) &&
		   (ticksThisSlice == maxTicksPerSlice || correct ||
		    (!training && MmaxOutputChange(net)<=MoutputChangeTolerance(net)))) ;
      if (sliceDone) numSlicesCorrect += correct ;
      /*
       * Run backward pass
       */
      if (training && 
	  (timeSinceBP >= bpWindow || (IS_LAST_SLICE(net) && sliceDone))) {
	timeSinceBP = 0 ;
	netForAllUnits(net, ALL, zeroDeriv, NULL) ;
	for (backTime = 0,            backTick = forwTick ;
	     backTime < timeWindow && backTick > 0 ;
	     backTime += dt,          backTick--) {
	  MerrorRamp(net) = ERROR_RAMP(net, timeThisSlice - backTime) ;
	  netBackward(net, backTick, backTime < bpWindow) ;
	}
      }
    }
    MaveTime(net) += timeThisSlice ;

    netForAllOtherUnits(net, BIAS, saveSliceOutput, (void *)Mslice(net)) ;
    if (forcing) netForAllUnits(net, OUTPUT, forceTarget, (void *)&McorrectTolerance(net)) ;
    IDoTrace(&forwardEachTime) ;
  } /* slice */
}
/***********************************************************************
 * 			Calculate Error Derivs
 ***********************************************************************/
static	Boolean	getExample(net, exampleSet, next) 
  Net		net ;
  ExampleSet	exampleSet ;
  int		next ;
{
  if (MbatchSize(net) < exampleSet->numExamples) {
    /* select random example according to it's probability */
    do {
      MexampleNum(net) = randomInteger(exampleSet->numExamples);
      MexampleFrequency(net) = exampleSet->example[MexampleNum(net)].frequency ;
    } while (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net))) ;
  } else {
    MexampleNum(net) = next ;
    MexampleFrequency(net) = exampleSet->example[next].frequency ;
    if (MsampleExamples(net) && !WITH_PROBABILITY(MexampleFrequency(net)))
      return FALSE ;       /* skip this example */
  }
  MgetNumber(exampleSet, MexampleNum(net)) ;
  return TRUE ;
}
/**********************************************************************/
static	void	netAdjustWeightDerivs(net)
  Net	net ;
{
  int	idx; 
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real	dt = Mdt(net) ;
  Real 	*dv;

  for (idx = 0, dv=Mdvs(net) ; idx < numCompiledLinks ; idx++) 
    (*dv++) *= dt ;
}

/**********************************************************************/
static void	unitDecayWts(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;
  Real	weightCost	= MweightCost(net) ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link link    = incoming[idx] ;
    Real wtCost  = link->weight * weightCost ;
    net->cost   += wtCost * link->weight ;
    link->deriv += 2.0 * wtCost ;
  }
}
/**********************************************************************/
static void	unitMeltWts(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;
  Real	weightCost	= MweightCost(net) ;
  Real	Wo	   	= MstandardWeight(net) ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link link    = incoming[idx] ;
    Real wt	 = link->weight ;
    Real ratio   = (wt * wt) / (Wo * Wo) ;
    net->cost   += weightCost * ratio / (1.0 + ratio) ;
    link->deriv += weightCost * 2.0 * wt / ((1.0 + ratio) * (1.0 + ratio)) ;
  }
}
/**********************************************************************/
static 	void	netUpdateCost(net)
  Net	net ;
{
  if (MweightCost(net) == 0.0) return ;
  if (MstandardWeight(net) == 0.0) 
    netForAllUnits(net, ALL, unitDecayWts, NULL) ;
  else
    netForAllUnits(net, ALL, unitMeltWts, NULL) ;
}
/**********************************************************************/
static void	zeroLinkDeriv(link)
  Link	link ;
{
  link->deriv = 0.0 ;
}
/**********************************************************************/
static void	unitFreezeWts(unit, data)
  Unit		unit ;
  void		*data ;
{
  unitForAllLinks(unit, FROZEN, zeroLinkDeriv, NULL) ;
}
/**********************************************************************/
static 	void	netFreezeWeights(net)
  Net	net ;
{
  netForAllUnits(net, ALL, unitFreezeWts, NULL) ;
}
/***********************************************************************/
static	void	clearStats(net)
  Net	net ;
{
  if (net->mz->repetitionCount==0 || net->mz->nIterations==0 ||
      (net->mz->nIterations-1)%net->mz->repetitionCount==0) {
    net->error 		   	= 0 ;
    net->cost 		   	= 0 ;
    MaveTime(net)	   	= 0 ;
    MaveFixations(net) 	   	= 0 ;
    MaveSlicesCorrect(net)	= 0 ;
    MaveExamplesCorrect(net)    = 0 ;
  }
}
/***********************************************************************/
static	void	normStats(net)
  Net	net ;
{
  Real	scale = 1.0 ;

  if (net->mz->repetitionCount==0 || 
      net->mz->nIterations%net->mz->repetitionCount==0) {
    if (net->mz->nIterations && net->mz->repetitionCount)
      scale = (Real)net->mz->repetitionCount ;
    net->error		     	*= Mdt(net) / scale ;
    net->cost		     	/= scale ;
    scale *= net->batchSize ;
    MaveTime(net)	     	/= scale ;
    MaveFixations(net)	     	/= scale ;
    MaveSlicesCorrect(net) 	/= scale ;
    MaveExamplesCorrect(net)    /= scale ;
  }
}
/**********************************************************************/
static Proc	updateNetActivity (net)
  Net	net ;
{
  Mrunning(net) = FALSE ;
  if (!Mcompiled(net)) netCompile(net) ;
  getCompiledWeights(net) ;
  net->error               	= 0.0 ;
  MexampleFrequency(net)	= 1.0 ;
  MaveTime(net)	 	   	= 0 ;
  MaveFixations(net)       	= 0 ;
  MaveSlicesCorrect(net)	= 0 ;
  MaveExamplesCorrect(net)	= 0 ;

  netIterate(net, FALSE) ;
  net->error *= Mdt(net) ;
  Mslice(net) = MnumSlices(net) - 1 ;
  IDoTrace(&afterExample) ;
}
/***********************************************************************/
static void	calculateError(net, exampleSet, calculateDerivs)
  Net		net ;
  ExampleSet	exampleSet ;
  Boolean	calculateDerivs ;
{
  int 	numExamples ;

  Mrunning(net)  = TRUE ;
  if (!Mcompiled(net)) netCompile(net) ;
  if (calculateDerivs) initEpoch(net, MalphaOnset(net)) ;
  clearStats(net) ;
  getCompiledWeights(net) ;

  for (numExamples = 0 ; numExamples < MbatchSize(net) ; ++numExamples) {
    if (!getExample(net, exampleSet, numExamples)) continue ;
    netIterate(net, calculateDerivs) ;
    if (calculateDerivs) IDoTrace(&afterExample) ;
  }
  if (numExamples <= 0)
    IErrorAbort("calculateError: no examples processed") ;
  if (calculateDerivs) {
    netAdjustWeightDerivs(net) ;
    putCompiledDerivs(net) ;
    netUpdateCost(net) ;
    netFreezeWeights(net) ;
  }
  normStats(net) ;
  Mrunning(net)  = FALSE ;
}
/***********************************************************************/
static void	calculateNetErrorDeriv(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  calculateError(net, exampleSet, TRUE) ;
}
/*********************************************************************/
static void	calculateNetError(net, exampleSet)
  Net		net ;
  ExampleSet	exampleSet ;
{
  calculateError(net, exampleSet, FALSE) ;
}
/**********************************************************************/
