
/**********************************************************************
 * $Id: bp.h,v 1.7 91/09/26 11:05:16 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	   Copyright 1990,1991 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#ifndef ELMAN_H
#define ELMAN_H

/* The following structures must be defined, even if you don't
 * need them in you simulator
 */
typedef struct NetExtensionRec {	/* BIND */
  Boolean	debug ;		/* netParam: */
  Boolean	running ;
  int	delayCount ;		/* netParam: */

  int	alphaOnset ;		/* netParam: */
  Real	weightCost ;		/* netParam: */
  Real	standardWeight ;	/* netParam: */
  Real	zeroErrorRadius ;	/* netParam: */
  Real	correctTolerance ;	/* netParam: */
  int	teacherForcing ;	/* netParam: */
  int	rampError ;		/* netParam: */
  int	errorFunction ;		/* netParam: 0=CE 1=SSE */
  Real	initialUnitOutput ;	/* netParam: */
  Real	unitNoise ;		/* netParam: */
  Real	inputNoise ;		/* netParam: */
  Real	inputStrength ;		/* netParam: */
  int	inputClamping ;		/* netParam: */
  int	exampleNum ;
  Real	exampleFrequency ;
  int	sampleExamples ;	/* netParam: */

  Real	maxOutputChange ;	
  Real	outputChangeTolerance ;	/* netParam: */

  int	refixate ;		/* netParam: */
  Real	aveTime ;	
  Real	aveFixations ;
  Real	aveSlicesCorrect ;	
  Real	aveExamplesCorrect ;	

  Boolean	compiled ;
  int	numCompiledLinks ;
  Unit	*preUnits ;		/* invis: counter: numCompiledLinks */
  Unit	*postUnits ;		/* invis: counter: numCompiledLinks */
  Real*	*weights ;		/* invis: counter: numCompiledLinks */
  Real	*wts ;			/* invis: counter: numCompiledLinks */
  Real*	*derivs ;		/* invis: counter: numCompiledLinks */
  Real	*dvs ;			/* invis: counter: numCompiledLinks */

  int	maxTicks ;		
  int	numSlices ;
  Real	errorRamp ;
  Real	graceTimePerSlice ;	/* netParam: */
  Real	minTimePerSlice ;	/* netParam: */
  Real	maxTimePerSlice ;	/* netParam: */
  Real	timeWindow  ;		/* netParam: */
  Real	bpWindow  ;		/* netParam: */
  Real  dt ;		        /* netParam: */

} NetExtensionRec ;

typedef struct GroupExtensionRec {	/* BIND */
  int	foo ;
} GroupExtensionRec ;

typedef struct UnitExtensionRec {	/* BIND */
  Boolean	lesioned ;
  Boolean	clamped ;
  Real		ext_x ;		/* external input (for input units) */
  Real*		y_t ;		/* output */
  Real*		d_t ;		/* target */
  Real		dydx ;		/* sigmoidDeriv(sigmoid(x)) */
  Real*		dydx_t ;	/* for each tick */
  Real		dEdydt ;	/* change in dEdy over time */
} UnitExtensionRec ;

typedef struct LinkExtensionRec {	/* BIND */
  Boolean	lesioned ;
  Real		lesionedWeight ;
} LinkExtensionRec ;

typedef struct ExampleExtensionRec {	/* BIND */
  void	*data ;
} ExampleExtensionRec ;

#define MbatchSize(net)		((net)->batchSize)
#define	Mslice(net)		((net)->currentTime)

#define Mdebug(net)		((net)->extension->debug)
#define Mrunning(net)		((net)->extension->running)
#define MdelayCount(net)	((net)->extension->delayCount)
#define MzeroErrorRadius(net)	((net)->extension->zeroErrorRadius)
#define McorrectTolerance(net)	((net)->extension->correctTolerance)
#define MweightCost(net)	((net)->extension->weightCost)
#define MstandardWeight(net)	((net)->extension->standardWeight)

#define MinitialUnitOutput(net)	((net)->extension->initialUnitOutput)
#define MunitNoise(net)		((net)->extension->unitNoise)
#define MinputNoise(net)	((net)->extension->inputNoise)
#define MinputStrength(net)	((net)->extension->inputStrength)
#define MinputClamping(net)	((net)->extension->inputClamping)

#define MteacherForcing(net)	((net)->extension->teacherForcing)
#define MrampError(net)		((net)->extension->rampError)
#define MerrorFunction(net)	((net)->extension->errorFunction)
#define MalphaOnset(net)	((net)->extension->alphaOnset)
#define MexampleNum(net)	((net)->extension->exampleNum)
#define MexampleFrequency(net)	((net)->extension->exampleFrequency)
#define MsampleExamples(net) 	((net)->extension->sampleExamples)

#define Mcompiled(net)		((net)->extension->compiled)
#define MpreUnits(net)		((net)->extension->preUnits)
#define MpostUnits(net)		((net)->extension->postUnits)
#define Mwts(net)		((net)->extension->wts)
#define Mweights(net)		((net)->extension->weights)
#define Mderivs(net)		((net)->extension->derivs)
#define Mdvs(net)		((net)->extension->dvs)
#define MnumCompiledLinks(net)	((net)->extension->numCompiledLinks)

#define MmaxOutputChange(net)	    ((net)->extension->maxOutputChange)
#define MoutputChangeTolerance(net) ((net)->extension->outputChangeTolerance)

#define Mrefixate(net)		((net)->extension->refixate)
#define MaveFixations(net)	((net)->extension->aveFixations)
#define MaveSlicesCorrect(net)	((net)->extension->aveSlicesCorrect)
#define MaveExamplesCorrect(net) ((net)->extension->aveExamplesCorrect)
#define MaveTime(net)		((net)->extension->aveTime)

#define MmaxTicks(net)		((net)->extension->maxTicks)
#define MnumSlices(net)		((net)->extension->numSlices)
#define MerrorRamp(net)		((net)->extension->errorRamp)
#define MgraceTimePerSlice(net)	((net)->extension->graceTimePerSlice)
#define MminTimePerSlice(net)	((net)->extension->minTimePerSlice)
#define MmaxTimePerSlice(net)	((net)->extension->maxTimePerSlice)
#define MtimeWindow(net)	((net)->extension->timeWindow)
#define MbpWindow(net)		((net)->extension->bpWindow)
#define Mdt(net)		((net)->extension->dt)

#define Mlesioned(unit_or_link)	((unit_or_link)->extension->lesioned)
#define MlesionedWeight(link)	((link)->extension->lesionedWeight)

#define Mclamped(unit)		((unit)->extension->clamped)

#define Mi_slice(unit)		((unit)->clampingHistory)
#define Mx(unit)		((unit)->totalInput)
#define Mext(unit)		((unit)->extension->ext_x)
#define My(unit)		((unit)->output)
#define My_slice(unit)		((unit)->outputHistory)
#define My_t(unit)		((unit)->extension->y_t)
#define Mdydx(unit)		((unit)->extension->dydx)
#define Mdydx_t(unit)		((unit)->extension->dydx_t)
#define Md(unit)		((unit)->target)
#define Md_slice(unit)		((unit)->targetHistory)
#define Md_t(unit)		((unit)->extension->d_t)
#define MdEdx(unit)		((unit)->inputDeriv)
#define MdEdy(unit)		((unit)->outputDeriv)
#define MdEdydt(unit)		((unit)->extension->dEdydt)

#define NULL_POS	-1
#define IS_LAST_SLICE(net)	(Mslice(net) == MnumSlices(net)-1)

/* types of clamping */
#define	SOFT	((Mask)(1<<0))	/* use external input */
#define	ONESHOT	((Mask)(1<<1))	/* only present unit states once */

#define ERROR_RAMP(net, sliceTime)				\
  ((sliceTime) <= MgraceTimePerSlice(net) ? 			\
   0.0 :							\
   (MrampError(net) ? 						\
    (((sliceTime) - MgraceTimePerSlice(net)) /			\
     (MmaxTimePerSlice(net) - MgraceTimePerSlice(net))) : 	\
    1.0))

extern struct TRACE	afterExample ; 		/* BIND */
extern struct TRACE	forwardEachTime ;	/* BIND */
extern struct TRACE    	inRelaxation ;		/* BIND */

/* compile-barak.c */
extern	void	netCompile		ARGS((Net net)) ;
extern	void	netCompileTicks		ARGS((Net net)) ;
extern	void	getCompiledWeights	ARGS((Net net)) ;
extern	void	putCompiledDerivs	ARGS((Net net)) ;

/* lesion-barak.c */
extern void	lesionUnit		ARGS((Unit unit)) ;
extern void	unlesionUnit		ARGS((Unit unit)) ;
extern Boolean	isLesionedUnit		ARGS((Unit unit)) ;
extern void	lesionLink		ARGS((Link link)) ;
extern void	unlesionLink		ARGS((Link link)) ;
extern Boolean	isLesionedLink		ARGS((Link link)) ;

#endif				/* ELMAN_H */
