#include <commands.h>
#include "barak.h"

/**********************************************************************
 * 				Compilation
 *********************************************************************/
static void	unitCompile(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numLinks	= MnumCompiledLinks(net) ;
  Unit	*preUnits	= MpreUnits(net) ;
  Unit	*postUnits	= MpostUnits(net) ;
  Real*	*weights	= Mweights(net) ;
  Real*	*derivs		= Mderivs(net) ;

  if (Mlesioned(unit)) return ;

  for (idx = 0 ; idx < unit->numIncoming ; ++idx) {
    Link	link  = unit->incomingLink[idx] ;
    if (!Mlesioned(link)) {
      preUnits[numLinks] 	= link->preUnit ;
      postUnits[numLinks] 	= unit ;
      weights[numLinks] 	= &(link->weight) ;
      derivs[numLinks] 		= &(link->deriv) ;
      numLinks++ ;
      if (link->type&BIDIRECTIONAL) {
	preUnits[numLinks] 	= unit ;
	postUnits[numLinks] 	= link->preUnit ;
	weights[numLinks] 	= &(link->weight) ;
	derivs[numLinks]	= &(link->deriv) ;
	numLinks++ ;
      }
    }
  }
  MnumCompiledLinks(net) = numLinks ;
}
/**********************************************************************/
static void	countCompiledLinks(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx, sum = 0 ;
  for (idx = 0 ; idx < unit->numIncoming ; ++idx) {
    Link	link  = unit->incomingLink[idx] ;
    if (!Mlesioned(link)) {
      sum++ ;
      if (link->type&BIDIRECTIONAL) sum++ ;
    }
  }
  MnumCompiledLinks(unit->net) += sum ;
}
/**********************************************************************/
static void	unitCompileTicks(unit, ticks)
  Unit		unit ;
  void		*ticks  ;
{
  int	oldTicks = MmaxTicks(unit->net) ;
  extendList((void *)&My_t(unit),	sizeof(Real), oldTicks, ticks);
  extendList((void *)&Mdydx_t(unit),	sizeof(Real), oldTicks, ticks);
  if (unit->group->type & OUTPUT)
    extendList((void *)&Md_t(unit),	sizeof(Real), oldTicks, ticks);
}
/***********************************************************************/
void	netCompileTicks(net)
  Net		net ;
{
  Real	time = (MtimeWindow(net)>0 ? MtimeWindow(net) :
		MnumSlices(net) * MmaxTimePerSlice(net)) ;
  int	maxTicks = 1 + (time / Mdt(net)) ;

  if (maxTicks > MmaxTicks(net)) {
    netForAllUnits(net, ALL, unitCompileTicks, (void *)maxTicks) ;
    MmaxTicks(net) = maxTicks ;
  }
}
/***********************************************************************/
void	netCompile(net)
  Net		net ;
{
  static int	maxLinks = 0;
  int		numLinks = 0;

  MnumCompiledLinks(net) = 0;
  netForAllUnits(net, ALL, countCompiledLinks, NULL) ;
  numLinks = MnumCompiledLinks(net);

  if (numLinks > maxLinks) {
    extendList((void *)&MpreUnits(net), sizeof(Unit),  maxLinks, numLinks);
    extendList((void *)&MpostUnits(net),sizeof(Unit),  maxLinks, numLinks);
    extendList((void *)&Mweights(net),	sizeof(Real*), maxLinks, numLinks);
    extendList((void *)&Mderivs(net),	sizeof(Real*), maxLinks, numLinks);

    extendList((void *)&Mwts(net), sizeof(Real), maxLinks, numLinks);
    extendList((void *)&Mdvs(net), sizeof(Real), maxLinks, numLinks);

    maxLinks = numLinks ;
  }
  MnumCompiledLinks(net) = 0;
  netForAllUnits(net, ALL, unitCompile, NULL) ;
  netCompileTicks(net) ;
  Mcompiled(net) = TRUE ;
}
/**********************************************************************/
void	getCompiledWeights(net)
  Net net;
{
  int	idx ;
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real	*wt, *dv, **weight, **deriv ;

  for (idx=0, wt=Mwts(net), weight=Mweights(net), dv=Mdvs(net), deriv=Mderivs(net); 
       idx < numCompiledLinks ; 
       idx++) {
    (*wt++) = *(*weight++) ;
    (*dv++) = *(*deriv++) = 0.0 ;
  }
}
/**********************************************************************/
void	putCompiledDerivs(net)
  Net net;
{
  int 	idx;
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real	*dv, **deriv ;

  for (idx = 0, dv=Mdvs(net), deriv=Mderivs(net);
       idx < numCompiledLinks ; 
       idx++) 
    *(*deriv++) += (*dv++) ;
}
/***********************************************************************/
int	command_compile(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Compiles network");
    IHelp
      (IHelpArgs,
       "Compiles network.",
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netCompile(currentNet) ;

  return 1 ;
}
/**********************************************************************/
