#include <commands.h>
#include "dbm.h"
/*********************************************************************
 *	Name:		unitCompile
 *	Description:	put incomingLinks in compiled arrays
 *	Parameters:
 *	  Unit		unit - the unit whose links are to be compiled
 *	  void		*data - UNUSED
 *	Return Value:	NONE
 *********************************************************************/
static void	unitCompile(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx ;
  Net	net 		= unit->net ;
  int	numLinks	= MnumCompiledLinks(net) ;
  Unit	*preUnits	= MpreUnits(net) ;
  Unit	*postUnits	= MpostUnits(net) ;
  Real*	*weights	= Mweights(net) ;
  Real  *wts		= Mwts(net) ;
  Real*	*derivs		= Mderivs(net) ;
  int	numIncoming 	= unit->numIncoming ;
  Link	*incoming   	= unit->incomingLink ;

  for (idx = 0 ; idx < numIncoming ; ++idx) {
    Link	link  = incoming[idx] ;
    if (!Mlesioned(link)) {
      preUnits[numLinks] 	= link->preUnit ;
      postUnits[numLinks] 	= unit ;
      weights[numLinks] 	= &(link->weight) ;
      wts[numLinks]		= link->weight ;
      derivs[numLinks] 		= &(link->deriv) ;
      numLinks++ ;
    }
  }
  MnumCompiledLinks(net) = numLinks ;
}
/**********************************************************************/
static void	countCompiledLinks(unit, data)
  Unit		unit ;
  void		*data ;
{
  MnumCompiledLinks(unit->net) += unit->numIncoming ;
}
/***********************************************************************/
void	netCompile(net)
  Net		net ;
{
  static int	maxLinks = 0;
  int		numLinks = 0;

  MnumCompiledLinks(net) = 0;
  netForAllUnits(net, ALL, countCompiledLinks, NULL) ;
  numLinks = MnumCompiledLinks(net);

  if (numLinks > maxLinks) {
    extendList((void *)&MpreUnits(net), sizeof(Unit),  maxLinks, numLinks);
    extendList((void *)&MpostUnits(net),sizeof(Unit),  maxLinks, numLinks);
    extendList((void *)&Mweights(net),	sizeof(Real*), maxLinks, numLinks);
    extendList((void *)&Mderivs(net),	sizeof(Real*), maxLinks, numLinks);

    extendList((void *)&Mwts(net), 	sizeof(Real),  maxLinks, numLinks);
    extendList((void *)&Mdvs(net), 	sizeof(Real),  maxLinks, numLinks);

    maxLinks = numLinks ;
  }
  MnumCompiledLinks(net) = 0;
  netForAllUnits(net, ALL, unitCompile, NULL) ;
  Mcompiled(net) = TRUE ;
}
/**********************************************************************/
void	getCompiledWeights(net)
  Net net;
{
  int	idx ;
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real	*wt, *dv, **weight ;

  if (!Mcompiled(net)) netCompile(net) ;

  for (idx = 0, wt=Mwts(net), weight=Mweights(net), dv = Mdvs(net) ; 
       idx < numCompiledLinks ; 
       idx++) {
    (*wt++) = *(*weight++) ;
    (*dv++) = 0.0 ;
  }
}
/**********************************************************************/
void	putCompiledDerivs(net)
  Net net;
{
  int 	idx;
  int	numCompiledLinks = MnumCompiledLinks(net) ;
  Real	*dv, **deriv ;

  for (idx = 0, dv=Mdvs(net), deriv=Mderivs(net);
       idx < numCompiledLinks ; 
       idx++) {
    *(*deriv++) = (*dv++) ;
  }
}
/**********************************************************************/
int	command_compile(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Compiles network");
    IHelp
      (IHelpArgs,
       "Compiles network.",
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netCompile(currentNet) ;

  return 1 ;
}
/**********************************************************************/
