#ifndef DBM_H
#define DBM_H

/* The following structures must be defined, even if you don't
 * need them in you simulator
 */
typedef struct NetExtensionRec {	/* BIND */
  Boolean	debug ;		/* netParam: */
  Boolean	running ;
  Real	zeroErrorRadius ;	/* netParam: */
  Real	weightCost ;		/* netParam: */
  Real	standardWeight ;	/* netParam: */
  Real	initialUnitOutput ;	/* netParam: */
  Real	inputNoise ;		/* netParam: */
  Real	unitNoise ;		/* netParam: */
  int	delayCount ;		/* netParam: */
  int	exampleNum ;
  Real	exampleFrequency ;
  int	numExamples ;
  int	alphaOnset ;		/* netParam: */
  int	sampleExamples ;	/* netParam: */

  int	phase ;			/* netParam: */
  Real	tMax ;			/* netParam: */
  Real	tMin ;			/* netParam: */
  Real	tDecay ;		/* netParam: */
  Real	temperature ;
  Real  relaxStepSize ;         /* netParam: */
  Real  relaxTolerance ;        /* netParam: */
  Real  maxDelta ;      
  int   unitFunction ;          /* netParam: */
  int	aveSweepCount ;		
  int	maxSweepCount ;		/* netParam: */
  Real	targetOffset ;		/* netParam: */
  Boolean	annealPositive ;

  Boolean	compiled ;
  int	numCompiledLinks ;
  Unit	*preUnits ;		/* invis: counter: numCompiledLinks */
  Unit	*postUnits ;		/* invis: counter: numCompiledLinks */
  Real*	*weights ;		/* invis: counter: numCompiledLinks */
  Real	*wts ;			/* invis: counter: numCompiledLinks */
  Real*	*derivs ;		/* invis: counter: numCompiledLinks */
  Real	*dvs ;			/* invis: counter: numCompiledLinks */
} NetExtensionRec ;

typedef struct GroupExtensionRec {	/* BIND */
  int	foo ;
} GroupExtensionRec ;

typedef struct UnitExtensionRec {	/* BIND */
  Boolean	clamped ;
  Boolean	lesioned ;
} UnitExtensionRec ;

typedef struct LinkExtensionRec {	/* BIND */
  Boolean	lesioned ;
  Real		lesionedWeight ;
} LinkExtensionRec ;

typedef struct ExampleExtensionRec {	/* BIND */
  void	*data ;
} ExampleExtensionRec ;

/* macros for getting at extension fields */
#define MbatchSize(net)		((net)->batchSize)

#define Mdebug(net)		((net)->extension->debug)
#define Mrunning(net)		((net)->extension->running)
#define MzeroErrorRadius(net)	((net)->extension->zeroErrorRadius)
#define MweightCost(net)	((net)->extension->weightCost)
#define MstandardWeight(net)	((net)->extension->standardWeight)
#define MinitialUnitOutput(net)	((net)->extension->initialUnitOutput)
#define MinputNoise(net)	((net)->extension->inputNoise)
#define MunitNoise(net)		((net)->extension->unitNoise)
#define MdelayCount(net)	((net)->extension->delayCount)
#define MexampleNum(net)	((net)->extension->exampleNum)
#define MexampleFrequency(net)	((net)->extension->exampleFrequency)
#define MnumExamples(net)	((net)->extension->numExamples)
#define MalphaOnset(net)	((net)->extension->alphaOnset)
#define MsampleExamples(net)	((net)->extension->sampleExamples)

#define Mphase(net)		((net)->extension->phase)
#define MtMax(net)		((net)->extension->tMax)
#define MtMin(net)		((net)->extension->tMin)
#define MtDecay(net)		((net)->extension->tDecay)
#define Mtemp(net)		((net)->extension->temperature)
#define MrelaxStepSize(net)	((net)->extension->relaxStepSize)
#define MrelaxTolerance(net)	((net)->extension->relaxTolerance)
#define MmaxDelta(net)		((net)->extension->maxDelta)
#define MunitFunction(net)	((net)->extension->unitFunction)
#define MdelayCount(net)	((net)->extension->delayCount)
#define MaveSweepCount(net)	((net)->extension->aveSweepCount)
#define MmaxSweepCount(net)	((net)->extension->maxSweepCount)
#define MtargetOffset(net)	((net)->extension->targetOffset)
#define MalphaOnset(net)	((net)->extension->alphaOnset)
#define MannealPositive(net)	((net)->extension->annealPositive)

#define Mcompiled(net)		((net)->extension->compiled)
#define MnumCompiledLinks(net)	((net)->extension->numCompiledLinks)
#define MpreUnits(net)		((net)->extension->preUnits)
#define MpostUnits(net)		((net)->extension->postUnits)
#define Mweights(net)		((net)->extension->weights)
#define Mwts(net)		((net)->extension->wts)
#define Mderivs(net)		((net)->extension->derivs)
#define Mdvs(net)		((net)->extension->dvs)
#define Mweights(net)		((net)->extension->weights)

#define Mclamped(unit)		((unit)->extension->clamped)

#define Mlesioned(unit_or_link)	((unit_or_link)->extension->lesioned)
#define MlesionedWeight(link)	((link)->extension->lesionedWeight)

#define CLEAR		0
#define POSITIVE	1
#define NEGATIVE	2

#define sym2asym(x)		(0.5*(x)+0.5)	/* [-1 1] => [ 0 1] */
#define asym2sym(x)		(2.0*(x)-1.0)	/* [ 0 1] => [-1 1] */

#define asym2either(net,x)	(MunitFunction(net) ? asym2sym(x) : (x))
#define either2asym(net,x)	(MunitFunction(net) ? sym2asym(x) : (x))

#define sym2either(net,x)	(MunitFunction(net) ? (x) : sym2asym(x))
#define either2sym(net,x)	(MunitFunction(net) ? (x) : asym2sym(x))

extern struct TRACE     inRelaxation ; 		/* BIND */
extern struct TRACE	afterExample ; 		/* BIND */

/* compile-dbm.c */
extern	void	netCompile		ARGS((Net net)) ;
extern	void	getCompiledWeights	ARGS((Net net)) ;
extern	void	putCompiledDerivs	ARGS((Net net)) ;

/* lesion-dbm.c */
extern void	lesionUnit		ARGS((Unit unit)) ;
extern void	unlesionUnit		ARGS((Unit unit)) ;
extern Boolean	isLesionedUnit		ARGS((Unit unit)) ;
extern void	lesionLink		ARGS((Link link)) ;
extern void	unlesionLink		ARGS((Link link)) ;
extern Boolean	isLesionedLink		ARGS((Link link)) ;

#endif				/* DBM_H */
