#include <commands.h>
#include "elman.h"

/*********************************************************************/
static void	unitCompile(unit, data)
  Unit		unit ;
  void		*data ;
{
  int	idx, cdx, numIncoming = unit->numIncoming ;
  Link	*link ;

  if (Mlesioned(unit)) {
    MnumCompiledIncoming(unit) = 0 ;
    return ;
  }
  if (!MpreOutputs(unit)) {
    MpreOutputs(unit)	   = ((Real **)calloc(numIncoming, sizeof(Real*))) ;
    MpreOutputDerivs(unit) = ((Real **)calloc(numIncoming, sizeof(Real*))) ;
    Mweights(unit)    	   = ((Real *) calloc(numIncoming, sizeof(Real))) ;
    Mderivs(unit)     	   = ((Real *) calloc(numIncoming, sizeof(Real))) ;
  }

  for (idx=0, cdx=0, link=unit->incomingLink ; idx < numIncoming ; idx++) {
    if (!Mlesioned(*link)) {
      MpreOutputs(unit)[cdx] 	  = &((*link)->preUnit->output) ;
      MpreOutputDerivs(unit)[cdx] = &((*link)->preUnit->outputDeriv) ;
      Mweights(unit)[cdx++]	  =   (*link++)->weight ;
    }
  }
  MnumCompiledIncoming(unit) = cdx ; 
}
/***********************************************************************/
void	netCompile(net)
  Net		net ;
{
  netForAllOtherUnits(net, INPUT|BIAS, unitCompile, NULL) ;
  Mcompiled(net) = TRUE ;
}
/**********************************************************************/
void	getCompiledWeights(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	idx ;
  int	numIncoming	= MnumCompiledIncoming(unit) ;
  Link	*link ;
  Real	*wt, *dv;

  for (idx=0, link=unit->incomingLink, wt=Mweights(unit), dv=Mderivs(unit) ;
       idx < numIncoming ; 
       idx++) {
    (*wt++) = (*link++)->weight ;
    (*dv++) = 0.0 ;
  }
}
/**********************************************************************/
void	putCompiledDerivs(unit, data)
  Unit	unit ;
  void	*data ;
{
  int	idx ;
  int	numIncoming = MnumCompiledIncoming(unit) ;
  Link	*link ;
  Real	*dv ;

  for (idx=0, link=unit->incomingLink, dv=Mderivs(unit) ;
       idx < numIncoming ; 
       idx++) 
    (*link++)->deriv  = (*dv++) ;
}
/***********************************************************************
 *	Name:		command_compile
 *	Description:	Compiles network.
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_template - 0 on failure, 1 on success
 ***********************************************************************/
int	command_compile(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{

  IUsage("");
  if (GiveHelp(tokc)) {
    ISynopsis("Compiles network");
    IHelp
      (IHelpArgs,
       "Compiles network.",
       "EXAMPLES",
       "SEE ALSO",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  netCompile(currentNet) ;

  return 1 ;
}
/**********************************************************************/
