
/**********************************************************************
 * $Id: bp.h,v 1.7 91/09/26 11:05:16 drew Exp Locker: drew $
 **********************************************************************/

/**********************************************************************
 *	   Copyright 1990,1991 by The University of Toronto,
 *		       Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute,  and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided  that the above copyright notice  appears in all copies and
 * that both the copyright notice and this permission notice  appear in
 * supporting documentation, and  that  the  name of The University  of
 * Toronto  not  be used  in advertising   or publicity pertaining   to
 * distribution  of   the software   without  specific, written   prior
 * permission.  The  University  of Toronto  makes   no representations
 * about the  suitability  of  this software  for  any purpose.   It is
 * provided "as is" without express or implied warranty.
 *
 * THE  UNIVERSITY OF  TORONTO DISCLAIMS ALL WARRANTIES  WITH REGARD TO
 * THIS SOFTWARE,  INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT  SHALL THE UNIVERSITY  OF TORONTO BE LIABLE
 * FOR ANY SPECIAL,  INDIRECT OR CONSEQUENTIAL  DAMAGES  OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF  USE, DATA OR PROFITS,  WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE  OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF OR  IN  CONNECTION   WITH  THE  USE OR  PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#ifndef BP_H
#define BP_H

/* The following structures must be defined, even if you don't
 * need them in you simulator
 */
typedef struct NetExtensionRec {	/* BIND */
  Boolean	debug ;		/* netParam: */
  int	errorFunction ;		/* netParam: 0=CE 1=SSE */
  Real	zeroErrorRadius ;	/* netParam: */
  Real	weightCost ;		/* netParam: */
  Real	unitCost ;		/* netParam: */

  /* adaptive weight melting */
  Real	standardWeight ;	/* netParam: */
  Real	weightCostIncrement ;	/* netParam: */
  Real	weightCostDecrement ;	/* netParam: */
  Real	lastError ;
  Real	aveError ;
  Real	aveErrorFactor ;	/* netParam: */
  Real	desiredError ;		/* netParam: */

  int	alphaOnset ;		/* netParam: */
  int	exampleNum ;
  Real	exampleFrequency ;
  int	sampleExamples ;	/* netParam: */
  Boolean	compiled ;
} NetExtensionRec ;

typedef struct GroupExtensionRec {	/* BIND */
  int	foo ;
} GroupExtensionRec ;

typedef struct UnitExtensionRec {	/* BIND */
  Boolean	lesioned ;
  int		numCompiledIncoming ;
  Real*		*preOutputs ;		
  Real*		*preOutputDerivs ;	
  Real		*weights ;		
  Real		*derivs ;		
} UnitExtensionRec ;

typedef struct LinkExtensionRec {	/* BIND */
  Boolean	lesioned ;
  Real		lesionedWeight ;
} LinkExtensionRec ;

typedef struct ExampleExtensionRec {	/* BIND */
  void	*data ;
} ExampleExtensionRec ;

#define MbatchSize(net)		((net)->batchSize)

#define Mdebug(net)		((net)->extension->debug)
#define MerrorFunction(net)	((net)->extension->errorFunction)
#define MzeroErrorRadius(net)	((net)->extension->zeroErrorRadius)
#define MweightCost(net)	((net)->extension->weightCost)
#define MunitCost(net) 		((net)->extension->unitCost)
#define MstandardWeight(net)	((net)->extension->standardWeight)
#define MweightCostIncrement(net) ((net)->extension->weightCostIncrement)
#define MweightCostDecrement(net) ((net)->extension->weightCostDecrement)
#define MlastError(net)		((net)->extension->lastError)
#define MaveError(net)		((net)->extension->aveError)
#define MaveErrorFactor(net)	((net)->extension->aveErrorFactor)
#define MdesiredError(net)	((net)->extension->desiredError)
#define MalphaOnset(net)	((net)->extension->alphaOnset)
#define MexampleNum(net)	((net)->extension->exampleNum)
#define MexampleFrequency(net)	((net)->extension->exampleFrequency)
#define MsampleExamples(net) 	((net)->extension->sampleExamples)

#define Mcompiled(net)		((net)->extension->compiled)

#define MnumCompiledIncoming(unit) ((unit)->extension->numCompiledIncoming)
#define MpreOutputs(unit)	((unit)->extension->preOutputs)
#define MpreOutputDerivs(unit)	((unit)->extension->preOutputDerivs)
#define Mweights(unit)		((unit)->extension->weights)
#define Mderivs(unit)		((unit)->extension->derivs)

#define Mlesioned(unit_or_link)	((unit_or_link)->extension->lesioned)
#define MlesionedWeight(link)	((link)->extension->lesionedWeight)

extern	struct TRACE	afterExample ; 		/* BIND */

/* compile-ffbp.c */
extern	void	netCompile		ARGS((Net net)) ;
extern	void	getCompiledWeights	ARGS((Unit unit, void* data)) ;
extern	void	putCompiledDerivs	ARGS((Unit unit, void* data)) ;

/* lesion-ffbp.c */
extern void	lesionUnit		ARGS((Unit unit)) ;
extern void	unlesionUnit		ARGS((Unit unit)) ;
extern Boolean	isLesionedUnit		ARGS((Unit unit)) ;
extern void	lesionLink		ARGS((Link link)) ;
extern void	unlesionLink		ARGS((Link link)) ;
extern Boolean	isLesionedLink		ARGS((Link link)) ;

#endif				/* BP_H */
